@@ -1,10 +1,10 @@
 /*
  *
- * d y n l o a d . c			-- All the stuff dealing with 
+ * d y n l o a d . c			-- All the stuff dealing with
  *					   dynamic loading
  *
- * Copyright  1993-1999 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ * Copyright © 1993-2005 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  *
  * Permission to use, copy, modify, distribute,and license this
  * software and its documentation for any purpose is hereby granted,
@@ -14,10 +14,10 @@
  * required for any of the authorized uses.
  * This software is provided ``AS IS'' without express or implied
  * warranty.
- * 
+ *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 23-Jan-1994 19:09
- * Last file update: 30-Nov-1999 12:51 (eg)
+ * Last file update: 11-Jan-2005 16:00 (eg)
  *
  * Win32 DLL support by Steve Pruitt <steve@pruitt.net>
  *
@@ -46,7 +46,7 @@
 #  include <dld.h>
 #endif
 
-#if defined(NETBSD1)	     
+#if defined(NETBSD)
 #  include <sys/types.h>
 #  include <nlist.h>
 #  include <link.h>
@@ -57,7 +57,7 @@
 #if defined(FREEBSD)
 #  define dlerror() "error unknown" /* dlerror() isn't implemented in FreeBSD too */
 #  define DLOPEN_COMPATIBLE
-#endif 
+#endif
 
 #ifdef HPUX
 #  include <dl.h>
@@ -98,7 +98,7 @@ static SCM cache_files = NULL;
 
 #include <libstack-Win32.h>
 
-#define dlerror() "error unknown" 
+#define dlerror() "error unknown"
 
 #define MAKE_STAT_PTR(p) (STk_make_Cpointer(ANONYMOUS_STAT_PTR_ID, (p), TRUE))
 #define MAKE_DYN_PTR(p)  (STk_make_Cpointer(ANONYMOUS_DYN_PTR_ID,  (p), FALSE))
@@ -110,7 +110,7 @@ HINSTANCE tclInstance;
 static void initialize_dynload(void)
 {
   void *handle;
-  
+
   handle = (void *) TclWinGetTclInstance();
   cache_files = LIST2(STk_makestring(""), MAKE_STAT_PTR(handle));
   STk_gc_protect(&cache_files);
@@ -136,14 +136,14 @@ static void *find_function(char *path, char *fname, int error_if_absent)
       fprintf(stderr, "find_function: cannot open object file : %s", dlerror());
     cache_files = Cons(str, Cons(MAKE_STAT_PTR(handle), cache_files));
   }
-  
+
   if ((fct = (void *) GetProcAddress(handle, fname)) == NULL && error_if_absent) {
     char msg[MAX_PATH_LENGTH];
-    
+
     sprintf(msg, "cannot find symbol ``%s'' in object file", fname);
     Serror(msg, str);
   }
-  
+
   return fct;
 }
 
@@ -164,7 +164,7 @@ static void load_and_call(char *path, char *fct_name)
   if (STk_member(str, cache_files) != STk_false) {
     Serror("module is already (dynamically) loaded", str);
   }
-  
+
   init_fct = find_function(path, fct_name, TRUE);
   (*init_fct)();
 }
@@ -172,7 +172,7 @@ static void load_and_call(char *path, char *fct_name)
 #endif /* MSC_VER WIN32 */
 
 
-#if defined(SUNOS4) || defined(SUNOS5) || defined(NETBSD1) || defined(FREEBSD) || defined(IRIX5) || defined(OSF1) ||defined(LINUX_ELF) ||defined(UNIXWARE)
+#if defined(SUNOS4) || defined(SUNOS5) || defined(NETBSD) || defined(FREEBSD) || defined(IRIX5) || defined(OSF1) ||defined(LINUX_ELF) ||defined(UNIXWARE)
 
 #define MAKE_STAT_PTR(p) (STk_make_Cpointer(ANONYMOUS_STAT_PTR_ID, (p), TRUE))
 #define MAKE_DYN_PTR(p)  (STk_make_Cpointer(ANONYMOUS_DYN_PTR_ID,  (p), FALSE))
@@ -181,10 +181,10 @@ static void load_and_call(char *path, char *fct_name)
 static void initialize_dynload(void)
 {
   void *handle;
-  
+
   if ((handle = (void *) dlopen(NULL, DYN_FLAG)) == NULL)
     Err("dynload: cannot initialize dynload.", STk_makestring(dlerror()));
-  
+
   cache_files = LIST2(STk_makestring(""), MAKE_STAT_PTR(handle));
   STk_gc_protect(&cache_files);
 }
@@ -209,14 +209,14 @@ static void *find_function(char *path, char *fname, int error_if_absent)
       fprintf(stderr, "find_function: cannot open object file : %s", dlerror());
     cache_files = Cons(str, Cons(MAKE_STAT_PTR(handle), cache_files));
   }
-  
+
   if ((fct = (void *) dlsym(handle, fname)) == NULL && error_if_absent) {
     char msg[MAX_PATH_LENGTH];
-    
+
     sprintf(msg, "cannot find symbol ``%s'' in object file", fname);
     Serror(msg, str);
   }
-  
+
   return fct;
 }
 
@@ -231,19 +231,19 @@ static void load_and_call(char *path, char *fct_name)
   /* Test if fct_name is already defined in the core interpreter */
   if ((init_fct = find_function("", fct_name, FALSE)) != NULL)
     Serror("module is already (statically) loaded", str);
-  
+
   if (STk_member(str, cache_files) != STk_false) {
     Serror("module is already (dynamically) loaded", str);
   }
-  
+
   init_fct = find_function(path, fct_name, TRUE);
   (*init_fct)();
 }
 #endif
 
 #if defined(LINUX_DLD)
-/* 
- * This code is for Linux, using the dld package. This code should not be used 
+/*
+ * This code is for Linux, using the dld package. This code should not be used
  * anymore when ELF will be completely accepted under Linux. In the meanwhile...
  *
  * This code is inspired from a code sent by Patrick Nguyen pnguyen@elde.epfl.ch.
@@ -272,25 +272,25 @@ static void load_and_call(char *path, char *fct_name)
   if(!dld_already_initialized) {
     if (dld_init (dld_find_executable (STk_Argv0)))
       dld_perror("dld: failed to init dld");
-    else 
+    else
       dld_already_initialized = TRUE;
   }
-  
+
   /* Load file */
   if (dld_link(path)) dld_perror("dld: cannot link");
 
   /* And get a pointer on function "fct_name" */
   init_fct = (void (*) ()) dld_get_func(fct_name);
   if (init_fct) {
-    /* When loading a function. It can induce some unresolved references 
+    /* When loading a function. It can induce some unresolved references
      * Most of the time, the unresolved references come from fucntions
      * which are in the libc but which are not used by the core interpreter.
      * For instance, if loaded module uses fork, which is not loaded in
      * the interpreter, we will have an unresolve reference for _fork.
      * To avoid this problem, we make again a link against standard libc
-     * Note: in general situation is more complicated than this, unresolved 
+     * Note: in general situation is more complicated than this, unresolved
      * symbols could be elsewhere than the libc....
-     * Those situation are not handled by current code, but I hope that ELF 
+     * Those situation are not handled by current code, but I hope that ELF
      * will supplant this way of doing dynamic loading soon...
      */
     if (dld_function_executable_p (fct_name))
@@ -298,8 +298,8 @@ static void load_and_call(char *path, char *fct_name)
       (*init_fct) ();
     else {
       /* Function is notexecutable = we have unresolved references */
-      if (dld_link("/usr/lib/libc.a") && dld_link("/usr/lib/libc.sa")) 
-	dld_perror("dld: cannot link"); 
+      if (dld_link("/usr/lib/libc.a") && dld_link("/usr/lib/libc.sa"))
+	dld_perror("dld: cannot link");
 
       if (dld_function_executable_p (fct_name)) {
 	init_fct = (void (*) ()) dld_get_func(fct_name);
@@ -307,11 +307,11 @@ static void load_and_call(char *path, char *fct_name)
       }
       else {
 	char **unresolved;
-	extern int dld_undefined_sym_count; 
-	int i;	  
+	extern int dld_undefined_sym_count;
+	int i;
 	Fprintf(STk_curr_eport, "dld: function %s not executable!\n", fct_name);
 	Fprintf(STk_curr_eport, "Unresolved symbols are:\n");
-	
+
 	unresolved= dld_list_undefined_sym();
 	for (i = 0; i < dld_undefined_sym_count; i++)
 	  Fprintf(STk_curr_eport, "\t%s\n",unresolved[i]);
@@ -335,14 +335,14 @@ static void load_and_call(char *path, char *fct_name)
 {
   shl_t handle;
   void (*init_fct)();
- 
-  if ((handle = shl_load(path, BIND_IMMEDIATE | BIND_VERBOSE, 0L)) == NULL) 
+
+  if ((handle = shl_load(path, BIND_IMMEDIATE | BIND_VERBOSE, 0L)) == NULL)
     Err("Cannot open file", STk_makestring(path));
-  
+
   handle = NULL;
   if (shl_findsym(&handle, fct_name, TYPE_PROCEDURE, &init_fct) == -1) {
     char msg[MAX_PATH_LENGTH];
-    
+
     sprintf(msg, "Cannot find function %s in object file", fct_name);
     Err(msg, STk_nil);
   }
@@ -360,7 +360,7 @@ void STk_load_object_file(char *path)
   for (p = path, slash = p-1; *p; p++)		/* Find position of last '/' */
     if (*p == '/') slash = p;
 
-#if defined(NETBSD1)
+#if defined(NETBSB)
   sprintf(fct_name, "_STk_init_%s", slash + 1);
 #else
   sprintf(fct_name, "STk_init_%s", slash + 1);
@@ -372,7 +372,7 @@ void STk_load_object_file(char *path)
   load_and_call(path, fct_name);
 }
 
-#if defined(CYGWIN32)		
+#if defined(CYGWIN32)
 
 #define MAKE_STAT_PTR(p) STk_nil
 #define MAKE_DYN_PTR(p)  STk_nil
@@ -381,7 +381,7 @@ void STk_load_object_file(char *path)
 static void initialize_dynload(void)
 {
   /* CYGWIN32 */
-  Err("dynload: cannot initialize dynload.", STk_makestring(dlerror())); 
+  Err("dynload: cannot initialize dynload.", STk_makestring(dlerror()));
 }
 
 
@@ -401,7 +401,7 @@ static void *find_function(char *path, char *fname, int error_if_absent)
 #endif
 
 /******************************************************************************
- *									       
+ *
  * FFI support
  *
  ******************************************************************************/
@@ -488,8 +488,8 @@ static void push_argument(char *proc_name, SCM value, SCM name, int type)
 		     }
   		     goto Error;
 
-    case EXT_STRING: 
-    case EXT_STAT_PTR: 
+    case EXT_STRING:
+    case EXT_STAT_PTR:
     case EXT_DYN_PTR: if (CPOINTERP(value)) {
       			res = push_ptr(EXTDATA(value));
 			break;
@@ -507,7 +507,7 @@ static void push_argument(char *proc_name, SCM value, SCM name, int type)
     Serror("too many values pushed on the stack", STk_nil);
   }
   return;
-Error: 
+Error:
   {
     char message[300];
     sprintf(message, "argument ``%s'' has a bad type", PNAME(name));
@@ -520,7 +520,7 @@ static void push_list(char *proc_name, SCM l)
   if (!NULLP(l)) {
     SCM x = CAR(l);
     int type;
-    
+
     if (CHARP(x))	  type = EXT_CHAR;
     else if (INTP(x)) 	  type = EXT_LONG;
     else if (FLONUMP(x))  type = EXT_DOUBLE;
@@ -584,7 +584,7 @@ PRIMITIVE STk_external_existsp(SCM entry_name, SCM library)
 
   if (NSTRINGP(entry_name)) Serror("bad string", entry_name);
   if (NSTRINGP(library))    Serror("bad string", library);
-  
+
   return find_function(CHARS(library), CHARS(entry_name), FALSE) ? STk_true : STk_false;
 }
 
@@ -594,10 +594,10 @@ PRIMITIVE STk_call_external(SCM l, int len)
   SCM libname, entryname, rettype, argnames, argtypes;
   char *c_entryname, *c_libname;
   int  c_rettype;
-  
+
   if (len < 5) Serror("not enough arguments", l);
 
-  libname   = CAR(l); l = CDR(l); 
+  libname   = CAR(l); l = CDR(l);
   entryname = CAR(l); l = CDR(l);
   rettype   = CAR(l); l = CDR(l);
   argnames  = CAR(l); l = CDR(l);
@@ -608,7 +608,7 @@ PRIMITIVE STk_call_external(SCM l, int len)
   if (!STRINGP(entryname)) Serror("bad entry name", entryname);
   if (!INTEGERP(rettype))  Serror("bad return type", rettype);
 
-  c_rettype   = STk_integer_value(rettype); 
+  c_rettype   = STk_integer_value(rettype);
   c_entryname = CHARS(entryname);
   c_libname   = CHARS(libname);
   init_ext_call();
@@ -627,9 +627,9 @@ PRIMITIVE STk_call_external(SCM l, int len)
     if (NULLP(l))
       /* no actual arguments and list of names is not terminated */
       STk_procedure_error(c_entryname, "not enought arguments", argnames);
-    
+
     /* Standard case */
-    push_argument(c_entryname, CAR(l), CAR(argnames), 
+    push_argument(c_entryname, CAR(l), CAR(argnames),
 		  STk_integer_value_no_overflow(CAR(argtypes)));
     l 	     = CDR(l);
     argnames = CDR(argnames);
@@ -643,13 +643,13 @@ PRIMITIVE STk_cstring2string(SCM pointer)
 {
   static char *proc_name = "c-string->string";
   char *str;
-  
+
   if (STRINGP(pointer)) str = CHARS(pointer);
   else
     if (CPOINTERP(pointer)) str = (char *) EXTDATA(pointer);
     else
       Serror("bad strng or C pointer", pointer);
-  
+
   return STk_makestring(str);
 }
 
@@ -674,7 +674,7 @@ PRIMITIVE STk_cstring2string(SCM pointer)
   ENTER_PRIMITIVE("c-string->string");
   Serror(msg, STk_nil);
 }
-#endif 
+#endif
 
 
 
