@@ -28,7 +28,7 @@
  * must trace the complex_descriptor.
  *
  * Note that descriptors inside objects may appear cleared, if we encounter a
- * false refrence to an object on a free list.  In the GC_descr case, this
+ * false reference to an object on a free list.  In the GC_descr case, this
  * is OK, since a 0 descriptor corresponds to examining no fields.
  * In the complex_descriptor case, we explicitly check for that case.
  *
@@ -36,18 +36,19 @@
  * since they are not accessible through the current interface.
  */
 
-#include "private/gc_priv.h"
-#include "private/gc_mark.h"
+#include "private/gc_pmark.h"
 #include "gc_typed.h"
 
 # define TYPD_EXTRA_BYTES (sizeof(word) - EXTRA_BYTES)
 
-GC_bool GC_explicit_typing_initialized = FALSE;
+STATIC GC_bool GC_explicit_typing_initialized = FALSE;
 
-int GC_explicit_kind;	/* Object kind for objects with indirect	*/
+STATIC int GC_explicit_kind;
+			/* Object kind for objects with indirect	*/
 			/* (possibly extended) descriptors.		*/
 
-int GC_array_kind;	/* Object kind for objects with complex		*/
+STATIC int GC_array_kind;
+			/* Object kind for objects with complex		*/
 			/* descriptors and GC_array_mark_proc.		*/
 
 /* Extended descriptors.  GC_typed_mark_proc understands these.	*/
@@ -66,16 +67,16 @@ typedef union ComplexDescriptor {
     struct LeafDescriptor {	/* Describes simple array	*/
         word ld_tag;
 #	define LEAF_TAG 1
-	word ld_size;		/* bytes per element	*/
+	size_t ld_size;		/* bytes per element	*/
 				/* multiple of ALIGNMENT	*/
-	word ld_nelements;	/* Number of elements.	*/
+	size_t ld_nelements;	/* Number of elements.	*/
 	GC_descr ld_descriptor; /* A simple length, bitmap,	*/
 				/* or procedure descriptor.	*/
     } ld;
     struct ComplexArrayDescriptor {
         word ad_tag;
 #	define ARRAY_TAG 2
-	word ad_nelements;
+	size_t ad_nelements;
 	union ComplexDescriptor * ad_element_descr;
     } ad;
     struct SequenceDescriptor {
@@ -87,50 +88,52 @@ typedef union ComplexDescriptor {
 } complex_descriptor;
 #define TAG ld.ld_tag
 
-ext_descr * GC_ext_descriptors;	/* Points to array of extended 	*/
-				/* descriptors.			*/
+STATIC ext_descr * GC_ext_descriptors;	/* Points to array of extended 	*/
+					/* descriptors.			*/
 
-word GC_ed_size = 0;	/* Current size of above arrays.	*/
+STATIC size_t GC_ed_size = 0;	/* Current size of above arrays.	*/
 # define ED_INITIAL_SIZE 100;
 
-word GC_avail_descr = 0;	/* Next available slot.		*/
+STATIC size_t GC_avail_descr = 0;	/* Next available slot.		*/
 
-int GC_typed_mark_proc_index;	/* Indices of my mark		*/
-int GC_array_mark_proc_index;	/* procedures.			*/
+STATIC int GC_typed_mark_proc_index;	/* Indices of my mark		*/
+STATIC int GC_array_mark_proc_index;	/* procedures.			*/
+
+static void GC_push_typed_structures_proc (void)
+{
+  GC_push_all((ptr_t)&GC_ext_descriptors, (ptr_t)&GC_ext_descriptors + sizeof(word));
+}
 
 /* Add a multiword bitmap to GC_ext_descriptors arrays.  Return	*/
 /* starting index.						*/
 /* Returns -1 on failure.					*/
 /* Caller does not hold allocation lock.			*/
-signed_word GC_add_ext_descriptor(bm, nbits)
-GC_bitmap bm;
-word nbits;
+STATIC signed_word GC_add_ext_descriptor(GC_bitmap bm, word nbits)
 {
-    register size_t nwords = divWORDSZ(nbits + WORDSZ-1);
-    register signed_word result;
-    register word i;
-    register word last_part;
-    register int extra_bits;
+    size_t nwords = divWORDSZ(nbits + WORDSZ-1);
+    signed_word result;
+    size_t i;
+    word last_part;
+    size_t extra_bits;
     DCL_LOCK_STATE;
 
-    DISABLE_SIGNALS();
     LOCK();
     while (GC_avail_descr + nwords >= GC_ed_size) {
     	ext_descr * new;
     	size_t new_size;
     	word ed_size = GC_ed_size;
     	
-    	UNLOCK();
-        ENABLE_SIGNALS();
     	if (ed_size == 0) {
+    	    GC_push_typed_structures = GC_push_typed_structures_proc;
+    	    UNLOCK();
     	    new_size = ED_INITIAL_SIZE;
     	} else {
+    	    UNLOCK();
     	    new_size = 2 * ed_size;
     	    if (new_size > MAX_ENV) return(-1);
     	} 
     	new = (ext_descr *) GC_malloc_atomic(new_size * sizeof(ext_descr));
     	if (new == 0) return(-1);
-    	DISABLE_SIGNALS();
         LOCK();
         if (ed_size == GC_ed_size) {
             if (GC_avail_descr != 0) {
@@ -155,7 +158,6 @@ word nbits;
     GC_ext_descriptors[result + i].ed_continued = FALSE;
     GC_avail_descr += nwords;
     UNLOCK();
-    ENABLE_SIGNALS();
     return(result);
 }
 
@@ -166,19 +168,19 @@ GC_descr GC_bm_table[WORDSZ/2];
 /* each of which is described by descriptor.				*/
 /* The result is known to be short enough to fit into a bitmap		*/
 /* descriptor.								*/
-/* Descriptor is a DS_LENGTH or DS_BITMAP descriptor.			*/
-GC_descr GC_double_descr(descriptor, nwords)
-register GC_descr descriptor;
-register word nwords;
+/* Descriptor is a GC_DS_LENGTH or GC_DS_BITMAP descriptor.		*/
+STATIC GC_descr GC_double_descr(GC_descr descriptor, word nwords)
 {
-    if ((descriptor & DS_TAGS) == DS_LENGTH) {
+    if ((descriptor & GC_DS_TAGS) == GC_DS_LENGTH) {
         descriptor = GC_bm_table[BYTES_TO_WORDS((word)descriptor)];
     };
-    descriptor |= (descriptor & ~DS_TAGS) >> nwords;
+    descriptor |= (descriptor & ~GC_DS_TAGS) >> nwords;
     return(descriptor);
 }
 
-complex_descriptor * GC_make_sequence_descriptor();
+STATIC complex_descriptor *
+GC_make_sequence_descriptor(complex_descriptor *first,
+			    complex_descriptor *second);
 
 /* Build a descriptor for an array with nelements elements,	*/
 /* each of which can be described by a simple descriptor.	*/
@@ -192,28 +194,24 @@ complex_descriptor * GC_make_sequence_descriptor();
 /* is returned in *simple_d.					*/
 /* If the result is NO_MEM, then				*/
 /* we failed to allocate the descriptor.			*/
-/* The implementation knows that DS_LENGTH is 0.		*/
+/* The implementation knows that GC_DS_LENGTH is 0.		*/
 /* *leaf, *complex_d, and *simple_d may be used as temporaries	*/
 /* during the construction.					*/
 # define COMPLEX 2
 # define LEAF 1
 # define SIMPLE 0
 # define NO_MEM (-1)
-int GC_make_array_descriptor(nelements, size, descriptor,
-			     simple_d, complex_d, leaf)
-word size;
-word nelements;
-GC_descr descriptor;
-GC_descr *simple_d;
-complex_descriptor **complex_d;
-struct LeafDescriptor * leaf;
+STATIC int GC_make_array_descriptor(size_t nelements, size_t size,
+				    GC_descr descriptor, GC_descr *simple_d,
+				    complex_descriptor **complex_d,
+				    struct LeafDescriptor * leaf)
 {
 #   define OPT_THRESHOLD 50
 	/* For larger arrays, we try to combine descriptors of adjacent	*/
 	/* descriptors to speed up marking, and to reduce the amount	*/
 	/* of space needed on the mark stack.				*/
-    if ((descriptor & DS_TAGS) == DS_LENGTH) {
-      if ((word)descriptor == size) {
+    if ((descriptor & GC_DS_TAGS) == GC_DS_LENGTH) {
+      if (descriptor == (GC_descr)size) {
     	*simple_d = nelements * descriptor;
     	return(SIMPLE);
       } else if ((word)descriptor == 0) {
@@ -232,7 +230,7 @@ struct LeafDescriptor * leaf;
         }
       }
     } else if (size <= BITMAP_BITS/2
-    	       && (descriptor & DS_TAGS) != DS_PROC
+    	       && (descriptor & GC_DS_TAGS) != GC_DS_PROC
     	       && (size & (sizeof(word)-1)) == 0) {
       int result =      
           GC_make_array_descriptor(nelements/2, 2*size,
@@ -299,9 +297,9 @@ struct LeafDescriptor * leaf;
     }
 }
 
-complex_descriptor * GC_make_sequence_descriptor(first, second)
-complex_descriptor * first;
-complex_descriptor * second;
+STATIC complex_descriptor *
+GC_make_sequence_descriptor(complex_descriptor *first,
+			    complex_descriptor *second)
 {
     struct SequenceDescriptor * result =
         (struct SequenceDescriptor *)
@@ -318,9 +316,8 @@ complex_descriptor * second;
 }
 
 #ifdef UNDEFINED
-complex_descriptor * GC_make_complex_array_descriptor(nelements, descr)
-word nelements;
-complex_descriptor * descr;
+complex_descriptor * GC_make_complex_array_descriptor(word nelements,
+						      complex_descriptor *descr)
 {
     struct ComplexArrayDescriptor * result =
         (struct ComplexArrayDescriptor *)
@@ -335,112 +332,70 @@ complex_descriptor * descr;
 }
 #endif
 
-ptr_t * GC_eobjfreelist;
+STATIC ptr_t * GC_eobjfreelist;
 
-ptr_t * GC_arobjfreelist;
+STATIC ptr_t * GC_arobjfreelist;
 
-mse * GC_typed_mark_proc GC_PROTO((register word * addr,
-				   register mse * mark_stack_ptr,
-				   mse * mark_stack_limit,
-				   word env));
+STATIC mse * GC_typed_mark_proc(word * addr, mse * mark_stack_ptr,
+				mse * mark_stack_limit, word env);
 
-mse * GC_array_mark_proc GC_PROTO((register word * addr,
-				   register mse * mark_stack_ptr,
-				   mse * mark_stack_limit,
-				   word env));
-
-GC_descr GC_generic_array_descr;
+STATIC mse * GC_array_mark_proc(word * addr, mse * mark_stack_ptr,
+				mse * mark_stack_limit, word env);
 
 /* Caller does not hold allocation lock. */
-void GC_init_explicit_typing()
+STATIC void GC_init_explicit_typing(void)
 {
     register int i;
     DCL_LOCK_STATE;
 
-    
-#   ifdef PRINTSTATS
-     	if (sizeof(struct LeafDescriptor) % sizeof(word) != 0)
-     	    ABORT("Bad leaf descriptor size");
-#   endif
-    DISABLE_SIGNALS();
+    GC_STATIC_ASSERT(sizeof(struct LeafDescriptor) % sizeof(word) == 0);
     LOCK();
     if (GC_explicit_typing_initialized) {
       UNLOCK();
-      ENABLE_SIGNALS();
       return;
     }
     GC_explicit_typing_initialized = TRUE;
     /* Set up object kind with simple indirect descriptor. */
-      GC_eobjfreelist = (ptr_t *)
-          GC_INTERNAL_MALLOC((MAXOBJSZ+1)*sizeof(ptr_t), PTRFREE);
-      if (GC_eobjfreelist == 0) ABORT("Couldn't allocate GC_eobjfreelist");
-      BZERO(GC_eobjfreelist, (MAXOBJSZ+1)*sizeof(ptr_t));
-      GC_explicit_kind = GC_n_kinds++;
-      GC_obj_kinds[GC_explicit_kind].ok_freelist = GC_eobjfreelist;
-      GC_obj_kinds[GC_explicit_kind].ok_reclaim_list = 0;
-      GC_obj_kinds[GC_explicit_kind].ok_descriptor =
-    		(((word)WORDS_TO_BYTES(-1)) | DS_PER_OBJECT);
-      GC_obj_kinds[GC_explicit_kind].ok_relocate_descr = TRUE;
-      GC_obj_kinds[GC_explicit_kind].ok_init = TRUE;
+      GC_eobjfreelist = (ptr_t *)GC_new_free_list_inner();
+      GC_explicit_kind = GC_new_kind_inner(
+		      	    (void **)GC_eobjfreelist,
+		      	    (((word)WORDS_TO_BYTES(-1)) | GC_DS_PER_OBJECT),
+			    TRUE, TRUE);
     		/* Descriptors are in the last word of the object. */
-      GC_typed_mark_proc_index = GC_n_mark_procs;
-      GC_mark_procs[GC_typed_mark_proc_index] = GC_typed_mark_proc;
-      GC_n_mark_procs++;
-        /* Moving this up breaks DEC AXP compiler.      */
+      GC_typed_mark_proc_index = GC_new_proc_inner(GC_typed_mark_proc);
     /* Set up object kind with array descriptor. */
-      GC_arobjfreelist = (ptr_t *)
-          GC_INTERNAL_MALLOC((MAXOBJSZ+1)*sizeof(ptr_t), PTRFREE);
-      if (GC_arobjfreelist == 0) ABORT("Couldn't allocate GC_arobjfreelist");
-      BZERO(GC_arobjfreelist, (MAXOBJSZ+1)*sizeof(ptr_t));
-      if (GC_n_mark_procs >= MAX_MARK_PROCS)
-      		ABORT("No slot for array mark proc");
-      GC_array_mark_proc_index = GC_n_mark_procs++;
-      if (GC_n_kinds >= MAXOBJKINDS)
-      		ABORT("No kind available for array objects");
-      GC_array_kind = GC_n_kinds++;
-      GC_obj_kinds[GC_array_kind].ok_freelist = GC_arobjfreelist;
-      GC_obj_kinds[GC_array_kind].ok_reclaim_list = 0;
-      GC_obj_kinds[GC_array_kind].ok_descriptor =
-    		MAKE_PROC(GC_array_mark_proc_index, 0);;
-      GC_obj_kinds[GC_array_kind].ok_relocate_descr = FALSE;
-      GC_obj_kinds[GC_array_kind].ok_init = TRUE;
-    		/* Descriptors are in the last word of the object. */
-            GC_mark_procs[GC_array_mark_proc_index] = GC_array_mark_proc;
+      GC_arobjfreelist = (ptr_t *)GC_new_free_list_inner();
+      GC_array_mark_proc_index = GC_new_proc_inner(GC_array_mark_proc);
+      GC_array_kind = GC_new_kind_inner(
+		      	    (void **)GC_arobjfreelist,
+			    GC_MAKE_PROC(GC_array_mark_proc_index, 0),
+			    FALSE, TRUE);
       for (i = 0; i < WORDSZ/2; i++) {
           GC_descr d = (((word)(-1)) >> (WORDSZ - i)) << (WORDSZ - i);
-          d |= DS_BITMAP;
+          d |= GC_DS_BITMAP;
           GC_bm_table[i] = d;
       }
-      GC_generic_array_descr = MAKE_PROC(GC_array_mark_proc_index, 0); 
     UNLOCK();
-    ENABLE_SIGNALS();
 }
 
-# if defined(__STDC__) || defined(__cplusplus)
-    mse * GC_typed_mark_proc(register word * addr,
-			     register mse * mark_stack_ptr,
-			     mse * mark_stack_limit,
-			     word env)
-# else
-    mse * GC_typed_mark_proc(addr, mark_stack_ptr, mark_stack_limit, env)
-    register word * addr;
-    register mse * mark_stack_ptr;
-    mse * mark_stack_limit;
-    word env;
-# endif
+STATIC mse * GC_typed_mark_proc(word * addr, mse * mark_stack_ptr,
+				mse * mark_stack_limit, word env)
 {
-    register word bm = GC_ext_descriptors[env].ed_bitmap;
-    register word * current_p = addr;
-    register word current;
-    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
-    register ptr_t least_ha = GC_least_plausible_heap_addr;
-    
+    word bm = GC_ext_descriptors[env].ed_bitmap;
+    word * current_p = addr;
+    word current;
+    ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
+    ptr_t least_ha = GC_least_plausible_heap_addr;
+    DECLARE_HDR_CACHE;
+
+    INIT_HDR_CACHE;
     for (; bm != 0; bm >>= 1, current_p++) {
     	if (bm & 1) {
     	    current = *current_p;
+	    FIXUP_POINTER(current);
     	    if ((ptr_t)current >= least_ha && (ptr_t)current <= greatest_ha) {
     	        PUSH_CONTENTS((ptr_t)current, mark_stack_ptr,
-			      mark_stack_limit, current_p, exit1);
+			      mark_stack_limit, (ptr_t)current_p, exit1);
     	    }
     	}
     }
@@ -453,9 +408,9 @@ void GC_init_explicit_typing()
         if (mark_stack_ptr >= mark_stack_limit) {
             mark_stack_ptr = GC_signal_mark_stack_overflow(mark_stack_ptr);
         }
-        mark_stack_ptr -> mse_start = addr + WORDSZ;
+        mark_stack_ptr -> mse_start = (ptr_t)(addr + WORDSZ);
         mark_stack_ptr -> mse_descr =
-        	MAKE_PROC(GC_typed_mark_proc_index, env+1);
+        	GC_MAKE_PROC(GC_typed_mark_proc_index, env+1);
     }
     return(mark_stack_ptr);
 }
@@ -463,8 +418,7 @@ void GC_init_explicit_typing()
 /* Return the size of the object described by d.  It would be faster to	*/
 /* store this directly, or to compute it as part of			*/
 /* GC_push_complex_descriptor, but hopefully it doesn't matter.		*/
-word GC_descr_obj_size(d)
-register complex_descriptor *d;
+STATIC word GC_descr_obj_size(complex_descriptor *d)
 {
     switch(d -> TAG) {
       case LEAF_TAG:
@@ -483,11 +437,8 @@ register complex_descriptor *d;
 
 /* Push descriptors for the object at addr with complex descriptor d	*/
 /* onto the mark stack.  Return 0 if the mark stack overflowed.  	*/
-mse * GC_push_complex_descriptor(addr, d, msp, msl)
-word * addr;
-register complex_descriptor *d;
-register mse * msp;
-mse * msl;
+STATIC mse * GC_push_complex_descriptor(word *addr, complex_descriptor *d,
+					mse *msp, mse *msl)
 {
     register ptr_t current = (ptr_t) addr;
     register word nelements;
@@ -504,7 +455,7 @@ mse * msl;
           sz = d -> ld.ld_size;
           for (i = 0; i < nelements; i++) {
               msp++;
-              msp -> mse_start = (word *)current;
+              msp -> mse_start = current;
               msp -> mse_descr = descr;
               current += sz;
           }
@@ -542,22 +493,13 @@ mse * msl;
 }
 
 /*ARGSUSED*/
-# if defined(__STDC__) || defined(__cplusplus)
-    mse * GC_array_mark_proc(register word * addr,
-			     register mse * mark_stack_ptr,
-			     mse * mark_stack_limit,
-			     word env)
-# else
-    mse * GC_array_mark_proc(addr, mark_stack_ptr, mark_stack_limit, env)
-    register word * addr;
-    register mse * mark_stack_ptr;
-    mse * mark_stack_limit;
-    word env;
-# endif
+STATIC mse * GC_array_mark_proc(word * addr, mse * mark_stack_ptr,
+				mse * mark_stack_limit, word env)
 {
-    register hdr * hhdr = HDR(addr);
-    register word sz = hhdr -> hb_sz;
-    register complex_descriptor * descr = (complex_descriptor *)(addr[sz-1]);
+    hdr * hhdr = HDR(addr);
+    size_t sz = hhdr -> hb_sz;
+    size_t nwords = BYTES_TO_WORDS(sz);
+    complex_descriptor * descr = (complex_descriptor *)(addr[nwords-1]);
     mse * orig_mark_stack_ptr = mark_stack_ptr;
     mse * new_mark_stack_ptr;
     
@@ -578,28 +520,22 @@ mse * msl;
     	/* the original array entry.					*/
     	GC_mark_stack_too_small = TRUE;
     	new_mark_stack_ptr = orig_mark_stack_ptr + 1;
-    	new_mark_stack_ptr -> mse_start = addr;
-    	new_mark_stack_ptr -> mse_descr = WORDS_TO_BYTES(sz) | DS_LENGTH;
+    	new_mark_stack_ptr -> mse_start = (ptr_t)addr;
+    	new_mark_stack_ptr -> mse_descr = sz | GC_DS_LENGTH;
     } else {
         /* Push descriptor itself */
         new_mark_stack_ptr++;
-        new_mark_stack_ptr -> mse_start = addr + sz - 1;
-        new_mark_stack_ptr -> mse_descr = sizeof(word) | DS_LENGTH;
+        new_mark_stack_ptr -> mse_start = (ptr_t)(addr + nwords - 1);
+        new_mark_stack_ptr -> mse_descr = sizeof(word) | GC_DS_LENGTH;
     }
-    return(new_mark_stack_ptr);
+    return new_mark_stack_ptr;
 }
 
-#if defined(__STDC__) || defined(__cplusplus)
-  GC_descr GC_make_descriptor(GC_bitmap bm, size_t len)
-#else
-  GC_descr GC_make_descriptor(bm, len)
-  GC_bitmap bm;
-  size_t len;
-#endif
+GC_API GC_descr GC_CALL GC_make_descriptor(GC_bitmap bm, size_t len)
 {
-    register signed_word last_set_bit = len - 1;
-    register word result;
-    register int i;
+    signed_word last_set_bit = len - 1;
+    GC_descr result;
+    signed_word i;
 #   define HIGH_BIT (((word)1) << (WORDSZ - 1))
     
     if (!GC_explicit_typing_initialized) GC_init_explicit_typing();
@@ -616,7 +552,7 @@ mse * msl;
       }
       if (all_bits_set) {
     	/* An initial section contains all pointers.  Use length descriptor. */
-        return(WORDS_TO_BYTES(last_set_bit+1) | DS_LENGTH);
+        return (WORDS_TO_BYTES(last_set_bit+1) | GC_DS_LENGTH);
       }
     }
 #   endif
@@ -628,131 +564,103 @@ mse * msl;
     	    result >>= 1;
     	    if (GC_get_bit(bm, i)) result |= HIGH_BIT;
     	}
-    	result |= DS_BITMAP;
+    	result |= GC_DS_BITMAP;
     	return(result);
     } else {
     	signed_word index;
     	
     	index = GC_add_ext_descriptor(bm, (word)last_set_bit+1);
-    	if (index == -1) return(WORDS_TO_BYTES(last_set_bit+1) | DS_LENGTH);
+    	if (index == -1) return(WORDS_TO_BYTES(last_set_bit+1) | GC_DS_LENGTH);
     				/* Out of memory: use conservative	*/
     				/* approximation.			*/
-    	result = MAKE_PROC(GC_typed_mark_proc_index, (word)index);
-    	return(result);
+    	result = GC_MAKE_PROC(GC_typed_mark_proc_index, (word)index);
+    	return result;
     }
 }
 
-ptr_t GC_clear_stack();
+void * GC_clear_stack(void *);
 
 #define GENERAL_MALLOC(lb,k) \
-    (GC_PTR)GC_clear_stack(GC_generic_malloc((word)lb, k))
+    (void *)GC_clear_stack(GC_generic_malloc((word)lb, k))
     
 #define GENERAL_MALLOC_IOP(lb,k) \
-    (GC_PTR)GC_clear_stack(GC_generic_malloc_ignore_off_page(lb, k))
+    (void *)GC_clear_stack(GC_generic_malloc_ignore_off_page(lb, k))
 
-#if defined(__STDC__) || defined(__cplusplus)
-  void * GC_malloc_explicitly_typed(size_t lb, GC_descr d)
-#else
-  char * GC_malloc_explicitly_typed(lb, d)
-  size_t lb;
-  GC_descr d;
-#endif
+GC_API void * GC_CALL GC_malloc_explicitly_typed(size_t lb, GC_descr d)
 {
-register ptr_t op;
-register ptr_t * opp;
-register word lw;
-DCL_LOCK_STATE;
+    ptr_t op;
+    ptr_t * opp;
+    size_t lg;
+    DCL_LOCK_STATE;
 
     lb += TYPD_EXTRA_BYTES;
-    if( SMALL_OBJ(lb) ) {
-#       ifdef MERGE_SIZES
-	  lw = GC_size_map[lb];
-#	else
-	  lw = ALIGNED_WORDS(lb);
-#       endif
-	opp = &(GC_eobjfreelist[lw]);
-	FASTLOCK();
-        if( !FASTLOCK_SUCCEEDED() || (op = *opp) == 0 ) {
-            FASTUNLOCK();
+    if(SMALL_OBJ(lb)) {
+	lg = GC_size_map[lb];
+	opp = &(GC_eobjfreelist[lg]);
+	LOCK();
+        if( (op = *opp) == 0 ) {
+            UNLOCK();
             op = (ptr_t)GENERAL_MALLOC((word)lb, GC_explicit_kind);
-	    if (0 == op) return(0);
-#	    ifdef MERGE_SIZES
-		lw = GC_size_map[lb];	/* May have been uninitialized.	*/            
-#	    endif
+	    if (0 == op) return 0;
+	    lg = GC_size_map[lb];	/* May have been uninitialized.	*/
         } else {
             *opp = obj_link(op);
 	    obj_link(op) = 0;
-            GC_words_allocd += lw;
-            FASTUNLOCK();
+            GC_bytes_allocd += GRANULES_TO_BYTES(lg);
+            UNLOCK();
         }
+	((word *)op)[GRANULES_TO_WORDS(lg) - 1] = d;
    } else {
        op = (ptr_t)GENERAL_MALLOC((word)lb, GC_explicit_kind);
-       if (op != NULL)
-	    lw = BYTES_TO_WORDS(GC_size(op));
+       if (op != NULL) {
+	    lg = BYTES_TO_GRANULES(GC_size(op));
+	    ((word *)op)[GRANULES_TO_WORDS(lg) - 1] = d;
+       }
    }
-   if (op != NULL)
-       ((word *)op)[lw - 1] = d;
-   return((GC_PTR) op);
+   return((void *) op);
 }
 
-#if defined(__STDC__) || defined(__cplusplus)
-  void * GC_malloc_explicitly_typed_ignore_off_page(size_t lb, GC_descr d)
-#else
-  char * GC_malloc_explicitly_typed_ignore_off_page(lb, d)
-  size_t lb;
-  GC_descr d;
-#endif
+GC_API void * GC_CALL GC_malloc_explicitly_typed_ignore_off_page(size_t lb,
+								GC_descr d)
 {
-register ptr_t op;
-register ptr_t * opp;
-register word lw;
+ptr_t op;
+ptr_t * opp;
+size_t lg;
 DCL_LOCK_STATE;
 
     lb += TYPD_EXTRA_BYTES;
     if( SMALL_OBJ(lb) ) {
-#       ifdef MERGE_SIZES
-	  lw = GC_size_map[lb];
-#	else
-	  lw = ALIGNED_WORDS(lb);
-#       endif
-	opp = &(GC_eobjfreelist[lw]);
-	FASTLOCK();
-        if( !FASTLOCK_SUCCEEDED() || (op = *opp) == 0 ) {
-            FASTUNLOCK();
+	lg = GC_size_map[lb];
+	opp = &(GC_eobjfreelist[lg]);
+	LOCK();
+        if( (op = *opp) == 0 ) {
+            UNLOCK();
             op = (ptr_t)GENERAL_MALLOC_IOP(lb, GC_explicit_kind);
-#	    ifdef MERGE_SIZES
-		lw = GC_size_map[lb];	/* May have been uninitialized.	*/            
-#	    endif
+	    if (0 == op) return 0;
+	    lg = GC_size_map[lb];	/* May have been uninitialized.	*/
         } else {
             *opp = obj_link(op);
 	    obj_link(op) = 0;
-            GC_words_allocd += lw;
-            FASTUNLOCK();
+            GC_bytes_allocd += GRANULES_TO_BYTES(lg);
+            UNLOCK();
         }
+	((word *)op)[GRANULES_TO_WORDS(lg) - 1] = d;
    } else {
        op = (ptr_t)GENERAL_MALLOC_IOP(lb, GC_explicit_kind);
-       if (op != NULL)
-       lw = BYTES_TO_WORDS(GC_size(op));
+       if (op != NULL) {
+         lg = BYTES_TO_WORDS(GC_size(op));
+	 ((word *)op)[GRANULES_TO_WORDS(lg) - 1] = d;
+       }
    }
-   if (op != NULL)
-       ((word *)op)[lw - 1] = d;
-   return((GC_PTR) op);
+   return((void *) op);
 }
 
-#if defined(__STDC__) || defined(__cplusplus)
-  void * GC_calloc_explicitly_typed(size_t n,
-  				    size_t lb,
-  				    GC_descr d)
-#else
-  char * GC_calloc_explicitly_typed(n, lb, d)
-  size_t n;
-  size_t lb;
-  GC_descr d;
-#endif
+GC_API void * GC_CALL GC_calloc_explicitly_typed(size_t n, size_t lb,
+						GC_descr d)
 {
-register ptr_t op;
-register ptr_t * opp;
-register word lw;
+ptr_t op;
+ptr_t * opp;
+size_t lg;
 GC_descr simple_descr;
 complex_descriptor *complex_descr;
 register int descr_type;
@@ -774,60 +682,51 @@ DCL_LOCK_STATE;
     	    break;
     }
     if( SMALL_OBJ(lb) ) {
-#       ifdef MERGE_SIZES
-	  lw = GC_size_map[lb];
-#	else
-	  lw = ALIGNED_WORDS(lb);
-#       endif
-	opp = &(GC_arobjfreelist[lw]);
-	FASTLOCK();
-        if( !FASTLOCK_SUCCEEDED() || (op = *opp) == 0 ) {
-            FASTUNLOCK();
+	lg = GC_size_map[lb];
+	opp = &(GC_arobjfreelist[lg]);
+	LOCK();
+        if( (op = *opp) == 0 ) {
+            UNLOCK();
             op = (ptr_t)GENERAL_MALLOC((word)lb, GC_array_kind);
 	    if (0 == op) return(0);
-#	    ifdef MERGE_SIZES
-		lw = GC_size_map[lb];	/* May have been uninitialized.	*/            
-#	    endif
+	    lg = GC_size_map[lb];	/* May have been uninitialized.	*/            
         } else {
             *opp = obj_link(op);
 	    obj_link(op) = 0;
-            GC_words_allocd += lw;
-            FASTUNLOCK();
+            GC_bytes_allocd += GRANULES_TO_BYTES(lg);
+            UNLOCK();
         }
    } else {
        op = (ptr_t)GENERAL_MALLOC((word)lb, GC_array_kind);
        if (0 == op) return(0);
-       lw = BYTES_TO_WORDS(GC_size(op));
+       lg = BYTES_TO_GRANULES(GC_size(op));
    }
    if (descr_type == LEAF) {
        /* Set up the descriptor inside the object itself. */
-       VOLATILE struct LeafDescriptor * lp =
+       volatile struct LeafDescriptor * lp =
            (struct LeafDescriptor *)
                ((word *)op
-                + lw - (BYTES_TO_WORDS(sizeof(struct LeafDescriptor)) + 1));
+                + GRANULES_TO_WORDS(lg)
+		- (BYTES_TO_WORDS(sizeof(struct LeafDescriptor)) + 1));
                 
        lp -> ld_tag = LEAF_TAG;
        lp -> ld_size = leaf.ld_size;
        lp -> ld_nelements = leaf.ld_nelements;
        lp -> ld_descriptor = leaf.ld_descriptor;
-       ((VOLATILE word *)op)[lw - 1] = (word)lp;
+       ((volatile word *)op)[GRANULES_TO_WORDS(lg) - 1] = (word)lp;
    } else {
-       extern unsigned GC_finalization_failures;
-       unsigned ff = GC_finalization_failures;
+       size_t lw = GRANULES_TO_WORDS(lg);
        
        ((word *)op)[lw - 1] = (word)complex_descr;
        /* Make sure the descriptor is cleared once there is any danger	*/
        /* it may have been collected.					*/
-       (void)
-         GC_general_register_disappearing_link((GC_PTR *)
-         					  ((word *)op+lw-1),
-       					          (GC_PTR) op);
-       if (ff != GC_finalization_failures) {
+       if (GC_general_register_disappearing_link((void * *)((word *)op+lw-1),
+						 op) == 2) {
 	   /* Couldn't register it due to lack of memory.  Punt.	*/
 	   /* This will probably fail too, but gives the recovery code  */
 	   /* a chance.							*/
 	   return(GC_malloc(n*lb));
        }			          
    }
-   return((GC_PTR) op);
+   return((void *) op);
 }
