@@ -2,32 +2,33 @@
  *
  * e n v . c			-- Environment management
  *
- * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ * Copyright © 1993-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
  *
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date: 23-Oct-1993 21:37
- * Last file update: 24-Apr-2002 11:41 (eg)
+ * Last file update: 27-May-2011 22:15 (eg)
  */
 
 #include "stklos.h"
 #include "hash.h"
-
+#include "vm.h"
+#include "thread-common.h"
 
 static void error_bad_module_name(SCM obj)
 {
@@ -61,16 +62,16 @@ static void error_bad_symbol(SCM symbol)
 
 
 /*===========================================================================*\
- * 
- * 				M O D U L E S 
- * 
+ *
+ * 				M O D U L E S
+ *
 \*===========================================================================*/
 struct module_obj {
   stk_header header;
   SCM name;			/* module name */
   SCM exported_symbols;		/* symbols declared as exported */
   SCM imports;			/* imported modules */
-  struct hash_table_obj hash;	/* The associated hash table */	
+  struct hash_table_obj hash;	/* The associated hash table */
 };
 
 
@@ -79,17 +80,16 @@ struct module_obj {
 #define MODULE_IMPORTS(m)	(((struct module_obj *) (m))->imports)
 #define MODULE_HASH_TABLE(m)	(((struct module_obj *) (m))->hash)
 
-#define VISIBLE_P(symb, mod)	(STk_memq((symb), MODULE_EXPORTS(mod))!=STk_false)
+#define VISIBLE_P(symb, mod)	(((mod) == STk_STklos_module) ||	\
+				 (STk_memq((symb), MODULE_EXPORTS(mod))!=STk_false))
 
 
-SCM STk_current_module;		/* The current module */
-SCM stklos_module;		/* The module whose name is STklos */ 
-static SCM all_modules;		/* List of all knowm modules */		
-
+SCM STk_STklos_module;		/* The module whose name is STklos */
+static SCM all_modules;		/* List of all knowm modules */
 
 static void print_module(SCM module, SCM port, int mode)
 {
-  STk_puts("#[module ", port);
+  STk_nputs(port, "#[module ", 9);
   STk_print(MODULE_NAME(module), port, mode);
   STk_putc(']', port);
 }
@@ -98,14 +98,19 @@ static void print_module(SCM module, SCM port, int mode)
 static SCM STk_makemodule(SCM name)
 {
   register SCM z;
-  
+
   NEWCELL(z, module);
   MODULE_NAME(z)	= name;
   MODULE_EXPORTS(z)	= STk_nil;
-  MODULE_IMPORTS(z)	= (name == STk_void)? STk_nil : LIST1(stklos_module);
-  /* Initialize the associated hash table & stor the module in the global list*/
+  MODULE_IMPORTS(z)	= (name == STk_void)? STk_nil : LIST1(STk_STklos_module);
+  /* Initialize the associated hash table & store the module in the global list*/
   STk_hashtable_init(&MODULE_HASH_TABLE(z), HASH_VAR_FLAG);
-  all_modules = STk_cons(z, all_modules);
+  {
+    MUT_DECL(lck);
+    MUT_LOCK(lck);
+    all_modules = STk_cons(z, all_modules);
+    MUT_UNLOCK(lck);
+  }
   return z;
 }
 
@@ -113,7 +118,10 @@ static SCM STk_makemodule(SCM name)
 static SCM find_module(SCM name, int create)
 {
   SCM tmp;
-  
+
+  if (name == STk_intern("STklos") || name == STk_intern("stklos"))
+    return STk_STklos_module;
+
   for (tmp = all_modules; !NULLP(tmp); tmp = CDR(tmp)) {
     if (MODULE_NAME(CAR(tmp)) == name)
       return CAR(tmp);
@@ -123,7 +131,7 @@ static SCM find_module(SCM name, int create)
 }
 
 /*===========================================================================*\
- * 
+ *
  * Module primitives
  *
 \*===========================================================================*/
@@ -132,54 +140,53 @@ static SCM find_module(SCM name, int create)
 
 DEFINE_PRIMITIVE("%create-module", create_module, subr1, (SCM name))
 {
-  ENTER_PRIMITIVE(create_module);
   if (!SYMBOLP(name)) error_bad_module_name(name);
   return find_module(name, TRUE);
 }
 
 DEFINE_PRIMITIVE("%select-module", select_module, subr1, (SCM module))
 {
-  ENTER_PRIMITIVE(select_module);
+  vm_thread_t *vm = STk_get_current_vm();
+
   if (!MODULEP(module)) error_bad_module(module);
-  
-  STk_current_module = module;
+  vm->current_module= module;
   return STk_void;
 }
 
-DEFINE_PRIMITIVE("%module-imports-set!", module_imports_set, subr2, 
+DEFINE_PRIMITIVE("%module-imports-set!", module_imports_set, subr2,
 		 (SCM importer,SCM imported))
 {
-  ENTER_PRIMITIVE(module_imports_set);
-
   if (!MODULEP(importer)) error_bad_module(importer);
-  if (!CONSP(imported))   error_bad_list(imported);
-  
-  MODULE_IMPORTS(importer) = imported;
+
+  if (CONSP(imported))
+    MODULE_IMPORTS(importer) = STk_dappend2(imported, LIST1(STk_STklos_module));
+  else if (NULLP(imported))
+    MODULE_IMPORTS(importer) = STk_nil;
+  else error_bad_list(imported);
+
   return STk_void;
 }
 
-DEFINE_PRIMITIVE("%module-exports-set!", module_exports_set, subr2, 
+DEFINE_PRIMITIVE("%module-exports-set!", module_exports_set, subr2,
 		 (SCM exporter,SCM exported))
 {
-  ENTER_PRIMITIVE(module_imports_set);
-
   if (!MODULEP(exporter)) error_bad_module(exporter);
   if (!CONSP(exported))   error_bad_list(exported);
-  
+
   MODULE_EXPORTS(exporter) = exported;
   return STk_void;
 }
 
 
 /*
-<doc ext module?
+<doc EXT module?
  * (module? object)
  *
- * Returns |t| if |object| is a module and |f| otherwise.
+ * Returns |#t| if |object| is a module and |#f| otherwise.
  * @lisp
- * (module? (find-module 'STklos))  => #t
- * (module? 'STklos)                => #f
- * (module? 1 'no)                  => no
+ * (module? (find-module 'ST\klos))  => #t
+ * (module? 'ST\klos)                => #f
+ * (module? 123 'no)                => no
  * @end lisp
 doc>
 */
@@ -189,21 +196,20 @@ DEFINE_PRIMITIVE("module?", modulep, subr1, (SCM obj))
 }
 
 /*
-<doc ext find-module
+<doc EXT find-module
  * (find-module name)
  * (find-module name default)
  *
- * @stklos{} modules are first class objects and |find-module| returns the
+ * STklos modules are first class objects and |find-module| returns the
  * module associated to |name| if it exists. If there is no module
  * associated to |name|, an error is signaled if no |default| is
- * provided, otherwise |find-module| returns |default|. 
+ * provided, otherwise |find-module| returns |default|.
 doc>
 */
 DEFINE_PRIMITIVE("find-module", scheme_find_module, subr12, (SCM name, SCM def))
 {
   SCM module;
 
-  ENTER_PRIMITIVE(scheme_find_module);
   if (!SYMBOLP(name)) error_bad_module_name(name);
 
   module = find_module(name, FALSE);
@@ -215,78 +221,80 @@ DEFINE_PRIMITIVE("find-module", scheme_find_module, subr12, (SCM name, SCM def))
 }
 
 /*
-<doc ext current-module
+<doc EXT current-module
  * (current-module)
  *
  * Returns the current module.
  * @lisp
  * (define-module M
- *   (display 
+ *   (display
  *       (cons (eq? (current-module) (find-module 'M))
  *             (eq? (current-module) (find-module 'STklos)))))
  *    @print{} (#t . #f)
  * @end lisp
 doc>
  */
-DEFINE_PRIMITIVE("current-module", scheme_current_module, subr0, (void))
+DEFINE_PRIMITIVE("current-module", current_module, subr0, (void))
 {
-  return STk_current_module;
+  if (STk_primordial_thread) {
+    vm_thread_t *vm = STk_get_current_vm();
+    return vm->current_module;
+  } else {
+    return STk_STklos_module;
+  }
 }
 
 
 /*
-<doc ext module-name
+<doc EXT module-name
  * (module-name module)
  *
- * Returns the the name (a symbol) associated to a |module|.
+ * Returns the name (a symbol) associated to a |module|.
 doc>
  */
 DEFINE_PRIMITIVE("module-name", module_name, subr1, (SCM module))
 {
-  ENTER_PRIMITIVE(module_name);
   if (!MODULEP(module)) error_bad_module(module);
   return MODULE_NAME(module);
 }
 
 
 /*
-<doc ext module-imports
+<doc EXT module-imports
  * (module-imports module)
  *
- * Returns the list of modules that |module| imports.
+ * Returns the list of modules that |module| (fully) imports.
 doc>
  */
 DEFINE_PRIMITIVE("module-imports", module_imports, subr1, (SCM module))
 {
-  ENTER_PRIMITIVE(module_imports);
   if (!MODULEP(module)) error_bad_module(module);
   return MODULE_IMPORTS(module);
 }
 
 
 /*
-<doc ext module-exports
+<doc EXT module-exports
  * (module-exports module)
  *
  * Returns the list of symbols exported by |module|. Note that this function
- * returns the list of symbols given in the module |export| clause and that 
+ * returns the list of symbols given in the module |export| clause and that
  * some of these symbols can be not yet defined.
 doc>
  */
 DEFINE_PRIMITIVE("module-exports", module_exports, subr1, (SCM module))
 {
-  ENTER_PRIMITIVE(module_exports);
   if (!MODULEP(module)) error_bad_module(module);
 
   /* STklos module is special: everything is exported ==> module-symbols */
-  return (module == stklos_module) ?
+  return (module == STk_STklos_module) ?
     		STk_hash_keys(&MODULE_HASH_TABLE(module)) :
     		MODULE_EXPORTS(module);
 }
 
 
 /*
-<doc ext module-symbols
+<doc EXT module-symbols
  * (module-symbols module)
  *
  * Returns the list of symbols already defined in |module|.
@@ -294,13 +302,12 @@ doc>
  */
 DEFINE_PRIMITIVE("module-symbols", module_symbols, subr1,  (SCM module))
 {
-  ENTER_PRIMITIVE(module_name);
   if (!MODULEP(module)) error_bad_module(module);
   return STk_hash_keys(&MODULE_HASH_TABLE(module));
 }
 
 /*
-<doc ext all-modules
+<doc EXT all-modules
  * (all-modules)
  *
  * Returns the list of all the living modules.
@@ -313,28 +320,27 @@ DEFINE_PRIMITIVE("all-modules", all_modules, subr0, (void))
 
 
 /*
-<doc EXT symbol-value 
+<doc EXT symbol-value
  * (symbol-value symbol module)
  * (symbol-value symbol module default)
  *
  * Returns the value bound to |symbol| in |module|. If |symbol| is not bound,
- * an error is signaled if no |default| is provided, otherwise |symbol-value| 
+ * an error is signaled if no |default| is provided, otherwise |symbol-value|
  * returns |default|.
 doc>
  */
-DEFINE_PRIMITIVE("symbol-value", symbol_value, subr23, 
+DEFINE_PRIMITIVE("symbol-value", symbol_value, subr23,
 		 (SCM symbol, SCM module, SCM default_value))
 {
   SCM res;
   int i;
 
-  ENTER_PRIMITIVE(symbol_value);
   if (!SYMBOLP(symbol)) error_bad_symbol(symbol);
   if (!MODULEP(module)) error_bad_module(module);
 
   res = STk_hash_get_variable(&MODULE_HASH_TABLE(module), symbol, &i);
   if (res) {
-    return CDR(res);
+    return BOX_VALUE(CDR(res));
   } else {
     if (!default_value) error_unbound_variable(symbol);
     return default_value;
@@ -342,11 +348,37 @@ DEFINE_PRIMITIVE("symbol-value", symbol_value, subr23,
 }
 
 
+DEFINE_PRIMITIVE("%redefine-module-exports", redefine_module_exports, subr12,
+		 (SCM from, SCM to))
+{
+  SCM lst, res;
+  int i;
+
+  if (!MODULEP(from)) error_bad_module(from);
+  if (!to)
+    to =  STk_current_module();
+  else
+    if (!MODULEP(to)) error_bad_module(to);
+
+  /* Compute the list of exported symbols */
+  if (from == STk_STklos_module)
+    lst = STk_hash_keys(&MODULE_HASH_TABLE(STk_STklos_module)); /* everybody */
+  else
+    lst = MODULE_EXPORTS(from);				/* explicitly exported */
+
+  for (     ; !NULLP(lst); lst = CDR(lst)) {
+    res = STk_hash_get_variable(&MODULE_HASH_TABLE(from), CAR(lst), &i);
+    if (res)
+      /* symbol (car lst) is bound in module from. redefine it in module to */
+      STk_define_variable(CAR(lst), BOX_VALUE(CDR(res)), to);
+  }
+  return STk_void;
+}
 
 /*===========================================================================*\
- * 
+ *
  * 		E n v i r o n m e n t   M a n a g e m e n t
- * 
+ *
 \*===========================================================================*/
 
 #define FRAME_ALLOC_BYTES(len) (sizeof(struct frame_obj) + ((len)-1)*sizeof(SCM))
@@ -377,10 +409,9 @@ void STk_define_variable(SCM symbol, SCM value, SCM module)
 }
 
 
-DEFINE_PRIMITIVE("%symbol-define", symbol_define, subr3, 
+DEFINE_PRIMITIVE("%symbol-define", symbol_define, subr3,
 		 (SCM symbol, SCM value, SCM module))
 {
-  ENTER_PRIMITIVE(symbol_define);
   if (!SYMBOLP(symbol)) error_bad_symbol(symbol);
   if (!MODULEP(module)) error_bad_module(module);
 
@@ -388,6 +419,26 @@ DEFINE_PRIMITIVE("%symbol-define", symbol_define, subr3,
   return value;
 }
 
+DEFINE_PRIMITIVE("%symbol-alias", symbol_alias, subr23,
+		 (SCM new, SCM old, SCM module))
+{
+  SCM res, mod = STk_current_module();;
+  int i;
+
+  if (!SYMBOLP(new)) error_bad_symbol(new);
+  if (!SYMBOLP(old)) error_bad_symbol(old);
+  if (!module)
+    module = mod;
+  else
+    if (!MODULEP(module)) error_bad_module(module);
+
+  res = STk_hash_get_variable(&MODULE_HASH_TABLE(module), old, &i);
+  if (!res)
+    error_unbound_variable(old);
+
+  STk_hash_set_alias(&MODULE_HASH_TABLE(mod), new, CDR(res));
+  return STk_void;
+}
 
 
 /*===========================================================================*\
@@ -401,38 +452,26 @@ SCM STk_lookup(SCM symbol, SCM env, SCM *ref, int err_if_unbound)
 
   while (FRAMEP(env)) env = FRAME_NEXT(env);
 
-
   res = STk_hash_get_variable(&MODULE_HASH_TABLE(env), symbol, &i);
   if (res) {
     *ref = res;
-    return CDR(res);
+    return BOX_VALUE(CDR(res));
   }
   else {
-    /* symbol was not found in the module. Try to find it in the 
-     * exported symbols of its imported modules
+    /* symbol was not found in the given env module. Try to find it in
+     * the  exported symbols of its imported modules.
      */
     for (l = MODULE_IMPORTS(env)  ; !NULLP(l); l = CDR(l)) {
       module = CAR(l);
       res    = STk_hash_get_variable(&MODULE_HASH_TABLE(module), symbol, &i);
       if (res && VISIBLE_P(symbol, module)) {
 	*ref = res;
-	return CDR(res);
+	return BOX_VALUE(CDR(res));
       }
     }
 
-    /* Not found in the imported modules. Try in the stklos module (if we 
-     * didn't had searched it yet
-     */
-    if (env != stklos_module) {
-      res = STk_hash_get_variable(&MODULE_HASH_TABLE(stklos_module), symbol, &i);
-      if (res) {
-	*ref = res;
-	return CDR(res);
-      }
-    }
-    
     /* It definitively does not exists  :-< */
-    if (err_if_unbound) 
+    if (err_if_unbound)
       error_unbound_variable(symbol);
     return STk_void;
   }
@@ -440,13 +479,13 @@ SCM STk_lookup(SCM symbol, SCM env, SCM *ref, int err_if_unbound)
 
 
 /*===========================================================================*\
- * 
+ *
  * 		E n v i r o n m e n t   P r i m i t i v e s
- * 
+ *
 \*===========================================================================*/
 
 /* The stucture which describes the modules type */
-struct extended_type_descr xtype_module = {
+static struct extended_type_descr xtype_module = {
   "module",			/* name */
   print_module			/* print function */
 };
@@ -455,7 +494,7 @@ struct extended_type_descr xtype_module = {
 
 
 /* The stucture which describes the frame type */
-struct extended_type_descr xtype_frame = { 
+static struct extended_type_descr xtype_frame = {
   "frame",			/* name */
   NULL				/* print function */
 };
@@ -466,9 +505,8 @@ int STk_init_env(void)
   all_modules = STk_nil;
 
   /* Create the stklos module */
-  stklos_module      = STk_makemodule(STk_void); /* will be changed later */
-  STk_current_module = stklos_module;
-  
+  STk_STklos_module  = STk_makemodule(STk_void); /* will be changed later */
+
   /* Declare the extended types module_obj and frame_obj */
   DEFINE_XTYPE(module, &xtype_module);
   DEFINE_XTYPE(frame,  &xtype_frame);
@@ -478,28 +516,33 @@ int STk_init_env(void)
 int STk_late_init_env(void)
 {
   /* Now that symbols are initialized change the STklos module name */
-  MODULE_NAME(stklos_module) = STk_intern("stklos");
-  
+  MODULE_NAME(STk_STklos_module) = STk_intern("stklos");
+  MODULE_IMPORTS(STk_STklos_module) = LIST1(STk_STklos_module);
+
+  /* Create the SCHEME module */
+  STk_makemodule(STk_intern("SCHEME"));
+
   /* ==== Undocumented primitives ==== */
   ADD_PRIMITIVE(create_module);
   ADD_PRIMITIVE(select_module);
   ADD_PRIMITIVE(module_imports_set);
   ADD_PRIMITIVE(module_exports_set);
+  ADD_PRIMITIVE(redefine_module_exports);
 
   /* ==== User primitives ==== */
   ADD_PRIMITIVE(modulep);
   ADD_PRIMITIVE(scheme_find_module);
-  ADD_PRIMITIVE(scheme_current_module);
+  ADD_PRIMITIVE(current_module);
   ADD_PRIMITIVE(module_name);
   ADD_PRIMITIVE(module_imports);
   ADD_PRIMITIVE(module_exports);
   ADD_PRIMITIVE(module_symbols);
   ADD_PRIMITIVE(all_modules);
 
-  ADD_PRIMITIVE(symbol_define);
   ADD_PRIMITIVE(symbol_value);
+  ADD_PRIMITIVE(symbol_define);
+  ADD_PRIMITIVE(symbol_alias);
 
 
   return TRUE;
 }
-
