@@ -1,11 +1,11 @@
 /*
  * gnu-getopt.c		-- Adaptation of the GNU getopt function for STklos
- * 
- * Copyright  2001-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ *
+ * Copyright © 2001-2006 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date:  1-Nov-2001 23:40 (eg)
- * Last file update: 10-Feb-2003 10:42 (eg)
+ * Last file update: 12-Apr-2006 12:43 (eg)
  *
  * The following code is constituted of the GNU getopt, getopt_long
  * and getopt_long_only implementation. Code is slightly modified (all
@@ -60,29 +60,31 @@
 
 
 /*===========================================================================*\
- * 
+ *
  * 				STklos adaptation
  *
 \*===========================================================================*/
 #include <stklos.h>
+#define GETOPT_BUF_SIZE 250
 
 static int we_are_in_scheme = 0;
 
-/* 
- * Use a traditional K&R function declaration to avoid the complexity
- * of such a function. Parameters are not controled here but there is
- * no problem.
- */
-static void getopt_fprintf(f, fmt, p1, p2, p3)
-     FILE *f; char *fmt;
+static void getopt_fprintf(FILE* f, const char *format, ...)
 {
+  char buffer[GETOPT_BUF_SIZE];
+  va_list ap;
+
+  va_start(ap, format);
+  vsnprintf(buffer, GETOPT_BUF_SIZE, format, ap);
+
   if (we_are_in_scheme) {
-    STk_fprintf(STk_curr_oport, fmt, p1, p2, p3);
+    STk_puts(buffer, STk_current_output_port());
   } else {
-    fprintf(stderr, fmt, p1, p2, p3);
+    fprintf(f, "%s", buffer);
   }
 }
 
+
 /* Redefine fprintf to our function for later uses in this file */
 #define fprintf getopt_fprintf
 
@@ -1037,8 +1039,8 @@ getopt (argc, argv, optstring)
 
 /*===========================================================================*/
 
-/* 
- * STklos: The following code is normally in a file name getopt1.c. 
+/*
+ * STklos: The following code is normally in a file name getopt1.c.
  * It has been placed here for commodity
  */
 
