@@ -1,9 +1,9 @@
 /*
  *
- * h a s h  . c			-- Hash Tables 
+ * h a s h  . c			-- Hash Tables (mostly SRFI-69)
+ *
+ * Copyright © 1994-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  *
- * Copyright  1994-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
  +=============================================================================
  ! This code is a rewriting of the file tclHash.c of the Tcl
  ! distribution.  Current writing is more modular, independent of Tcl
@@ -11,32 +11,32 @@
  ! Tcl implementation have been droped here. Furthermore, current
  ! implementation of the API also takes into account the fact that we
  ! have a GC approach. In fact, the code is very different, the only thing
- ! which is kept is the Tcl strategy for sizing and hashing 
- ! The original code from which this file is derived was copyrighted 
+ ! which is kept is the Tcl strategy for sizing and hashing
+ ! The original code from which this file is derived was copyrighted
  ! as follow:
  !
  ! Copyright (c) 1991-1993 The Regents of the University of California.
  ! Copyright (c) 1994 Sun Microsystems, Inc.
  +=============================================================================
- * 
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 17-Jan-1994 17:49
- * Last file update: 24-Apr-2002 12:18 (eg)
+ * Last file update: 11-Aug-2011 23:38 (eg)
  */
 
 #include "stklos.h"
@@ -57,8 +57,9 @@
 #define HASH_WORD(h1, h2)  ((((h1) << 4) + (h1)) ^ (h2))  /* Good repartition ? */
 
 
+
 /*===========================================================================*\
- * 
+ *
  *  		       H a s h i n g - f u n c t i o n s
  *
 \*===========================================================================*/
@@ -82,7 +83,7 @@ static unsigned long hash_string(register char *string)
    *    the high-order bits to fill out the hash value.  This seems
    *    works well both for decimal and non-decimal strings.
    */
-  
+
   for (c = *string; c ; c = *string++) {
     result += (result<<3) + c;
   }
@@ -90,12 +91,12 @@ static unsigned long hash_string(register char *string)
 }
 
 
-static unsigned long hash_scheme_string(SCM str) 
+static unsigned long hash_scheme_string(SCM str)
 {					     /* The same one for Scheme strings */
   char *string;
   unsigned long result = 0;
   int i, l;
-  
+
   if (!STRINGP(str)) STk_error("bad string key ~S", str);
 
   string = STRING_CHARS(str);
@@ -107,15 +108,15 @@ static unsigned long hash_scheme_string(SCM str)
 
 
 /*
- * sxhash calculates a "universal" hash value  la CL sxhash  function
+ * sxhash calculates a "universal" hash value à la CL sxhash  function
  *
  */
-unsigned long sxhash(SCM obj)
+static unsigned long sxhash(SCM obj)
 {
   unsigned long h;
   SCM tmp;
   int i;
-  
+
   if (!BOXED_OBJP(obj)) return (AS_LONG(obj) >> 2);
 
   switch (BOXED_TYPE(obj)) {
@@ -124,7 +125,7 @@ unsigned long sxhash(SCM obj)
 			  h = HASH_WORD(h, sxhash(CAR(tmp)));
 			h = HASH_WORD(h, sxhash(tmp));
 			return h;
-    case tc_bignum:	return (unsigned long) STk_integer_value(obj);
+    case tc_bignum:	return sxhash(STk_number2string(obj, MAKE_INT(16)));
     case tc_real:	return (unsigned long) REAL_VAL(obj);
     case tc_rational:   return HASH_WORD(sxhash(RATIONAL_NUM(obj)),
 					 sxhash(RATIONAL_DEN(obj)));
@@ -134,21 +135,21 @@ unsigned long sxhash(SCM obj)
     case tc_keyword:	return hash_string(KEYWORD_PNAME(obj));
     case tc_string:	return hash_scheme_string(obj);
     case tc_vector:	h = 0;
-			for (i=VECTOR_SIZE(obj)-1; i >= 0; i--) 
+			for (i=VECTOR_SIZE(obj)-1; i >= 0; i--)
 			  h = HASH_WORD(h, sxhash(VECTOR_DATA(obj)[i]));
 			return h;
     default:	        /* A complex type (STklos object, user defined type,
 			 * hashtable...). In this case we return the type of the
 			 * object. This is very  inneficient but it should be
-			 * rare to use a  structured object as a key. 
+			 * rare to use a  structured object as a key.
 			 */
       			 return (unsigned long) BOXED_TYPE(obj);
   }
 }
 
 /*===========================================================================*\
- * 
- * 				H a s h    S t a t s 
+ *
+ * 				H a s h    S t a t s
  *
 \*===========================================================================*/
 #define _MAX_COUNT 5
@@ -156,38 +157,38 @@ unsigned long sxhash(SCM obj)
 static void hash_stats(struct hash_table_obj *h, SCM port)
 {
   int i, j, more, len, count[_MAX_COUNT] = {0};
-  
+
   STk_fprintf(port, "Hash table statistics\n");
-  more = 0; 
+  more = 0;
   for (i = 0; i < HASH_NBUCKETS(h); i++) {
     len = STk_int_length(HASH_BUCKETS(h)[i]);
-    
+
     STk_fprintf(port, "%d: ", i);
     for (j = 0; j < len; j++) STk_putc('#', port);
     STk_putc('\n', port);
-    
+
     if (len < _MAX_COUNT)
       count[len] += 1;
     else
       more += 1;
   }
 
-  STk_fprintf(port, "  %d entries in table, %d buckets (ratio %g)\n", 
-	      HASH_NENTRIES(h), HASH_NBUCKETS(h), 
+  STk_fprintf(port, "  %d entries in table, %d buckets (ratio %g)\n",
+	      HASH_NENTRIES(h), HASH_NBUCKETS(h),
 	      (double) HASH_NENTRIES(h)/ HASH_NBUCKETS(h));
   STk_fprintf(port, "Repartition\n");
 
   for (i = 0; i < _MAX_COUNT; i++) {
-    if (count[i]) 
+    if (count[i])
       STk_fprintf(port, "  %d buckets with %d entries\n", count[i], i);
   }
-  if (more) 
+  if (more)
     STk_fprintf(port, "  %d buckets with more than %d entries\n", more, _MAX_COUNT);
 }
 
 
 /*===========================================================================*\
- * 
+ *
  * 				E n l a r g e   t a b l e
  *
 \*===========================================================================*/
@@ -210,7 +211,7 @@ static void enlarge_table(register struct hash_table_obj *h)
   HASH_MASK(h)	    = (HASH_MASK(h) << 2) + 3;
 
   /* Allocate and initialize the new bucket array. */
-  HASH_BUCKETS(h)   = (SCM *) 
+  HASH_BUCKETS(h)   = (SCM *)
     		STk_must_malloc((size_t) (HASH_NBUCKETS(h) * sizeof(SCM)));
   for (i = 0; i < HASH_NBUCKETS(h); i++) {
     HASH_BUCKETS(h)[i] = STk_nil;
@@ -220,24 +221,24 @@ static void enlarge_table(register struct hash_table_obj *h)
   for (i = 0; i < old_size; i++) {
     for (tmp = old_buckets[i]; !NULLP(tmp); tmp = CDR(tmp)) {
       switch (BOXED_INFO(h)) {
-	case HASH_OBARRAY_FLAG: 
+	case HASH_OBARRAY_FLAG:
 	  index = hash_string(SYMBOL_PNAME(CAR(tmp))) & HASH_MASK(h);
 	  break;
-	case HASH_VAR_FLAG: 
+	case HASH_VAR_FLAG:
 	  index = hash_string(SYMBOL_PNAME(CAR(CAR(tmp)))) & HASH_MASK(h);
 	  break;
 	case HASH_SCM_FLAG: {
 	  SCM key = CAR(CAR(tmp));
 	  switch (HASH_TYPE(h)) {
-	    case hash_eqp: 
+	    case hash_eqp:
 	      index = RANDOM_INDEX(h, key);
 	      break;
 	    case hash_stringp:
 	      index = hash_scheme_string(key) & HASH_MASK(h);
 	      break;
 	    case hash_general:
-	      index = RANDOM_INDEX(h, STk_integer_value(STk_C_apply(HASH_HASH(h), 
-								    1, 
+	      index = RANDOM_INDEX(h, STk_integer_value(STk_C_apply(HASH_HASH(h),
+								    1,
 								    key)));
 	      break;
 	    default: ;
@@ -256,7 +257,7 @@ static void enlarge_table(register struct hash_table_obj *h)
 
 
 /*===========================================================================*\
- * 
+ *
  *  		       H a s h - t a b l e   c r e a t i o n
  *
 \*===========================================================================*/
@@ -281,11 +282,11 @@ void STk_hashtable_init(struct hash_table_obj *h, int flag)
 
 
 /*===========================================================================*\
- * 
+ *
  *  	          O b a r r a y   h a s h t a b l e   f u n t i o n s
  *
  *
- * Keys are symbols or keywords. The value associated to a bucket is a list 
+ * Keys are symbols or keywords. The value associated to a bucket is a list
  * of the  symbols (or keywords) with the same hash value
  *
 \*===========================================================================*/
@@ -307,7 +308,7 @@ SCM STk_hash_intern_symbol(struct hash_table_obj *h, char *s, SCM (*create) (cha
 {
   SCM z;
   int index;
-  
+
   z = hash_get_symbol(h, s, &index);
 
   if (!z) {
@@ -323,12 +324,12 @@ SCM STk_hash_intern_symbol(struct hash_table_obj *h, char *s, SCM (*create) (cha
 
 
 /*===========================================================================*\
- * 
+ *
  *  	       v a r i a b l e s    h a s h t a b l e   f u n t i o n s
  *
  *
- * Here variable are the variables defined in a module. Keys are symbols. 
- * The value associated to a bucket is an A-list (symbol . value) of all the 
+ * Here variable are the variables defined in a module. Keys are symbols.
+ * The value associated to a bucket is an A-list (symbol . value) of all the
  * symbols with the same hash value
  *
 \*===========================================================================*/
@@ -354,6 +355,31 @@ void STk_hash_set_variable(struct hash_table_obj *h, SCM v, SCM value)
 
   z = STk_hash_get_variable(h, v, &index);
 
+  if (z) {
+    /* Variable already exists. Change its value*/
+    BOX_VALUE(CDR(z)) = value;
+  } else {
+    SCM z;
+
+    /* Create a new box for this value */
+    z = STk_make_box(value);
+
+    /* Enter the new variable in table */
+    HASH_BUCKETS(h)[index] = STk_cons(STk_cons(v, z),
+				      HASH_BUCKETS(h)[index]);
+    HASH_NENTRIES(h) += 1;
+    /* If the table has exceeded a decent size, rebuild it */
+    if (HASH_NENTRIES(h) >= HASH_NEWSIZE(h)) enlarge_table(h);
+  }
+}
+
+void STk_hash_set_alias(struct hash_table_obj *h, SCM v, SCM value)
+{
+  SCM z;
+  int index;
+
+  z = STk_hash_get_variable(h, v, &index);
+
   if (z) {
     /* Variable already exists. Change its value*/
     CDR(z) = value;
@@ -368,8 +394,10 @@ void STk_hash_set_variable(struct hash_table_obj *h, SCM v, SCM value)
 }
 
 
+
+
 /*===========================================================================*\
- * 
+ *
  * 			Utilities on hash tables
  *
 \*===========================================================================*/
@@ -391,11 +419,11 @@ SCM STk_hash_keys(struct hash_table_obj *h)
 
 
 /*===========================================================================*\
- * 
+ *
  * 		H a s h - t a b l e s   i n   t h e
  *
  * 		      S c h e m e   W o r l d
- * 
+ *
 \*===========================================================================*/
 static void error_bad_hash_table(SCM obj)
 {
@@ -408,6 +436,10 @@ static void error_bad_procedure(SCM obj)
   STk_error("bad procedure ~S", obj);
 }
 
+static void error_not_present(SCM key, SCM ht)
+{
+  STk_error("entry not defined for key ~S in table ~S", key, ht);
+}
 
 
 DEFINE_PRIMITIVE("%make-hash-table", make_hash, subr2, (SCM compar, SCM hashfct))
@@ -415,16 +447,14 @@ DEFINE_PRIMITIVE("%make-hash-table", make_hash, subr2, (SCM compar, SCM hashfct)
   SCM z;
   hash_type type = hash_general;
 
-  ENTER_PRIMITIVE(make_hash);
-
-  if (STk_procedurep(compar) == STk_false) 
+  if (STk_procedurep(compar) == STk_false)
     STk_error("bad comparison function ~S", compar);
-  if (STk_procedurep(hashfct) == STk_false) 
+  if (STk_procedurep(hashfct) == STk_false)
     STk_error("bad hash function ~S", hashfct);
 
   if (BOXED_TYPE(compar) == tc_subr2) {
     /* See if comparison function is 'eq?' or 'string?'.
-     * If so, we implement the hash table in the most efficient way. 
+     * If so, we implement the hash table in the most efficient way.
      */
     if (PRIMITIVE_FUNC(compar) == STk_eq)    type = hash_eqp;
     if (PRIMITIVE_FUNC(compar) == STk_streq) type = hash_stringp;
@@ -443,7 +473,7 @@ DEFINE_PRIMITIVE("%make-hash-table", make_hash, subr2, (SCM compar, SCM hashfct)
 
 
 /*
-<doc ext hash-table?
+<doc EXT hash-table?
  * (hash-table? obj)
  *
  * Returns |#t| if |obj| is a hash table, returns |#f| otherwise.
@@ -455,54 +485,96 @@ DEFINE_PRIMITIVE("hash-table?", hashtablep, subr1, (SCM obj))
 }
 
 
-
 /*
-<doc ext hash-table-put!
- * (hash-table-put! hash key value)
+<doc EXT hash-table-size
+ * (hash-table-size hash)
  *
- * Enters an association between |key| and |value| in the|hash| table. 
- * The value returned by |hash-table-put!| is @emph{void}.
+ * Returns the number of entries in the |hash|.
 doc>
 */
-DEFINE_PRIMITIVE("hash-table-put!", hash_put, subr3, (SCM ht, SCM key, SCM val))
+DEFINE_PRIMITIVE("hash-table-size", hash_table_size, subr1, (SCM ht))
+{
+  if(!HASHP(ht)) error_bad_hash_table(ht);
+  return MAKE_INT(HASH_NENTRIES(ht));
+}
+
+/*
+<doc EXT hash-table-equivalence-function
+ * (hash-table-equivalence-function hash)
+ *
+ * Returns the equivalence predicate used for keys in |hash|.
+doc>
+*/
+DEFINE_PRIMITIVE("hash-table-equivalence-function", hash_table_eqv_func,
+		 subr1, (SCM ht))
+{
+  if(!HASHP(ht)) error_bad_hash_table(ht);
+  return HASH_COMPAR(ht);
+}
+
+
+/*
+<doc EXT hash-table-hash-function
+ * (hash-table-hash-function hash)
+ *
+ * Returns the hash function used for keys in |hash|.
+doc>
+*/
+DEFINE_PRIMITIVE("hash-table-hash-function", hash_table_hash_func,
+		 subr1, (SCM ht))
+{
+  if(!HASHP(ht)) error_bad_hash_table(ht);
+  return HASH_HASH(ht);
+}
+
+
+
+/*
+<doc EXT hash-table-set!
+ * (hash-table-set! hash key value)
+ *
+ * Enters an association between |key| and |value| in the|hash| table.
+ * The value returned by |hash-table-set!| is ,(emph "void").
+doc>
+*/
+DEFINE_PRIMITIVE("hash-table-set!", hash_set, subr3, (SCM ht, SCM key, SCM val))
 {
   int index = 0;
   SCM func, l = STk_nil;
-  
-  ENTER_PRIMITIVE(hash_put);
+
   if (!HASHP(ht)) error_bad_hash_table(ht);
 
   switch (HASH_TYPE(ht)) {
-    case hash_eqp: 
+    case hash_eqp:
       index = RANDOM_INDEX(ht, key);
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (CAR(CAR(l)) == key) break;
       break;
 
     case hash_stringp:
       index = hash_scheme_string(key) & HASH_MASK(ht);
 
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (STk_streq(CAR(CAR(l)), key) != STk_false) break;
       break;
     case hash_general:
       func  = HASH_COMPAR(ht);
-      index = RANDOM_INDEX(ht, 
+      index = RANDOM_INDEX(ht,
 			   STk_integer_value(STk_C_apply(HASH_HASH(ht), 1, key)));
 
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) break;
     default:  break;
   }
 
-  /* When we are here, index is on the good bucket list. if l is null, 
+  /* When we are here, index is on the good bucket list. if l is null,
    * the item was not in the given list otherwise CAR(l) = (key . old-value)
    */
   if (NULLP(l)) {
     HASH_BUCKETS(ht)[index] = STk_cons(STk_cons(key, val), HASH_BUCKETS(ht)[index]);
     HASH_NENTRIES(ht) += 1;
     /* If the table has exceeded a decent size, rebuild it */
-    if (HASH_NENTRIES(ht) >= HASH_NEWSIZE(ht)) 
+    if (HASH_NENTRIES(ht) >= HASH_NEWSIZE(ht))
       enlarge_table((struct hash_table_obj *) ht);
   }
   else
@@ -513,114 +585,157 @@ DEFINE_PRIMITIVE("hash-table-put!", hash_put, subr3, (SCM ht, SCM key, SCM val))
 
 
 /*
-<doc ext hash-table-get
- * (hash-table-get hash key)
- * (hash-table-get hash key default)
+<doc EXT hash-table-ref
+ * (hash-table-ref hash key)
+ * (hash-table-ref hash key thunk)
  *
- * Returns the value associated with |key| in the given |hash| table. If no 
- * value has been associated with |key| in |hash|, the specified |default| is 
- * returned if given; otherwise an eerror is raised.
- * @lisp 
+ * Returns the value associated with |key| in the given |hash| table. If no
+ * value has been associated with |key| in |hash|, the specified |thunk| is
+ * called and its value is returned; otherwise an error is raised.
+ * @lisp
  * (define h1 (make-hash-table))
- * (hash-table-put! h1 'foo (list 1 2 3))
- * (hash-table-get  h1 'foo)              =>  (1 2 3)
- * (hash-table-get  h1 'bar 'absent)      =>  absent
- * (hash-table-get  h1 'bar)              =>  @w{@emph{error}}
- * (hash-table-put! h1 '(a b c) 'present)
- * (hash-table-get  h1 '(a b c) 'absent)  => absent
- * 
+ * (hash-table-set! h1 'foo (list 1 2 3))
+ * (hash-table-ref  h1 'foo)                 =>  (1 2 3)
+ * (hash-table-ref  h1 'bar
+                       (lambda () 'absent))  =>  absent
+ * (hash-table-ref  h1 'bar)                 =>  error
+ * (hash-table-set! h1 '(a b c) 'present)
+ * (hash-table-ref  h1 '(a b c)
+                        (lambda () 'absent)) => absent
+ *
  * (define h2 (make-hash-table equal?))
- * (hash-table-put! h2 '(a b c) 'present)
- * (hash-table-get  h2 '(a b c))          => present
+ * (hash-table-set! h2 '(a b c) 'present)
+ * (hash-table-ref  h2 '(a b c))             => present
  * @end lisp
 doc>
 */
-DEFINE_PRIMITIVE("hash-table-get", hash_get, subr23, (SCM ht, SCM key, SCM def))
+static SCM Inline hash_table_search(SCM ht, SCM key)
 {
   int index = 0;
   SCM func, l = STk_nil;
 
-  ENTER_PRIMITIVE(hash_get);
   if (!HASHP(ht)) error_bad_hash_table(ht);
 
   switch (HASH_TYPE(ht)) {
-    case hash_eqp: 
+    case hash_eqp:
       index = RANDOM_INDEX(ht, key);
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (CAR(CAR(l)) == key) return CDR(CAR(l));
       break;
 
     case hash_stringp:
       index = hash_scheme_string(key) & HASH_MASK(ht);
 
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (STk_streq(CAR(CAR(l)), key) != STk_false) return CDR(CAR(l));
       break;
 
     case hash_general:
       func = HASH_COMPAR(ht);
-      index = RANDOM_INDEX(ht, 
+      index = RANDOM_INDEX(ht,
 			   STk_integer_value(STk_C_apply(HASH_HASH(ht), 1, key)));
 
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l))
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) return CDR(CAR(l));
       break;
     default: break;
   }
 
   /* Not found */
-  if (def) return def;
-  STk_error("entry not defined for key ~S in table ~S", key, ht);
+  return NULL;
+}
+
+DEFINE_PRIMITIVE("hash-table-ref", hash_ref, subr23, (SCM ht, SCM key, SCM def))
+{
+  SCM res = hash_table_search(ht, key);
+
+  if (res) return res;
+  if (def) return STk_C_apply(def, 0);
+  error_not_present(key, ht);
   return STk_void; 			/* never reached */
 }
-		 
 
 /*
-<doc ext hash-table-remove!
- * (hash-table-remove hash key)
+<doc EXT hash-table-ref/default
+ * (hash-table-ref/default hash key default)
  *
- * Deletes the entry for |key| in |hash|, if it exists. Result of 
- * |hash-table-remove!| is @emph{void}.
- * 
+ * This function is equivalent to
  * @lisp
- * (define h (make-hash-table))
- * (hash-table-put! h 'foo (list 1 2 3))
- * (hash-table-get h 'foo)          => (1 2 3)
- * (hash-table-remove! h 'foo) 
- * (hash-table-get h 'foo 'absent)  => absent
+ * (hash-table-ref hash key (lambda () default))
  * @end lisp
 doc>
 */
-DEFINE_PRIMITIVE("hash-table-remove!", hash_remove, subr2, (SCM ht, SCM key))
+DEFINE_PRIMITIVE("hash-table-ref/default", hash_ref_default, subr3,
+		 (SCM ht, SCM key, SCM def))
+{
+  SCM res = hash_table_search(ht, key);
+
+  if (res) return res;
+  if (def) return def;
+  error_not_present(key,ht);
+  return STk_void; 			/* never reached */
+}
+
+/*
+<doc EXT hash-table-exists?
+ * (hash-table-exists? hash key)
+ *
+ * Returns |#t| if there is any association of |key| in
+ * |hash|. Returns |#f| otherwise.
+doc>
+*/
+DEFINE_PRIMITIVE("hash-table-exists?", hash_existp, subr2, (SCM ht, SCM key))
+{
+  return MAKE_BOOLEAN(hash_table_search(ht,key) != NULL);
+}
+
+
+/*
+<doc EXT hash-table-delete!
+ * (hash-table-delete! hash key)
+ *
+ * Deletes the entry for |key| in |hash|, if it exists. Result of
+ * |hash-table-delete!| is ,(emph "void").
+ *
+ * @lisp
+ * (define h (make-hash-table))
+ * (hash-table-set! h 'foo (list 1 2 3))
+ * (hash-table-ref h 'foo)                => (1 2 3)
+ * (hash-table-delete! h 'foo)
+ * (hash-table-ref h 'foo
+                     (lambda () 'absent)  => absent
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("hash-table-delete!", hash_delete, subr2, (SCM ht, SCM key))
 {
   int index = 0;
   SCM func, l, prev;
 
-  ENTER_PRIMITIVE(hash_remove);
   if (!HASHP(ht)) error_bad_hash_table(ht);
 
   l = prev = STk_nil;
-  
+
   switch (HASH_TYPE(ht)) {
-    case hash_eqp: 
+    case hash_eqp:
       index = RANDOM_INDEX(ht, key);
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); prev=l, l=CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); prev=l, l=CDR(l))
 	if (CAR(CAR(l)) == key) break;
       break;
 
     case hash_stringp:
       index = hash_scheme_string(key) & HASH_MASK(ht);
-      
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l);  prev=l, l=CDR(l)) 
+
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l);  prev=l, l=CDR(l))
 	if (STk_streq(CAR(CAR(l)), key) != STk_false) break;
       break;
 
     case hash_general:
       func = HASH_COMPAR(ht);
-      index = RANDOM_INDEX(ht, 
+      index = RANDOM_INDEX(ht,
 			   STk_integer_value(STk_C_apply(HASH_HASH(ht), 1, key)));
 
-      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); prev=l, l=CDR(l)) 
+      for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); prev=l, l=CDR(l))
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) break;
       break;
     default: break;
@@ -630,7 +745,7 @@ DEFINE_PRIMITIVE("hash-table-remove!", hash_remove, subr2, (SCM ht, SCM key))
     HASH_NENTRIES(ht) -= 1;
     if (NULLP(prev))
       HASH_BUCKETS(ht)[index] = CDR(l); /* somewhere in the chain */
-    else 
+    else
       CDR(prev) = CDR(l);		/* delete the first item of the chain */
   }
   return STk_void;
@@ -639,32 +754,36 @@ DEFINE_PRIMITIVE("hash-table-remove!", hash_remove, subr2, (SCM ht, SCM key))
 
 
 /*
-<doc ext hash-table-for-each
+<doc EXT hash-table-walk hash-table-for-each
  * (hash-table-for-each hash proc)
+ * (hash-table-walk hash proc)
  *
  * |Proc| must be a procedure taking two arguments. |Hash-table-for-each|
- * calls |proc| on each key/value association in |hash|, with the key as 
- * the first argument and the value as the second.  The value returned by 
- * |hash-table-for-each| is @emph{void}. 
- * 
- * @strong{Note}: The order of application of |proc| is unspecified.
+ * calls |proc| on each key/value association in |hash|, with the key as
+ * the first argument and the value as the second.  The value returned by
+ * |hash-table-for-each| is ,(emph "void").
+ *  ,(linebreak)
+ * ,(bold "Note:") The order of application of |proc| is unspecified.
+ *  ,(linebreak)
+ * ,(bold "Note:") |hash-table-walk| is another name for |hash-table-for-each|
+ * (this is the name used in ,(link-srfi 69)).
+ *
  * @lisp
  * (let ((h   (make-hash-table))
  *       (sum 0))
- *   (hash-table-put! h 'foo 2)
- *   (hash-table-put! h 'bar 3)
+ *   (hash-table-set! h 'foo 2)
+ *   (hash-table-set! h 'bar 3)
  *   (hash-table-for-each h (lambda (key value)
  *                            (set! sum (+ sum value))))
- *   sum)           =>  5 
+ *   sum)           =>  5
  * @end lisp
 doc>
 */
 DEFINE_PRIMITIVE("hash-table-for-each", hash_for_each, subr2, (SCM ht, SCM proc))
 {
-  int i, n; 
+  int i, n;
   SCM l = STk_nil;;
-  
-  ENTER_PRIMITIVE(hash_for_each);
+
   if (!HASHP(ht)) 			 error_bad_hash_table(ht);
   if (STk_procedurep(proc) == STk_false) error_bad_procedure(proc);
 
@@ -677,32 +796,31 @@ DEFINE_PRIMITIVE("hash-table-for-each", hash_for_each, subr2, (SCM ht, SCM proc)
 
 
 /*
-<doc ext hash-table-map
+<doc EXT hash-table-map
  * (hash-table-map hash proc)
  *
  * |Proc| must be a procedure taking two arguments. |Hash-table-map|
- * calls |proc| on each key/value association in |hash|, with the key as 
- * the first argument and the value as the second.  The result of 
- * |hash-table-map| is a list of the values returned by |proc|, in an 
+ * calls |proc| on each key/value association in |hash|, with the key as
+ * the first argument and the value as the second.  The result of
+ * |hash-table-map| is a list of the values returned by |proc|, in an
  * unspecified order.
- * 
- * @strong{Note}: The order of application of |proc| is unspecified.
+ * ,(linebreak)
+ * ,(bold "Note:") The order of application of |proc| is unspecified.
  * @lisp
  * (let ((h (make-hash-table)))
- *   (dotimes (i 5) 
- *     (hash-table-put! h i (number->string i)))
+ *   (dotimes (i 5)
+ *     (hash-table-set! h i (number->string i)))
  *   (hash-table-map h (lambda (key value)
- *                        (cons key value))))    
+ *                        (cons key value))))
  *              => ((3 . "3") (4 . "4") (0 . "0") (1 . "1") (2 . "2"))
  * @end lisp
 doc>
 */
 DEFINE_PRIMITIVE("hash-table-map", hash_map, subr2, (SCM ht, SCM proc))
 {
-  int i, n; 
+  int i, n;
   SCM res, l;
-  
-  ENTER_PRIMITIVE(hash_map);
+
   if (!HASHP(ht)) 			 error_bad_hash_table(ht);
   if (STk_procedurep(proc) == STk_false) error_bad_procedure(proc);
 
@@ -717,11 +835,11 @@ DEFINE_PRIMITIVE("hash-table-map", hash_map, subr2, (SCM ht, SCM proc))
 }
 
 /*
-<doc ext hash-table-hash
+<doc EXT hash-table-hash
  * (hash-table-hash obj)
  *
  * Computes a hash code for an object and returns this hash code as a
- * non negative integer. A property of |hash-table-hash| is that 
+ * non negative integer. A property of |hash-table-hash| is that
  * @lisp
  * (equal? x y) => (equal? (hash-table-hash x) (hash-table-hash y)
  * @end lisp
@@ -738,36 +856,35 @@ DEFINE_PRIMITIVE("hash-table-hash", hash_hash, subr1, (SCM obj))
 
 
 /*
-<doc ext hash-table-stats
+<doc EXT hash-table-stats
  * (hash-table-stats hash)
  * (hash-table-stats hash port)
  *
  * Prints  overall information about |hash|, such as the number of entries
  * it contains, the number of buckets in its hash array, and the utilization
  * of the buckets. Informations are printed on |port|. If no |port| is given
- * to |hash-table-stats|, information are printed on the current output port 
- * (see @pxref{current-output-port}).
+ * to |hash-table-stats|, information are printed on the current output port
+ * (see ,(ref :mark "current-output-port")).
 doc>
 */
 DEFINE_PRIMITIVE("hash-table-stats", hash_stats, subr12, (SCM ht, SCM port))
 {
-  ENTER_PRIMITIVE(hash_stats);
-
   if(!HASHP(ht)) error_bad_hash_table(ht);
-  if (!port) port = STk_curr_oport;
+  if (!port) port = STk_current_output_port();
   else if (!OPORTP(port)) STk_error("bad port ~S", port);
-  
+
   hash_stats((struct hash_table_obj *) ht, port);
   return STk_void;
 }
 
+
 /*===========================================================================*\
- * 
+ *
  * 	Initialization code
  *
 \*===========================================================================*/
 
-struct extended_type_descr xtype_hash = { "hash-table", NULL };
+static struct extended_type_descr xtype_hash = { "hash-table", NULL };
 
 int STk_init_hash(void)
 {
@@ -777,14 +894,22 @@ int STk_init_hash(void)
   /* Define primitives */
   ADD_PRIMITIVE(make_hash);
   ADD_PRIMITIVE(hashtablep);
-  ADD_PRIMITIVE(hash_put);
-  ADD_PRIMITIVE(hash_get);
-  ADD_PRIMITIVE(hash_remove);
-  
+  ADD_PRIMITIVE(hash_table_size);
+  ADD_PRIMITIVE(hash_table_eqv_func);
+  ADD_PRIMITIVE(hash_table_hash_func);
+
+  ADD_PRIMITIVE(hash_set);
+  ADD_PRIMITIVE(hash_ref);
+  ADD_PRIMITIVE(hash_ref_default);
+  ADD_PRIMITIVE(hash_existp);
+
+  ADD_PRIMITIVE(hash_delete);
+
   ADD_PRIMITIVE(hash_for_each);
   ADD_PRIMITIVE(hash_map);
 
   ADD_PRIMITIVE(hash_hash);
   ADD_PRIMITIVE(hash_stats);
+
   return TRUE;
 }
