@@ -2,25 +2,25 @@
  *
  * k e y w o r d . c				-- Keywords management
  *
- * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright © 1993-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 19-Nov-1993 16:12
- * Last file update: 11-Apr-2002 13:42 (eg)
+ * Last file update: 27-Jul-2011 22:53 (eg)
  */
 
 #include "stklos.h"
@@ -43,27 +43,32 @@ static void error_bad_list(SCM obj)
   STk_error("bad list of keywords ~S", obj);
 }
 
+static void error_const_cell(SCM x)
+{
+  STk_error("changing the constant ~s is not allowed", x);
+}
 
 static SCM make_uninterned_keyword(char *name)
 {
   SCM z;
-  
+
   NEWCELL(z, keyword);
   SYMBOL_PNAME(z) = name;	/* already duplicated in STk_makekey */
-  if (STk_need_bars(name))
-    BOXED_INFO(z) |= KEYWORD_NEEDS_BARS;
+  BOXED_INFO(z) |= STk_symbol_flags(name);
   return z;
 }
 
 
 SCM STk_makekey(char *token)
 {
+  SCM res;
   char *s;
+  MUT_DECL(lck);
 
-  /* We accept two kinds of keywords :xy and xy:. In anycase, the value 
+  /* We accept two kinds of keywords :xy and xy:. In anycase, the value
    * stored does not contain the ':' char.
    */
-  if (*token == ':') 
+  if (*token == ':')
     s = STk_strdup(token + 1);
   else {
     int len = strlen(token);
@@ -74,11 +79,15 @@ SCM STk_makekey(char *token)
       s[len - 1] = '\0';
     }
   }
-  return STk_hash_intern_symbol(&keyword_table, s, make_uninterned_keyword);
+  MUT_LOCK(lck);
+  res =  STk_hash_intern_symbol(&keyword_table, s, make_uninterned_keyword);
+  MUT_UNLOCK(lck);
+
+  return res;
 }
 
 /*===========================================================================*\
- * 
+ *
  * 				PRIMITIVES
  *
 \*===========================================================================*/
@@ -88,7 +97,7 @@ SCM STk_makekey(char *token)
  * (make-keyword s)
  *
  * Builds a keyword from the given |s|. The parameter |s| must be a symbol
- * or a string. 
+ * or a string.
  * @lisp
  * (make-keyword "test")    => :test
  * (make-keyword 'test)     => :test
@@ -100,8 +109,7 @@ DEFINE_PRIMITIVE("make-keyword", make_keyword, subr1, (SCM str))
 {
   char *s = "";
 
-  ENTER_PRIMITIVE(make_keyword);
-  if (STRINGP(str)) 
+  if (STRINGP(str))
     s = STRING_CHARS(str);
   else if (SYMBOLP(str))
     s = SYMBOL_PNAME(str);
@@ -142,13 +150,12 @@ DEFINE_PRIMITIVE("keyword->string", keyword2string, subr1, (SCM obj))
 {
  SCM res;
 
- ENTER_PRIMITIVE(keyword2string);
  if (!KEYWORDP(obj)) error_bad_keyword(obj);
  res = STk_Cstring2string(KEYWORD_PNAME(obj));
  return res;
 }
 
- 
+
 /*
 <doc EXT  key-get
  * (key-get list key)
@@ -163,7 +170,7 @@ DEFINE_PRIMITIVE("keyword->string", keyword2string, subr1, (SCM obj))
  * @lisp
  * (key-get '(:one 1 :two 2) :one)     => 1
  * (key-get '(:one 1 :two 2) :four #f) => #f
- * (key-get '(:one 1 :two 2) :four)    => @w{@emph{error}}
+ * (key-get '(:one 1 :two 2) :four)    => error
  * @end lisp
 doc>
  */
@@ -171,7 +178,6 @@ DEFINE_PRIMITIVE("key-get", key_get, subr23, (SCM l, SCM key, SCM dflt))
 {
   SCM save = l;
 
-  ENTER_PRIMITIVE(key_get);
   if (!KEYWORDP(key)) error_bad_keyword(key);
   if (CONSP(l) || NULLP(l)) {
     int i, len = STk_int_length(l);
@@ -197,14 +203,14 @@ DEFINE_PRIMITIVE("key-get", key_get, subr23, (SCM l, SCM key, SCM dflt))
  * (key-set! list key value)
  *
  * |List| must be a list of keywords and their respective values.
- * |key-set!| sets the value associated to |key| in the keyword list. 
- * If the key is already present in |list|, the keyword list is 
- * @emph{physically} changed. 
+ * |key-set!| sets the value associated to |key| in the keyword list.
+ * If the key is already present in |list|, the keyword list is
+ * ,(emph "physically") changed.
  * @lisp
- * (let ((l '(:one 1 :two 2)))
+ * (let ((l (list :one 1 :two 2)))
  *   (set! l (key-set! l :three 3))
- *   (cons (key-get l :one) 
- *         (key-get l :three))            => (1 . 3)
+ *   (cons (key-get l :one)
+ *         (key-get l :three)))            => (1 . 3)
  * @end lisp
 doc>
   */
@@ -212,7 +218,6 @@ DEFINE_PRIMITIVE("key-set!", key_set, subr3, (SCM l, SCM key, SCM val))
 {
   SCM save = l;
 
-  ENTER_PRIMITIVE(key_set);
   if (!KEYWORDP(key)) error_bad_keyword(key);
   if (CONSP(l) || NULLP(l)) {
     int i, len = STk_int_length(l);
@@ -222,13 +227,14 @@ DEFINE_PRIMITIVE("key-set!", key_set, subr3, (SCM l, SCM key, SCM val))
     for (i = 0; i < len; i+=2) {
       if (!KEYWORDP(CAR(l))) error_bad_keyword(CAR(l));
       if (strcmp(KEYWORD_PNAME(CAR(l)), KEYWORD_PNAME(key))==0) {
+	if (BOXED_INFO(l) & CONS_CONST) error_const_cell(l);
 	CAR(CDR(l)) = val;
 	return save;
       }
       l = CDR(CDR(l));
     }
     /* Key was not here. Add it to the end of the list to allow step/step build */
-    return STk_dappend2(save, LIST2(key, val));
+    return STk_dappend2(save, LIST2(key, val)); /* Test on constantness OK */
   } else {
     error_bad_list(l);
     return STk_void; /* never reached */
@@ -236,17 +242,18 @@ DEFINE_PRIMITIVE("key-set!", key_set, subr3, (SCM l, SCM key, SCM val))
 }
 
 /*
-<doc EXT key-delete key-delete! 
+<doc EXT key-delete key-delete!
+*  (key-delete  list key)
  * (key-delete! list key)
  *
  * |List| must be a list of keywords and their respective values.
- * |key-delete| remove the |key| and its associated value of the keyword 
- * list. The key can be absent of the list. 
- * 
+ * |key-delete| remove the |key| and its associated value of the keyword
+ * list. The key can be absent of the list.
+ * ,(linebreak)
  * |key-delete!| does the
  * same job than |key-delete| by physically modifying its |list| argument.
  * @lisp
- * (key-delete '(:one 1 :two 2) :one)    => (:one 1)
+ * (key-delete '(:one 1 :two 2) :two)    => (:one 1)
  * (key-delete '(:one 1 :two 2) :three)  => (:one 1 :two 2)
  * @end lisp
 doc>
@@ -264,7 +271,8 @@ static SCM key_del(SCM l, SCM key)
     for (i = 0; i < len; i+=2) {
       if (!KEYWORDP(CAR(l))) error_bad_keyword(CAR(l));
       if (strcmp(KEYWORD_PNAME(CAR(l)), KEYWORD_PNAME(key))==0) {
-	if (prev == l) 
+	if (BOXED_INFO(l) & CONS_CONST) error_const_cell(l);
+	if (prev == l)
 	  return CDR(CDR(l));
 	else {
 	  CDR(prev) = CDR(CDR(l));
@@ -280,17 +288,15 @@ static SCM key_del(SCM l, SCM key)
     return STk_void; /* never reached */
   }
 }
-  
+
 
 DEFINE_PRIMITIVE("key-delete!", dkey_delete, subr2, (SCM l, SCM key))
 {
-  ENTER_PRIMITIVE(dkey_delete);
   return key_del(l, key);
 }
 
 DEFINE_PRIMITIVE("key-delete", key_delete, subr2, (SCM l, SCM key))
 {
-  ENTER_PRIMITIVE(key_delete);
   return key_del(STk_copy_tree(l), key);
 }
 
