@@ -2,7 +2,7 @@
  *
  * l i s t . c			-- Lists procedures
  *
- * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright © 1993-2005 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  *
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: ??-Oct-1993 21:37
- * Last file update: 19-Dec-2002 11:12 (eg)
+ * Last file update: 10-Oct-2005 19:08 (eg)
  */
 
 #include "stklos.h"
@@ -75,7 +75,7 @@ int STk_int_length(SCM l)
 SCM STk_argv2list(int argc, SCM *argv)
 {
   SCM res = STk_nil;
-  
+
   while (argc--) {
     res = STk_cons(argv[-argc], res);
   }
@@ -136,13 +136,11 @@ DEFINE_PRIMITIVE("car", car, subr1, (SCM x))
  *     (car '(a b c))          =>  a
  *     (car '((a) b c d))      =>  (a)
  *     (car '(1 . 2))          =>  1
- *     (car '())               =>  @w{@emph{error}}
+ *     (car '())               =>  error
  * @end lisp
 doc>
  */
 {
-  ENTER_PRIMITIVE(car);
-
   if (CONSP(x)) return CAR(x);
   error_wrong_type(x);
   return STk_void; /* never reached */
@@ -159,13 +157,11 @@ DEFINE_PRIMITIVE("cdr", cdr, subr1, (SCM x))
  * @lisp
  *     (cdr '((a) b c d))      =>  (b c d)
  *     (cdr '(1 . 2))          =>  2
- *     (cdr '())               =>  @w{@emph{error}}
+ *     (cdr '())               =>  error
  * @end lisp
 doc>
  */
 {
-  ENTER_PRIMITIVE(cdr);
-
   if (CONSP(x)) return CDR(x);
   error_wrong_type(x);
   return STk_void; /* never reached */
@@ -178,18 +174,16 @@ DEFINE_PRIMITIVE("set-car!", setcar, subr2, (SCM cell, SCM value))
  * (set-car! pair obj)
  *
  * Stores |obj| in the car field of |pair|.
- * The value returned by |set-car!| is @emph{void}.
+ * The value returned by |set-car!| is ,(emph "void").
  * @lisp
  *    (define (f) (list 'not-a-constant-list))
  *    (define (g) '(constant-list))
  *    (set-car! (f) 3)
- *    (set-car! (g) 3)             =>  @w{@emph{error}}
+ *    (set-car! (g) 3)             =>  error
  * @end lisp
 doc>
  */
 {
-  ENTER_PRIMITIVE(setcar);
-
   if (!CONSP(cell)) 		     error_wrong_type(cell);
   if (BOXED_INFO(cell) & CONS_CONST) error_const_cell(cell);
 
@@ -204,13 +198,11 @@ DEFINE_PRIMITIVE("set-cdr!", setcdr, subr2, (SCM cell, SCM value))
  * (set-cdr! pair obj)
  *
  * Stores |obj| in the cdr field of |pair|.
- * The value returned by |set-cdr!| is @emph{void}.
- * 
+ * The value returned by |set-cdr!| is ,(emph "void").
+ *
 doc>
  */
 {
-  ENTER_PRIMITIVE(setcdr);
-
   if (!CONSP(cell)) error_wrong_type(cell);
   if (BOXED_INFO(cell) & CONS_CONST) error_const_cell(cell);
 
@@ -291,7 +283,6 @@ doc>
 {
   int len = STk_int_length(l);
 
-  ENTER_PRIMITIVE(list_length);
   if (len >= 0) return MAKE_INT(len);
   STk_error("length of ~S is not calculable", l);
   return STk_void; /* never reached */
@@ -353,7 +344,6 @@ Error:
 
 DEFINE_PRIMITIVE("append", append, vsubr, (int argc, SCM* argv))
 {
-  ENTER_PRIMITIVE(append);
   switch (argc) {
     case 0:  return STk_nil;
     case 1:  return *argv;
@@ -381,7 +371,6 @@ DEFINE_PRIMITIVE("reverse", reverse, subr1, (SCM l))
 {
   SCM p, n = STk_nil;
 
-  ENTER_PRIMITIVE(reverse);
   for(p=l; !NULLP(p); p=CDR(p)) {
     if (!CONSP(p)) STk_error("bad list ~W", l);
     n = STk_cons(CAR(p), n);
@@ -410,8 +399,6 @@ DEFINE_PRIMITIVE("list-tail", list_tail, subr2, (SCM list, SCM k))
   register long x;
   SCM l;
 
-  ENTER_PRIMITIVE(list_tail);
-
   if (!CONSP(list) && !NULLP(list)) STk_error("bad list ~W", list);
 
   x = STk_integer_value(k);
@@ -448,8 +435,6 @@ DEFINE_PRIMITIVE("list-ref", list_ref, subr2, (SCM list, SCM k))
   register long x;
   SCM l = list;
 
-  ENTER_PRIMITIVE(list_ref);
-
   if (!CONSP(list)) STk_error("bad list ~W", list);
 
   x = STk_integer_value(k);
@@ -501,7 +486,6 @@ doc>
 {								\
   register SCM ptr;						\
 								\
-  ENTER_PRIMITIVE(prim)						\
   if (!CONSP(list) && !NULLP(list)) goto Error;			\
   for (ptr=list; !NULLP(ptr); ) { 				\
     if (CONSP(ptr)) {						\
@@ -554,7 +538,7 @@ DEFINE_PRIMITIVE("member", member, subr2, (SCM obj, SCM list))
  *                               =>  (5 7)
  * @end lisp
  *
- * @b{Rationale}: Although they are ordinarily used as predicates,
+ * ,(bold "Rationale:") Although they are ordinarily used as predicates,
  * |memq|, |memv|, |member|, |assq|, |assv|, and |assoc| do not have question
  * marks in their names because they return useful values rather than just
  * |#t| or #|f|.
@@ -565,7 +549,6 @@ doc>
 {								\
   register SCM l,tmp;						\
 								\
-  ENTER_PRIMITIVE(prim);					\
   for(l=alist; CONSP(l); ) {					\
     tmp = CAR(l);						\
     if (CONSP(tmp) && compare(CAR(tmp), obj)) return tmp;	\
@@ -597,8 +580,8 @@ DEFINE_PRIMITIVE("assoc", assoc, subr2, (SCM obj, SCM alist))
 /*
 <doc EXT pair-mutable?
  * (pair-mutable? obj)
- * 
- * Returns |#t| if |obj| is a mutable pair, otherwise returns |#f|. 
+ *
+ * Returns |#t| if |obj| is a mutable pair, otherwise returns |#f|.
  * @lisp
  * (pair-mutable? '(1 . 2))    => #f
  * (pair-mutable? (cons 1 2))  => #t
@@ -617,7 +600,7 @@ DEFINE_PRIMITIVE("pair-mutable?", pair_mutable, subr1, (SCM obj))
  * (list* obj ...)
  *
  * |list*| is like |list| except that the last argument to |list*| is
- * used as the @emph{cdr} of the last pair constructed.
+ * used as the ,(emph "cdr") of the last pair constructed.
  * @lisp
  *    (list* 1 2 3)        => (1 2 . 3)
  *    (list* 1 2 3 '(4 5)) => (1 2 3 4 5)
@@ -671,8 +654,7 @@ doc>
 DEFINE_PRIMITIVE("last-pair", last_pair, subr1, (SCM l))
 {
   SCM tmp;
-  
-  ENTER_PRIMITIVE(last_pair);
+
   if (!CONSP(l)) error_wrong_type(l);
   for (tmp=l; CONSP(CDR(l)); l = CDR(l))
     /* Nothing */;
@@ -680,16 +662,16 @@ DEFINE_PRIMITIVE("last-pair", last_pair, subr1, (SCM l))
 }
 
 /*
-<doc EXT filter filter! 
+<doc EXT filter filter!
  * (filter  pred list)
  * (filter! pred list)
  *
  * |Filter| returns all the elements of |list| that satisfy predicate
  * |pred|. The |list| is not disordered: elements that appear in the
  * result list occur in the same order as they occur in the argument
- * list. |Filter!| does the same job than |filter| by physically 
- * modifying its |list| argument 
- * @lisp 
+ * list. |Filter!| does the same job than |filter| by physically
+ * modifying its |list| argument
+ * @lisp
  * (filter even? '(0 7 8 8 43 -4)) => (0 8 8 -4)
  * (let* ((l1 (list 0 7 8 8 43 -4))
  *        (l2 (filter! even? l1)))
@@ -703,13 +685,12 @@ DEFINE_PRIMITIVE("filter", filter, subr2, (SCM pred, SCM list))
   register SCM ptr, l;
   SCM result;
 
-  ENTER_PRIMITIVE(filter);
   if (STk_procedurep(pred) != STk_true) error_bad_proc(pred);
   if (!CONSP(list) && !NULLP(list)) error_bad_list(list);
 
   for (ptr=l=list, result=STk_nil; !NULLP(l); ) {
     if (!CONSP(l)) error_bad_list(list);
-    
+
     if (STk_C_apply(pred, 1, CAR(l)) != STk_false) {
       if (NULLP(result)) {
 	NEWCELL(result, cons);
@@ -732,7 +713,6 @@ DEFINE_PRIMITIVE("filter!", dfilter, subr2, (SCM pred, SCM list))
 {
   SCM previous, l, start=list;
 
-  ENTER_PRIMITIVE(dfilter);
   if (STk_procedurep(pred) != STk_true) error_bad_proc(pred);
   if (!CONSP(list) && !NULLP(list)) error_bad_list(list);
 
@@ -743,7 +723,7 @@ DEFINE_PRIMITIVE("filter!", dfilter, subr2, (SCM pred, SCM list))
     if (STk_C_apply(pred, 1, CAR(l)) == STk_false) {
       if (previous == STk_nil)
 	list = CDR(list);
-      else 
+      else
 	CDR(previous) = CDR(l);
     } else {
       previous = l;
@@ -759,8 +739,8 @@ DEFINE_PRIMITIVE("filter!", dfilter, subr2, (SCM pred, SCM list))
  *
  * Returns a list consisting of the elements of the first list
  * followed by the elements of the other lists.
- * Contrarily to |append|, the parameter lists (except the last one) are 
- * physically modified: their last pair is changed to the value of the next 
+ * Contrarily to |append|, the parameter lists (except the last one) are
+ * physically modified: their last pair is changed to the value of the next
  * list in the |append!| formal parameter list.
  * @lisp
  * (let* ((l1 (list 1 2))
@@ -779,7 +759,7 @@ SCM STk_dappend2(SCM l1, SCM l2)
 
   if (NULLP(l1)) return l2;
 
-  for (tmp = l1; CONSP(tmp); tmp = CDR(tmp)) {
+  for (tmp = l1; ; tmp = CDR(tmp)) {
     if (BOXED_INFO(tmp) & CONS_CONST) error_const_cell(tmp);
     if (NULLP(CDR(tmp))) {
       CDR(tmp) = l2;
@@ -793,7 +773,6 @@ SCM STk_dappend2(SCM l1, SCM l2)
 
 DEFINE_PRIMITIVE("append!", dappend, vsubr, (int argc, SCM* argv))
 {
-  ENTER_PRIMITIVE(dappend);
   switch (argc) {
     case 0:  return STk_nil;
     case 1:  return *argv;
@@ -813,7 +792,7 @@ DEFINE_PRIMITIVE("append!", dappend, vsubr, (int argc, SCM* argv))
  * @lisp
  * (let ((l '(a b c)))
  *   (list (reverse! l) l))        =>  ((c b a) (a))
- * (reverse! '(a constant list))   =>  @w{@emph{errror}}   
+ * (reverse! '(a constant list))   =>  error
  * @end lisp
 doc>
  */
@@ -821,8 +800,6 @@ DEFINE_PRIMITIVE("reverse!", dreverse, subr1, (SCM l))
 {
   SCM tmp, p, prev;
 
-  ENTER_PRIMITIVE(dreverse);
-  
   for(p=l, prev=STk_nil; !NULLP(p); prev=p, p=tmp) {
     if (!CONSP(p)) STk_error("bad list ~W", l);
     if (BOXED_INFO(p) & CONS_CONST) error_const_cell(p);
@@ -848,7 +825,7 @@ SCM STk_dremq(SCM obj, SCM list)
     if (obj == CAR(l)) {
       if (previous == STk_nil)
 	list = CDR(list);
-      else 
+      else
 	CDR(previous) = CDR(l);
     } else {
       previous = l;
@@ -858,6 +835,89 @@ SCM STk_dremq(SCM obj, SCM list)
   return list;
 }
 
+/*
+ *
+ * Fast version of assq; for internal use only (alist must be well formed)
+ *
+ */
+SCM STk_int_assq(SCM obj, SCM alist)
+{
+  register SCM l;
+
+  for(l=alist; !NULLP(l); l = CDR(l)) {
+    if (CAR(CAR(l)) == obj) return CAR(l);
+  }
+  return STk_false;
+}
+
+
+
+/* ======================================================================
+ *
+ * Extended pairs
+ *
+ * ======================================================================
+ */
+
+struct econs_obj {		/* Use a mapping wchich is identical to cons */
+  stk_header header;
+  SCM car;
+  SCM cdr;
+  char *file;
+  int line;
+  int pos;
+};
+
+#define ECONS_FILE(p)	(((struct econs_obj *) (p))->file)
+#define ECONS_LINE(p)	(((struct econs_obj *) (p))->line)
+#define ECONS_POS(p)	(((struct econs_obj *) (p))->pos)
+#define ECONSP(obj)	(CONSP(obj) && (BOXED_INFO(obj) & CONS_ECONS))
+
+
+SCM STk_econs(SCM car, SCM cdr, char *file, int line, int pos)
+{
+  SCM z;
+
+  NEWCELL_WITH_LEN(z, cons, sizeof(struct econs_obj));
+  CAR(z) 	= car;
+  CDR(z) 	= cdr;
+  ECONS_FILE(z) = file;
+  ECONS_LINE(z) = line;
+  ECONS_POS(z)  = pos;
+  BOXED_INFO(z) |= CONS_ECONS;
+
+  return z;
+}
+
+DEFINE_PRIMITIVE("%epair?", epairp, subr1, (SCM obj))
+{
+  return MAKE_BOOLEAN(ECONSP(obj));
+}
+
+DEFINE_PRIMITIVE("%epair-file", epair_file, subr1, (SCM obj))
+{
+  if (ECONSP(obj)) return STk_Cstring2string(ECONS_FILE(obj));
+  error_wrong_type(obj);
+  return STk_void; /* never reached */
+}
+
+
+DEFINE_PRIMITIVE("%epair-line", epair_line, subr1, (SCM obj))
+{
+  if (ECONSP(obj)) return MAKE_INT(ECONS_LINE(obj));
+  error_wrong_type(obj);
+  return STk_void; /* never reached */
+}
+
+
+DEFINE_PRIMITIVE("%epair-position", epair_position, subr1, (SCM obj))
+{
+  if (ECONSP(obj)) return MAKE_INT(ECONS_POS(obj));
+  error_wrong_type(obj);
+  return STk_void; /* never reached */
+}
+
+
 
 int STk_init_list(void)
 {
@@ -890,5 +950,11 @@ int STk_init_list(void)
   ADD_PRIMITIVE(dfilter);
   ADD_PRIMITIVE(dappend);
   ADD_PRIMITIVE(dreverse);
+
+  ADD_PRIMITIVE(epairp);
+  ADD_PRIMITIVE(epair_file);
+  ADD_PRIMITIVE(epair_line);
+  ADD_PRIMITIVE(epair_position);
+
   return TRUE;
 }
