@@ -1,30 +1,30 @@
 /*
  * p a t h . h			-- Path names management
- * 
- * Copyright  2000-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
- * 
+ *
+ * Copyright © 2000-2004 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
- * 
+ *
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date:  9-Jan-2000 14:25 (eg)
- * Last file update: 16-Jan-2003 19:28 (eg)
+ * Last file update: 14-Jul-2004 16:15 (eg)
  */
 
-#ifdef WIN32
+#ifdef OLD_WIN32
 #  include <windows.h>
 #  include <io.h>
 #  define F_OK   00
@@ -38,7 +38,7 @@
 #  define DIRSEP 	 '\\'
 #  define SDIRSEP  	 "\\"
 #  define PATHSEP	 ';'
-#else /* WIN32 */
+#else /* OLD_WIN32 */
 #  include <unistd.h>
 #  include <pwd.h>
 #  define ISDIRSEP(ch) 	 ((ch)=='/')
@@ -46,17 +46,17 @@
 #  define DIRSEP 	 '/'
 #  define SDIRSEP  	 "/"
 #  define PATHSEP	 ':'
-#endif /* WIN32 */
+#endif /* OLD_WIN32 */
 
 
 #include <sys/types.h>
 #include <sys/stat.h>
 
-#ifdef WIN32
-   /* One of above files includes <stdarg.h> with BC++  (and stdarg and 
-    * vararg are not compatible 
+#ifdef OLD_WIN32
+   /* One of above files includes <stdarg.h> with BC++  (and stdarg and
+    * vararg are not compatible
     */
-#  undef __STDARG_H 
+#  undef __STDARG_H
 
 #  ifdef _MSC_VER
 #  include <direct.h>
