@@ -1,44 +1,48 @@
 /*
  * p r i n t . c				-- writing stuff
  *
- * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ * Copyright © 1993-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
  *
  *           Author: Erick Gallesio [eg@unice.fr]
- *    Creation date: ??-Oct-1993 ??:?? 
- * Last file update:  8-May-2002 21:14 (eg)
+ *    Creation date: ??-Oct-1993 ??:??
+ * Last file update: 23-Jul-2011 12:12 (eg)
  *
  */
-
+#include <ctype.h>
 #include "stklos.h"
 
+static int pretty_quotes = 1;
+
 
 static void printlist(SCM exp, SCM port, int mode)
 {
   register SCM tmp;
   char *s;
 
-  /* Special case for pretty printing of quoted expressions */
-  s = STk_quote2str(CAR(exp));
-  if (s && !NULLP(CDR(exp)) && NULLP(CDR(CDR(exp)))) {
-    STk_puts(s, port);
-    STk_print(CAR(CDR(exp)), port, mode);
-    return;
+  if (pretty_quotes) {
+    /* Special case for pretty printing of quoted expressions */
+    s = STk_quote2str(CAR(exp));
+    if (s && !NULLP(CDR(exp)) && NULLP(CDR(CDR(exp)))) {
+      STk_puts(s, port);
+      STk_print(CAR(CDR(exp)), port, mode);
+      return;
+    }
   }
 
   STk_putc('(', port);
@@ -49,7 +53,7 @@ static void printlist(SCM exp, SCM port, int mode)
     STk_print(CAR(tmp), port, mode);
   }
   if (!NULLP(tmp)) {
-    STk_puts(" . ", port);
+    STk_nputs(port, " . ", 3);
     STk_print(tmp, port, mode);
   }
   STk_putc(')', port);
@@ -59,11 +63,13 @@ static void printlist(SCM exp, SCM port, int mode)
 static void Inline printsymbol(SCM symb, SCM port, int mode)
 {
   char *s = SYMBOL_PNAME(symb);
-  
-  if ((mode==WRT_MODE) && (BOXED_INFO(symb) & SYMBOL_NEEDS_BARS)) {
+
+  if ((mode==WRT_MODE) &&
+      ((BOXED_INFO(symb) & SYMBOL_NEEDS_BARS) ||
+       (!STk_read_case_sensitive && (BOXED_INFO(symb) & SYMBOL_HAS_UPPER)))) {
     STk_putc('|', port);  STk_puts(s, port); STk_putc('|', port);
   } else
-    STk_puts(s, port);
+    STk_puts(*s ? s: "||", port); /* print bars around the "null" symbol */
 }
 
 static void Inline printkeyword(SCM key, SCM port, int mode)
@@ -71,8 +77,9 @@ static void Inline printkeyword(SCM key, SCM port, int mode)
   char *s = KEYWORD_PNAME(key);
 
   if (mode==WRT_MODE) {
-    if (BOXED_INFO(key) & SYMBOL_NEEDS_BARS) {
-      STk_puts("|:", port);  STk_puts(s, port); STk_putc('|', port);
+    if ((BOXED_INFO(key) & SYMBOL_NEEDS_BARS) ||
+	(!STk_read_case_sensitive && (BOXED_INFO(key) & SYMBOL_HAS_UPPER))) {
+      STk_nputs(port, "|:", 2);  STk_puts(s, port); STk_putc('|', port);
       return;
     }
     STk_putc(':', port);
@@ -81,20 +88,20 @@ static void Inline printkeyword(SCM key, SCM port, int mode)
 }
 
 
+static char printhexa(int x)
+{
+  return (x >= 10) ? (x - 10 + 'a') : (x + '0');
+}
+
 static void printstring(SCM s, SCM port, int mode)
 {
-  register unsigned char *p  = STRING_CHARS(s);
-  register size_t len        = STRING_SIZE(s);
-  
   if (mode == DSP_MODE) {
-    if (strlen(p) == len)
-      /* No null in the string. We can use a puts, instead of the slow STk_putc */
-      STk_puts((char *) p, port);
-    else
-      for (len = STRING_SIZE(s); len; len--, p++) STk_putc(*p, port);
+    STk_putstring(s, port);
   } else {
+    register char *p    = STRING_CHARS(s);
+    register size_t len = STRING_SIZE(s);
     char buffer[MAX_TOKEN_SIZE], *buff = buffer;
-    
+
     *buff++ = '"';
     for (   ; len; len--, p++) {
       if (buff >= buffer + MAX_TOKEN_SIZE - 7) { /* 7 because we can add \X" and */
@@ -113,16 +120,20 @@ static void printstring(SCM s, SCM port, int mode)
         case '\r' : *buff++ = '\\'; *buff++ = 'r'; break;
         case '\t' : *buff++ = '\\'; *buff++ = 't'; break;
         case '\v' : *buff++ = '\\'; *buff++ = 'v'; break;
-        case '"'  : 
+        case '"'  :
         case '\\' : *buff++ = '\\'; *buff++ = *p;  break;
-        default   : if (*p < ' ') { 
-	  	      /* Non printable character (This work only for ASCII!!) */
+      default   : if (STk_use_utf8)
+		    *buff++ = *p;
+		  else {
+		    if ((((unsigned char) *p) & 0177) < (unsigned char) ' ') {
+	  	      /* Non printable character (It works only for ISO 8859-x !!) */
 	  	      *buff++ = '\\';
-	  	      *buff++ = '0';
-		      *buff++ = (*p / 8) + '0';
-		      *buff++ = (*p % 8) + '0';	  
-		    } 
+	  	      *buff++ = 'x';
+		      *buff++ = printhexa((unsigned char) *p / 16);
+		      *buff++ = printhexa((unsigned char) *p % 16);
+		    }
 		    else *buff++ = *p;
+		  }
       }
     }
     *buff++ = '"';
@@ -140,27 +151,43 @@ void STk_print(SCM exp, SCM port, int mode)
   if (SCONSTP(exp)) {
     /* Expression is a small constant */
     switch (AS_LONG(exp)) {
-      case AS_LONG(STk_nil):   STk_puts("()", port); 		  	return;
-      case AS_LONG(STk_false): STk_puts("#f", port); 		  	return;
-      case AS_LONG(STk_true):  STk_puts("#t", port); 		  	return;
-      case AS_LONG(STk_eof):   STk_puts("#[eof]", port);	  	return;
-      case AS_LONG(STk_void):  STk_puts("#[void]", port);	 	return;
+      case AS_LONG(STk_nil):   STk_nputs(port, "()", 2); 	  	return;
+      case AS_LONG(STk_false): STk_nputs(port, "#f", 2); 	  	return;
+      case AS_LONG(STk_true):  STk_nputs(port, "#t", 2); 	  	return;
+      case AS_LONG(STk_eof):   STk_nputs(port, "#eof", 4);	  	return;
+      case AS_LONG(STk_void):  STk_nputs(port, "#void", 5);	 	return;
       default:		       STk_panic("Bad small constant %d", exp); return;
     }
   }
-  
+
   if (INTP(exp)) {
-    sprintf(buffer, "%ld", INT_VAL(exp));
-    STk_puts(buffer, port);
+    int len = sprintf(buffer, "%ld", INT_VAL(exp));
+    STk_nputs(port, buffer, len);
     return;
   }
 
   if (CHARACTERP(exp)) {
+    char buffer[5] = {0};
+    int c = CHARACTER_VAL(exp);
+
     if (mode!=DSP_MODE){
+      char *s = STk_char2string(c);
+
       STk_puts("#\\", port);
-      STk_puts(STk_char2string(CHARACTER_VAL(exp)), port);
+      if (s)
+	STk_puts(s, port);
+      else
+	if (c < 0x80)
+	  STk_putc(c, port);
+	else {
+	  STk_char2utf8(c, buffer);
+	  STk_puts((char *) buffer, port);
+	}
+    }
+    else {
+      STk_char2utf8(c, buffer);
+      STk_puts((char *) buffer, port);
     }
-    else STk_putc(CHARACTER_VAL(exp), port);
     return;
   }
 
@@ -169,10 +196,7 @@ void STk_print(SCM exp, SCM port, int mode)
       printlist(exp, port, mode);
       return;
     case tc_real:
-      /* //FIXME: Traiter NaN et INFINITY */
-      sprintf(buffer, "%.*g", REAL_FORMAT_SIZE, REAL_VAL(exp));
-      if (strchr(buffer, '.') == NULL && strchr(buffer, 'e') == NULL) 
-	strcat(buffer, ".0");
+      STk_double2Cstr(buffer, REAL_VAL(exp));
       STk_puts(buffer, port);
       return;
     case tc_symbol:
@@ -180,11 +204,28 @@ void STk_print(SCM exp, SCM port, int mode)
       return;
     case tc_keyword:
       printkeyword(exp, port, mode);
-      break;
+      return;
     case tc_string:
       printstring(exp, port, mode);
       return;
-    case tc_subr0:       /* ==================> Utiliser un type tendu //FIXME */
+    case tc_box:
+      STk_putc('#', port);
+      STk_putc('&', port);
+      STk_print(BOX_VALUE(exp), port, mode);
+      return;
+    case tc_pointer:
+      if (CPOINTER_TYPE(exp) == STk_void) {
+	sprintf(buffer, "#[C-pointer %lx @ %lx]",
+		(unsigned long) CPOINTER_VALUE(exp), (unsigned long) exp);
+      } else {
+	STk_puts("#[", port);
+	STk_print(CPOINTER_TYPE(exp), port, mode);
+	sprintf(buffer, "-pointer %lx @ %lx]", (unsigned long) CPOINTER_VALUE(exp),
+		(unsigned long) exp);
+      }
+      STk_puts(buffer, port);
+      return;
+    case tc_subr0:       /* ==================> Utiliser un type étendu //FIXME */
     case tc_subr1:
     case tc_subr2:
     case tc_subr3:
@@ -199,10 +240,21 @@ void STk_print(SCM exp, SCM port, int mode)
       STk_puts(PRIMITIVE_NAME(exp), port);
       STk_putc(']', port);
       return;
-    default: 
+#ifdef HAVE_FFI
+    case tc_ext_func:
+      STk_puts("#[external-func ", port);
+      STk_puts(STRING_CHARS(STk_ext_func_name(exp)), port);
+      STk_putc(']', port);
+      return;
+    case tc_callback:
+      sprintf(buffer, "#[callback %lx]", (unsigned long) exp);
+      STk_puts(buffer, port);
+      return;
+#endif
+    default:
       {
 	struct extended_type_descr *xdescr = BOXED_XTYPE(exp);
-	
+
 	if (xdescr) {
 	  void (*p)() = XTYPE_PRINT(xdescr);
 
@@ -223,9 +275,9 @@ void STk_print(SCM exp, SCM port, int mode)
 
 
 
-/*============================================================================= 
- * 
- *			Printing of circular structures 
+/*=============================================================================
+ *
+ *			Printing of circular structures
  *
  *=============================================================================*/
 
@@ -237,11 +289,12 @@ static void pass2(SCM exp, SCM port);	/* pass 2: print      */
 
 
 static void print_cycle(SCM exp, SCM port)
-{  
+{
   SCM value, tmp;
 
   if ((tmp = STk_assv(exp, cycles)) != STk_false) {
-    if (INTP(value = CDR(tmp))) {
+    value= CDR(tmp);
+    if (INTP(value)) {
       STk_fprintf(port, "#%ld#", INT_VAL(value));
       return;
     }
@@ -258,12 +311,14 @@ static void printlist_star(SCM exp, SCM port)
 
   tmp = STk_nil;		/* for GCC */
 
-  /* Special case for pretty printing of quoted expressions */
-  s = STk_quote2str(CAR(exp));
-  if (s && !NULLP(CDR(exp)) && NULLP(CDR(CDR(exp)))) {
-    STk_puts(s, port);
-    print_cycle(CAR(CDR(exp)), port);
-    return;
+  if (pretty_quotes) {
+    /* Special case for pretty printing of quoted expressions */
+    s = STk_quote2str(CAR(exp));
+    if (s && !NULLP(CDR(exp)) && NULLP(CDR(CDR(exp)))) {
+      STk_puts(s, port);
+      print_cycle(CAR(CDR(exp)), port);
+      return;
+    }
   }
 
   STk_putc('(', port);
@@ -274,9 +329,9 @@ static void printlist_star(SCM exp, SCM port)
     if (NULLP(exp=CDR(exp))) break;
 
     if (!CONSP(exp) || (tmp = STk_assv(exp, cycles)) != STk_false) {
-      if (!CONSP(exp) || (value = CDR(tmp)) == STk_true || INTP(value)) { 
+      if (!CONSP(exp) || (value = CDR(tmp)) == STk_true || INTP(value)) {
 	/* either  ". X" or ". #0=(...)" or ". #0#" */
-	STk_puts(" . ", port);
+	STk_nputs(port, " . ", 3);
 	print_cycle(exp, port);
 	break;
       }
@@ -290,8 +345,8 @@ static void printlist_star(SCM exp, SCM port)
 static void printvector_star(SCM exp, SCM port)
 {
   int j, n = VECTOR_SIZE(exp);
-  
-  STk_puts("#(", port);
+
+  STk_nputs(port, "#(", 2);
   for(j=0; j < n; j++) {
     print_cycle(VECTOR_DATA(exp)[j], port);
     if ((j + 1) < n) STk_putc(' ', port);
@@ -313,7 +368,7 @@ Top:
 
     if (CONSP(exp)) {			/* it's a cons */
       pass1(CAR(exp));
-      exp = CDR(exp); 
+      exp = CDR(exp);
       goto Top;
     }
     else { 				/* it's a vector */
@@ -321,7 +376,7 @@ Top:
       for (i = 0; i < len; i++) pass1(VECTOR_DATA(exp)[i]);
       if (len >= 0) {exp = VECTOR_DATA(exp)[len]; goto Top;}
     }
-  } 
+  }
   else {
     /* This item was already seen. Note that this is the second time */
     CDR(tmp) = STk_true;
@@ -352,9 +407,53 @@ static void pass2(SCM exp, SCM port)
 
 void STk_print_star(SCM exp, SCM port)
 {
+  MUT_DECL(lck);
+
   if (!CONSP(exp) &&  !VECTORP(exp)) return STk_print(exp, port, WRT_MODE);
+  MUT_LOCK(lck);
   cycles      = STk_nil;
   index_label = 0;
 
   pass1(exp); pass2(exp, port);
+  MUT_UNLOCK(lck);
+}
+
+/*
+<doc EXT write-pretty-quotes
+ * (write-pretty-quotes)
+ * (write-pretty-quotes value)
+ *
+ * This parameter object permits to change the default behaviour of
+ * the |display| or |write| primitives when they write a list which starts with
+ *  the symbol quote,  quasiquote, unquote or unquote-splicing. If this parameter
+ * has a false value, the writer uses the list notation instead of a
+ * more human-readable value.
+ * By default, this parameter value is set to |#t|.
+ * @lisp
+ * (let ((x ''a))
+ *   (display x)
+ *   (display " ")
+ *   (write-pretty-quotes #f)
+ *   (display x))               @print 'a (quote a)
+ * @end lisp
+doc>
+*/
+static SCM write_pretty_quotes_conv(SCM value)
+{
+  pretty_quotes = (value != STk_false);
+  return MAKE_BOOLEAN(pretty_quotes);
+}
+
+/*===========================================================================*\
+ *
+ *			I n i t i a l i z a t i o n
+ *
+\*===========================================================================*/
+int STk_init_printer(void)
+{
+  STk_make_C_parameter("write-pretty-quotes",
+		       MAKE_BOOLEAN(pretty_quotes),
+		       write_pretty_quotes_conv,
+		       STk_STklos_module);
+  return TRUE;
 }
