@@ -1,27 +1,27 @@
 /*
  * socket.c				-- Socket acess for STklos
- * 
- * Copyright  2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
- * 
+ *
+ * Copyright © 2003-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
- * 
+ *
  *           Author: Erick Gallesio [eg@essi.fr]
  *    Creation date:  3-Jan-2003 18:45 (eg)
- * Last file update: 20-Jun-2003 23:54 (eg)
+ * Last file update: 27-May-2011 22:55 (eg)
  */
 
 #include <sys/types.h>
@@ -33,53 +33,23 @@
 #include <errno.h>
 
 #include <stklos.h>
-
+#include "socket.h"
 
 
 /*===========================================================================*\
- * 
- *			Definition of the socket type 
+ *
+ *			Definition of the socket type
  *
 \*===========================================================================*/
-
-static int tc_socket;		/* the key for socket objects */
-
-struct socket_obj {
-  stk_header header;
-  long portnum;
-  SCM hostname, hostip;
-  int type, fd;
-  SCM input, output;
-  SCM ready_event;
-  SCM data; 	/* parent for a client socket , children for a server socket */
-};
-
-#define SOCKETP(x)  		(BOXED_TYPE_EQ((x),tc_socket))
-#define SOCKET_PORTNUM(x) 	(((struct socket_obj *) (x))->portnum)
-#define SOCKET_HOSTNAME(x)	(((struct socket_obj *) (x))->hostname)
-#define SOCKET_HOSTIP(x)	(((struct socket_obj *) (x))->hostip)
-#define SOCKET_FD(x)		(((struct socket_obj *) (x))->fd)
-#define SOCKET_TYPE(x)		(((struct socket_obj *) (x))->type)
-#define SOCKET_IN(x)		(((struct socket_obj *) (x))->input)
-#define SOCKET_OUT(x)		(((struct socket_obj *) (x))->output)
-#define SOCKET_READY(x)		(((struct socket_obj *) (x))->ready_event)
-#define SOCKET_PARENT(x)	(((struct socket_obj *) (x))->data)
-#define SOCKET_CHILDREN(x)	(((struct socket_obj *) (x))->data) /* also */
-
-#define SOCKET_SERVER		0
-#define SOCKET_CLIENT		1
-
-
 static void print_socket(SCM sock, SCM port, int mode)
 {
   char buffer[1000];
   SCM name = SOCKET_HOSTNAME(sock);
-  
-  sprintf(buffer, "#[%s-socket %s:%ld %lx]", 
+
+  sprintf(buffer, "#[%s-socket %s:%ld (%d)]",
 	  (SOCKET_TYPE(sock) == SOCKET_SERVER) ? "server" : "client",
 	  ((name == STk_false) ?  "*none*" : (char *) STRING_CHARS(name)),
-	  SOCKET_PORTNUM(sock),
-	  (unsigned long) sock);
+	  SOCKET_PORTNUM(sock), SOCKET_FD(sock));
   STk_puts(buffer, port);
 }
 
@@ -89,9 +59,10 @@ static struct extended_type_descr xtype_socket = {
 };
 
 
+static void socket_finalizer(SCM socket);
 
 /*===========================================================================*\
- * 
+ *
  *				UTILITIES
  *
 \*===========================================================================*/
@@ -129,26 +100,20 @@ static void set_socket_io_ports(SCM sock, int line_buffered)
   SCM in, out;
 
   hostname = STRING_CHARS(SOCKET_HOSTNAME(sock));
-  fname    = STk_must_malloc_atomic(strlen(hostname) + 30);  
+  fname    = STk_must_malloc_atomic(strlen(hostname) + 30);
   sprintf(fname, "%s:%ld", hostname, SOCKET_PORTNUM(sock));
-		       
-  
-  /* duplicate handles so that we are able to access one socket channel via 
-   * two scheme ports 
-   */
-  s = SOCKET_FD(sock);
-  t = dup(s); 
-  if (t == -1) STk_error("cannot duplicate io port");
 
+  s   = SOCKET_FD(sock);
+  t   = dup(s);
   in  = STk_fd2scheme_port(s, "r", fname);
-  out = STk_fd2scheme_port(s, "w", fname);
+  out = STk_fd2scheme_port(t, "w", fname);
 
-  if (NULLP(in) || NULLP(out)) 
+  if (NULLP(in) || NULLP(out))
     STk_error("cannot create socket IO ports");
 
   SOCKET_IN(sock)  = in;
   SOCKET_OUT(sock) = out;
-  
+
   /* Set line buffered mode (if needed) */
   if (line_buffered) {
     STk_set_line_buffered_mode(in); /* not really useful */
@@ -156,8 +121,9 @@ static void set_socket_io_ports(SCM sock, int line_buffered)
   }
 }
 
+
 /*===========================================================================*\
- * 
+ *
  *				MAKE-SERVER-SOCKET
  *
 \*===========================================================================*/
@@ -166,20 +132,19 @@ static void set_socket_io_ports(SCM sock, int line_buffered)
 <doc EXT make-server-socket
  * (make-server-socket)
  * (make-server-socket port-number)
- * 
+ *
  * |make-server-socket| returns a new socket object. If |port-number|
- * is specified, the socket is listening on the specified port; 
- * otherwise, the communication port is chosen by the system. 
+ * is specified, the socket is listening on the specified port;
+ * otherwise, the communication port is chosen by the system.
 doc>
 */
 DEFINE_PRIMITIVE("make-server-socket", make_server_socket, subr01, (SCM port))
 {
   struct sockaddr_in sin;
-  int s, len;
+  socklen_t len;
   int long portnum;
   SCM z;
-  
-  ENTER_PRIMITIVE(make_server_socket);
+  int s, yes = 1;
 
   /* Determine port to use */
   portnum = (port) ? STk_integer_value(port) : 0;
@@ -188,7 +153,13 @@ DEFINE_PRIMITIVE("make-server-socket", make_server_socket, subr01, (SCM port))
   /* Create a socket */
   s = socket(AF_INET, SOCK_STREAM, 0);
   if (s < 0) STk_error("cannot create socket");
-  
+
+  /* Allow reusing the socket port */
+  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) < 0) {
+    close(s);
+    STk_error(strerror(errno));
+  }
+
   /* Bind the socket to a name */
   sin.sin_family      = AF_INET;
   sin.sin_port 	      = htons(portnum);
@@ -201,7 +172,7 @@ DEFINE_PRIMITIVE("make-server-socket", make_server_socket, subr01, (SCM port))
 
   /* Query the socket name (permits to get the true socket number if 0 was given */
   len = sizeof(sin);
-  if (getsockname(s, (struct sockaddr *) &sin, (int *) &len) < 0) {
+  if (getsockname(s, (struct sockaddr *) &sin,  &len) < 0) {
     close(s);
     STk_error(strerror(errno));
   }
@@ -223,12 +194,16 @@ DEFINE_PRIMITIVE("make-server-socket", make_server_socket, subr01, (SCM port))
   SOCKET_OUT(z)		= STk_false;
   SOCKET_READY(z)	= STk_false;
   SOCKET_CHILDREN(z)	= STk_nil;
+  SOCKET_USER_DATA(z)	= NULL;
+
+  STk_register_finalizer(z, socket_finalizer);
+
   return z;
 }
 
 
 /*===========================================================================*\
- * 
+ *
  *				MAKE-CLIENT-SOCKET
  *
 \*===========================================================================*/
@@ -236,26 +211,26 @@ DEFINE_PRIMITIVE("make-server-socket", make_server_socket, subr01, (SCM port))
 /*
 <doc EXT make-client-socket
  *  (make-client-socket hostname port-number)
- *  (make-client-socket hostname port_number line-buffered) 
- * 
+ *  (make-client-socket hostname port_number line-buffered)
+ *
  * |make-client-socket| returns a new socket object. This socket
  * establishes a link between the running program and the application
  * listening on port |port-number| of |hostname|.  If  the optional argument
- * |line-buffered| has a true value, a line buffered policy is used when writing 
- * to the client socket (i.e. characters on the socket are tranmitted as soon 
- * as a @code{#\newline} character is encountered). The default value of 
- * |line-buffered| is |t|.
- * 
+ * |line-buffered| has a true value, a line buffered policy is used when writing
+ * to the client socket (i.e. characters on the socket are tranmitted as soon
+ * as a ,(code "#\\newline") character is encountered). The default value of
+ * |line-buffered| is |#t|.
+ *
 doc>
 */
 static SCM internal_init_client_socket(SCM hostname, SCM port)
 {
   struct hostent *hp;
   struct sockaddr_in server;
-  int s;
+  int s, yes=1;
   SCM z;
   long val = STk_integer_value(port);
-  
+
   /* Verify arguments */
   if(!STRINGP(hostname)) error_bad_hostname(hostname);
   if(val == LONG_MIN)    error_bad_portnum(port);
@@ -266,7 +241,7 @@ static SCM internal_init_client_socket(SCM hostname, SCM port)
 
   /* Get a socket */
   if ((s=socket(AF_INET,SOCK_STREAM,0)) < 0) error_cannot_crate_socket();
-  
+
   /* Setup a connect address */
   memset(&server, 0, sizeof(server));
   memcpy((char*)&server.sin_addr, hp->h_addr, hp->h_length);
@@ -274,14 +249,15 @@ static SCM internal_init_client_socket(SCM hostname, SCM port)
   server.sin_port   = htons(val);
 
   /* Try to connect */
-  if (connect(s, (struct sockaddr *) &server, sizeof(server)) < 0) {
+  if ((connect(s, (struct sockaddr *) &server, sizeof(server))    < 0) ||
+      (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) < 0)) {
     close(s);
     STk_error("system reported \"%s\"", strerror(errno));
   }
-  
+
   /* Create a new Scheme socket object */
   NEWCELL(z, socket);
-  
+
   SOCKET_PORTNUM(z)  = ntohs(server.sin_port); /* Query true value */
   SOCKET_HOSTNAME(z) = STk_Cstring2string((char *) hp->h_name);
   SOCKET_HOSTIP(z)   = STk_Cstring2string((char *) inet_ntoa(server.sin_addr));
@@ -291,6 +267,9 @@ static SCM internal_init_client_socket(SCM hostname, SCM port)
   SOCKET_OUT(z)	     = STk_false;
   SOCKET_READY(z)    = STk_false;
   SOCKET_PARENT(z)   = STk_nil;
+  SOCKET_USER_DATA(z)= NULL;
+
+  STk_register_finalizer(z, socket_finalizer);
 
   return z;
 }
@@ -299,7 +278,6 @@ static SCM internal_init_client_socket(SCM hostname, SCM port)
 DEFINE_PRIMITIVE("%initialize-client-socket", init_client_socket, subr2,
 				 (SCM hostname, SCM port))
 {
-  ENTER_PRIMITIVE(init_client_socket);
   return internal_init_client_socket(hostname, port);
 }
 
@@ -309,7 +287,6 @@ DEFINE_PRIMITIVE("make-client-socket", make_client_socket, subr23,
 {
  SCM sock;
 
- ENTER_PRIMITIVE(make_client_socket);
  if (!line_buffered) line_buffered = STk_true;
 
  sock = internal_init_client_socket(hostname, port);
@@ -318,8 +295,9 @@ DEFINE_PRIMITIVE("make-client-socket", make_client_socket, subr23,
  return sock;
 }
 
+
 /*===========================================================================*\
- * 
+ *
  *				SOCKET-SHUTDOWN
  *
 \*===========================================================================*/
@@ -331,56 +309,57 @@ DEFINE_PRIMITIVE("make-client-socket", make_client_socket, subr23,
  *
  * |Socket-shutdown| shutdowns the connection associated to
  * |socket|. If the socket is a server socket, |socket-shutdown| is called
- * on all the clientsockets connected to this server.
+ * on all the client sockets connected to this server.
  * |Close| indicates if the the socket must be closed or not, when
- * the connection is destroyed. Closing the socket forbids further 
+ * the connection is destroyed. Closing the socket forbids further
  * connections on the same port with the |socket-accept| procedure.
  * Omitting a value for |close| implies the closing of socket.
- * 
- * The following example shows a simple server: when there is 
- * a new connection on the port number 12345, the server displays 
+ * @l
+ * The following example shows a simple server: when there is
+ * a new connection on the port number 12345, the server displays
  * the first line sent to it by the client, discards the others and
  * go back waiting for further client connections.
  * @lisp
  * (let ((s (make-server-socket 12345)))
  *    (let loop ()
  *       (let ((ns (socket-accept s)))
- *         (format t "I've read: ~A\n" (read-line (socket-input ns)))
- *         (socket-shutdown ns f)
+ *         (format #t "I've read: ~A\\n"
+ *                 (read-line (socket-input ns)))
+ *         (socket-shutdown ns #f)
  *         (loop))))
  * @end lisp
 doc>
 */
-DEFINE_PRIMITIVE("socket-shutdown", socket_shutdown, subr12, (SCM sock, SCM close))
+DEFINE_PRIMITIVE("socket-shutdown", socket_shutdown, subr12, (SCM sock, SCM closeit))
 {
-  ENTER_PRIMITIVE(socket_shutdown);
-
   if (!SOCKETP(sock)) error_bad_socket(sock);
-  if (close != STk_false) close = STk_true;
+  if (closeit != STk_false) closeit = STk_true;
 
   /* Eventually close the associated sockets */
   if (SOCKET_TYPE(sock) == SOCKET_SERVER) { /* Shutdown associated client sockets */
     SCM tmp;
 
     for (tmp = STk_copy_tree(SOCKET_CHILDREN(sock)); !NULLP(tmp); tmp = CDR(tmp)){
-      STk_socket_shutdown(CAR(tmp), close);
+      STk_socket_shutdown(CAR(tmp), closeit);
     }
-  } else {    /* Delete this socket form the server socket (if needed) */
+  } else {    /* Delete this socket from the server socket (if needed) */
     SCM parent = SOCKET_PARENT(sock);
 
     if (!NULLP(parent)) {
       SOCKET_CHILDREN(parent) = STk_dremq(sock, SOCKET_CHILDREN(parent));
+      SOCKET_PARENT(sock) = STk_nil;
     }
   }
-    
-  /* 
+
+  /*
    * Warning: input and output can have already be garbaged: if the
    * socket is no more used, the input and output are not marked as
    * used and can (eventually) be released before the call to shutdown
    * by the socket finalizer.
    */
-  if ((close == STk_true) && SOCKET_FD(sock) >= 0) {
-    shutdown(SOCKET_FD(sock), 2);
+  if (SOCKET_FD(sock) >= 0) {
+    if (closeit == STk_true) shutdown(SOCKET_FD(sock), 2);
+    close(SOCKET_FD(sock));
     SOCKET_FD(sock) = -1;
   }
   if (SOCKET_IN(sock) != STk_false) {
@@ -391,12 +370,23 @@ DEFINE_PRIMITIVE("socket-shutdown", socket_shutdown, subr12, (SCM sock, SCM clos
     STk_close_port(SOCKET_OUT(sock));
     SOCKET_OUT(sock) = STk_false;
   }
+
+  /* Call user shutdown hook */
+  if (SOCKET_USER_DATA(sock))
+    (*SOCKET_USER_SHUTDOWN(sock))(sock);
+
   return STk_void;
 }
 
+
+static void socket_finalizer(SCM socket)
+{
+  STk_socket_shutdown(socket, STk_true);
+}
+
 /*===========================================================================*\
- * 
- *				SOCKET-ACCEPT 
+ *
+ *				SOCKET-ACCEPT
  *
 \*===========================================================================*/
 
@@ -404,59 +394,55 @@ DEFINE_PRIMITIVE("socket-shutdown", socket_shutdown, subr12, (SCM sock, SCM clos
 <doc EXT socket-accept
  * (socket-accept socket)
  * (socket-accept socket line-buffered)
- * 
+ *
  * |socket-accept| waits for a client connection on the given
  * |socket|. If no client is already waiting for a connection, this
- * procedure blocks its caller; otherwise, the first connection request 
- * on the queue of pending connections is connected and |socket-accept| 
- * returns a new client socket to serve this request. 
+ * procedure blocks its caller; otherwise, the first connection request
+ * on the queue of pending connections is connected and |socket-accept|
+ * returns a new client socket to serve this request.
  * This procedure must be called on a server socket created
- * with |make-server-socket|. The result of |socket-accept| is undefined. 
- * |Line-buffered| indicates if the port should be considered as a 
- * line buffered. If |line-buffered| is omitted, it defaults to t.
- * 
+ * with |make-server-socket|. The result of |socket-accept| is undefined.
+ * |Line-buffered| indicates if the port should be considered as a
+ * line buffered. If |line-buffered| is omitted, it defaults to |#t|.
+ * @l
  * The following example is a simple server which waits for a connection
- * on the port 12345 @footnote{Under Unix, you can simply connect to
+ * on the port 12345 ,(footnote [Under Unix, you can simply connect to
  * a listening socket with the |telnet| command. With the given
- * example, this can be achieved by typing the following command in a 
- * window shell: |$ telnet localhost 12345|}. Once the connection with the
+ * example, this can be achieved by typing the following command in a
+ * window shell:
+ * ,(raw-code "$ telnet localhost 12345")]).
+ *  Once the connection with the
  * distant program is established, we read a line on the input port
  * associated to the socket and we write the length of this line on its
- * output port. 
+ * output port.
  * @lisp
  * (let* ((server (make-server-socket 13345))
  *        (client (socket-accept server))
  *        (l      (read-line (socket-input client))))
- *   (format (socket-output client) "Length is: ~a\n" (string-length l))
+ *   (format (socket-output client)
+ *           "Length is: ~a\n" (string-length l))
  *   (socket-shutdown server))
  * @end lisp
  *
- * @noindent
- * Note that shutting down the |server| socket suffices here to close
- * also the connection to |client|. 
+ *  Note that shutting down the |server| socket suffices here to close
+ * also the connection to |client|.
 doc>
 */
-DEFINE_PRIMITIVE("socket-accept", socket_accept, subr01, 
+DEFINE_PRIMITIVE("socket-accept", socket_accept, subr12,
 		 (SCM serv, SCM line_buffered))
 {
   char *s;
   struct sockaddr_in sin;
   struct hostent *host;
-  int len = sizeof(sin);
-  int new_s, new_fd;
+  socklen_t len = sizeof(sin);
+  int new_s;
   SCM z;
 
-  ENTER_PRIMITIVE(socket_accept);
-
   if (!SOCKETP(serv)) STk_error("bad socket ~S", serv);
   if (!line_buffered) line_buffered = STk_true;
- 
-  /* Dup the file descriptor of the server socket */
-  new_fd = dup(SOCKET_FD(serv));
-  if (new_fd <0) STk_error("cannot duplicate file descriptor");
-  
-  /* Accept the connection on the new file */
-  new_s = accept(new_fd, (struct sockaddr *) &sin, &len);
+
+  /* Accept the connection */
+  new_s = accept(SOCKET_FD(serv), (struct sockaddr *) &sin, &len);
   if (new_s < 0) STk_error(strerror(errno));
 
   /* Set the client info (if possible its name, otherwise its IP number) */
@@ -474,19 +460,22 @@ DEFINE_PRIMITIVE("socket-accept", socket_accept, subr01,
   SOCKET_OUT(z)	     = STk_false;
   SOCKET_READY(z)    = STk_false;
   SOCKET_PARENT(z)   = serv;
+  SOCKET_USER_DATA(z)= NULL;
 
   set_socket_io_ports(z, (line_buffered != STk_false));
 
   /* Place this socket in the socket list of its parent */
   SOCKET_CHILDREN(serv) = STk_cons(z, SOCKET_CHILDREN(serv));
 
+  STk_register_finalizer(z, socket_finalizer);
+
   return z;
 }
 
 
 /*===========================================================================*\
- * 
- *				SOCKET PRIMITIVES 
+ *
+ *				SOCKET PRIMITIVES
  *
 \*===========================================================================*/
 
@@ -494,7 +483,7 @@ DEFINE_PRIMITIVE("socket-accept", socket_accept, subr01,
 <doc EXT socket?
  * (socket? obj)
  *
- * Returns t if |socket| is a socket, otherwise returns f.
+ * Returns |#t| if |socket| is a socket, otherwise returns |#f|.
 doc>
 */
 DEFINE_PRIMITIVE("socket?", socketp, subr1, (SCM obj))
@@ -507,7 +496,7 @@ DEFINE_PRIMITIVE("socket?", socketp, subr1, (SCM obj))
 <doc EXT socket-client?
  * (socket-client? obj)
  *
- * Returns t if |socket| is a client socket, otherwise returns f.
+ * Returns |#t| if |socket| is a client socket, otherwise returns |#f|.
 doc>
 */
 DEFINE_PRIMITIVE("socket-client?", socket_clientp, subr1, (SCM obj))
@@ -520,10 +509,10 @@ DEFINE_PRIMITIVE("socket-client?", socket_clientp, subr1, (SCM obj))
 <doc EXT socket-server?
  * (socket-server? obj)
  *
- * Returns t if |socket| is a server socket, otherwise returns f.
+ * Returns |#t| if |socket| is a server socket, otherwise returns |#f|.
 doc>
 */
-DEFINE_PRIMITIVE("socket-client?", socket_serverp, subr1, (SCM obj))
+DEFINE_PRIMITIVE("socket-server?", socket_serverp, subr1, (SCM obj))
 {
   return MAKE_BOOLEAN(SOCKETP(obj) && SOCKET_TYPE(obj) == SOCKET_CLIENT);
 }
@@ -538,8 +527,6 @@ doc>
 */
 DEFINE_PRIMITIVE("socket-port-number", socket_port_number, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_port_number);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return MAKE_INT(SOCKET_PORTNUM(sock));
 }
@@ -552,24 +539,22 @@ DEFINE_PRIMITIVE("socket-port-number", socket_port_number, subr1, (SCM sock))
  *
  * Returns the file port associated for reading or writing with the
  * program connected with |socket|. If no connection has already been
- * established, these functions return f.
+ * established, these functions return |#f|.
  *
  * The following example shows how to make a client socket. Here we
- * create a socket on port 13 of the machine |kaolin.unice.fr|@footnote{
- * Port 13 is generally used for testing:
+ * create a socket on port 13 of the machine |kaolin.unice.fr|,(footnote
+ * [Port 13 is generally used for testing:
  * making a connection to it permits to know the distant system's idea
- * of the time of day.}:
+ * of the time of day.]):
  * @lisp
  * (let ((s (make-client-socket "kaolin.unice.fr" 13)))
- *   (format t "Time is: ~A~%" (read-line (socket-input s)))
+ *   (format #t "Time is: ~A~%" (read-line (socket-input s)))
  *   (socket-shutdown  s))
  * @end lisp
 doc>
 */
 DEFINE_PRIMITIVE("socket-input", socket_input, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_input);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return SOCKET_IN(sock);
 }
@@ -577,8 +562,6 @@ DEFINE_PRIMITIVE("socket-input", socket_input, subr1, (SCM sock))
 
 DEFINE_PRIMITIVE("socket-output", socket_output, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_output);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return SOCKET_OUT(sock);
 }
@@ -593,14 +576,12 @@ DEFINE_PRIMITIVE("socket-output", socket_output, subr1, (SCM sock))
  * |make-client-socket| this procedure returns the official name of
  * the distant machine used for connection. If |socket| has been
  * created with |make-server-socket|, this function returns the
- * official name of the client connected to the socket. If no client
- * has used yet |socket|, this function returns f.  
+< * official name of the client connected to the socket. If no client
+ * has used yet |socket|, this function returns |#f|.
 doc>
 */
 DEFINE_PRIMITIVE("socket-host-name", socket_host_name, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_host_name);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return SOCKET_HOSTNAME(sock);
 }
@@ -616,13 +597,11 @@ DEFINE_PRIMITIVE("socket-host-name", socket_host_name, subr1, (SCM sock))
  * distant machine used for connection. If |socket| has been created
  * with |make-server-socket|, this function returns the address of the
  * client connected to the socket.  If no client has used yet
- * |socket|, this function returns f.  
+ * |socket|, this function returns |#f|.
 doc>
 */
 DEFINE_PRIMITIVE("socket-host-address", socket_host_address, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_host_address);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return SOCKET_HOSTIP(sock);
 }
@@ -630,8 +609,6 @@ DEFINE_PRIMITIVE("socket-host-address", socket_host_address, subr1, (SCM sock))
 
 DEFINE_PRIMITIVE("socket-down?", socket_downp, subr1, (SCM sock))
 {
-  ENTER_PRIMITIVE(socket_downp);
-  
   if (!SOCKETP(sock)) error_bad_socket(sock);
   return MAKE_BOOLEAN(SOCKET_FD(sock) == -1);
 }
@@ -641,33 +618,32 @@ DEFINE_PRIMITIVE("socket-down?", socket_downp, subr1, (SCM sock))
 <doc EXT socket-local-address
  * (socket-local-address socket)
  *
- * Returns a string which contains the IP number of the local host 
+ * Returns a string which contains the IP number of the local host
  * attached to |socket|.
 doc>
-*/ 
+*/
 DEFINE_PRIMITIVE("socket-local-address", socket_local_addr, subr1, (SCM sock))
 {
   struct sockaddr_in sin;
-  int len = sizeof(sin);
-  
-  ENTER_PRIMITIVE(socket_local_addr);
+  socklen_t len = sizeof(sin);
+
   if (!SOCKETP(sock)) error_bad_socket(sock);
 
   if (getsockname(SOCKET_FD(sock), (struct sockaddr *) &sin, &len))
     STk_error("cannot get socket name of ~S", sock);
-  
+
   return STk_Cstring2string((char *) inet_ntoa(sin.sin_addr));
 }
 
 
 /*===========================================================================*\
- * 
+ *
  *				Initialization
- * 
+ *
 \*===========================================================================*/
 int STk_init_socket(void)
 {
-  DEFINE_USER_TYPE(tc_socket, & xtype_socket);
+  DEFINE_XTYPE(socket, & xtype_socket);
 
   ADD_PRIMITIVE(init_client_socket);
   ADD_PRIMITIVE(make_server_socket);
