@@ -1,35 +1,35 @@
 /*
  * s p o r t . c			-- String ports management
  *
- * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ * Copyright © 1993-2010 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
  *
  *            Author: Erick Gallesio [eg@unice.fr]
  *    Creation date: 17-Feb-1993 12:27
- * Last file update: 22-Feb-2002 12:31 (eg)
+ * Last file update:  9-Aug-2010 10:59 (eg)
  *
  */
 
 #include "stklos.h"
 
 /*===========================================================================*\
- * 
+ *
  * 			Utilities
- * 
+ *
 \*===========================================================================*/
 
 static void error_bad_string(SCM s)
@@ -38,36 +38,37 @@ static void error_bad_string(SCM s)
 }
 
 /*===========================================================================*\
- * 
+ *
  * Low level plugins
- * 
+ *
 \*===========================================================================*/
 
-#define START_ALLOC_SIZE  128	/* Initial size of an ouput string port */
+#define START_ALLOC_SIZE  100	/* Initial size of an ouput string port */
 
 struct sstream {
-  int		cnt;
-  unsigned char *ptr;
-  unsigned char *base;
-  int		bufsize;
-  SCM		str;	/* keep a ref on original string to avoid GC problems */
+  char *ptr;
+  char *base;
+  char *end;
+  int  bufsize;
+  SCM  str;	/* keep a ref on original string to avoid GC problems */
 };
 
-#define PORT_CNT(x)     (((struct sstream *) (x))->cnt)
-#define PORT_PTR(x)	(((struct sstream *) (x))->ptr)
 #define PORT_BASE(x)	(((struct sstream *) (x))->base)
+#define PORT_PTR(x)	(((struct sstream *) (x))->ptr)
+#define PORT_END(x)	(((struct sstream *) (x))->end)
 #define PORT_BUFSIZE(x) (((struct sstream *) (x))->bufsize)
 #define PORT_STR(x)	(((struct sstream *) (x))->str)
 
 static Inline int Sgetc(void *stream)
 {
-  return (--(PORT_CNT(stream))>=0 ? ((int) *PORT_PTR(stream)++): EOF);
+  return (PORT_PTR(stream) < PORT_END(stream)) ?
+    ((unsigned char) *PORT_PTR(stream)++) : EOF;
 }
 
 
 static Inline int Seof(void * stream)
 {
-  return PORT_CNT(stream) <= 0;
+  return (PORT_PTR(stream) >= PORT_END(stream));
 }
 
 static Inline int Sreadyp(void *stream)
@@ -80,26 +81,70 @@ static Inline int Sclose(void * stream)
   return 0;
 }
 
+static Inline int Sread(void *stream, void *buffer, int count)
+{
+  int avail = PORT_END(stream) - PORT_PTR(stream);
+
+  if (count > avail) count = avail;
+  if (count) {
+    memcpy(buffer, PORT_PTR(stream), count);
+    PORT_PTR(stream) += count;
+  }
+  return count;
+}
+
 static Inline int Sputc(int c, void *stream)
 {
   register unsigned int tmp;
-  
-  if (++PORT_CNT(stream) == PORT_BUFSIZE(stream)) {
-    tmp	  = PORT_BUFSIZE(stream);
-    tmp	 += tmp/2;
-    PORT_BASE(stream) = STk_must_realloc(PORT_BASE(stream), tmp);
-    PORT_PTR(stream)  = PORT_BASE(stream)+PORT_BUFSIZE(stream)-1;/* base can move */
-    PORT_BUFSIZE(stream) =  tmp;
+
+  if (PORT_PTR(stream) >= PORT_END(stream)) {
+    if (PORT_END(stream) == PORT_BASE(stream) + PORT_BUFSIZE(stream)) {
+      /* No more room => allocate a new buffer */
+      tmp	  = PORT_BUFSIZE(stream);
+      tmp	 += tmp/2;
+      PORT_BASE(stream)= STk_must_realloc(PORT_BASE(stream), tmp);
+      PORT_PTR(stream) = PORT_BASE(stream) + PORT_BUFSIZE(stream);/*base can move*/
+      PORT_BUFSIZE(stream) = tmp;
+    }
+    PORT_END(stream) = PORT_PTR(stream) + 1;
   }
- *PORT_PTR(stream)++ = (unsigned char) c;
- return 0;
+  *PORT_PTR(stream)++ = (unsigned char) c;
+
+  return c;
+}
+
+
+static Inline int Swrite(void *stream, void *buffer, int count)
+{
+  int tmp, pos;
+
+  if (PORT_PTR(stream) + count >= PORT_END(stream)) {
+    tmp = PORT_BUFSIZE(stream) + count + START_ALLOC_SIZE;
+    pos = PORT_PTR(stream) - PORT_BASE(stream);
+    PORT_BASE(stream) = STk_must_realloc(PORT_BASE(stream), tmp);
+    PORT_PTR(stream)  = PORT_BASE(stream)+ pos; /* base can move */
+    PORT_BUFSIZE(stream) = tmp;
+    PORT_END(stream) = PORT_PTR(stream) + count;
+  }
+  memcpy(PORT_PTR(stream), buffer, count);
+  PORT_PTR(stream) += count;
+
+  return count;
 }
 
 static Inline int Sputs(char *s, void *stream)
 {
-  while (*s)
-    Sputc(*s++, stream);
-  return 0;
+  return Swrite(stream, s, strlen(s));
+}
+
+static Inline int Sputstring(SCM s, void *stream)
+{
+  return Swrite(stream, STRING_CHARS(s), STRING_SIZE(s));
+}
+
+static Inline int Snputs(void *stream, char *s, int len)
+{
+  return Swrite(stream, s, len);
 }
 
 static Inline int Sflush(void *stream)
@@ -108,33 +153,36 @@ static Inline int Sflush(void *stream)
 }
 
 
-static Inline int Sread(void *stream, void *buffer, int count)
+static off_t Sseek(void *stream, off_t offset, int whence)
 {
-  int i, c;
+  char* p;
 
-  for (i = 0; i < count; i++) {
-    c = Sgetc(stream);
-    if (c == EOF) return i;
-    *(char*)buffer++ = c;
+  switch (whence) {
+  case SEEK_SET:
+    p = PORT_BASE(stream) + offset;
+    break;
+  case SEEK_CUR:
+    p = PORT_PTR(stream) + offset;
+    break;
+  default: /* SEEK_END */
+    p = PORT_END(stream) + offset;
+    break;
   }
-  return count;
+
+  if ((PORT_BASE(stream) <= p) && (p <= PORT_END(stream))) {
+    PORT_PTR(stream) = p;
+    return p - PORT_BASE(stream);
+  }
+  else
+    return -1;
 }
 
-static Inline int Swrite(void *stream, void *buffer, int count)
-{
-  int i;
-
-  for (i = 0; i < count; i++) {
-    Sputc(*(char*)buffer++, stream);
-  }
-  return count;
-}
 
 static void sport_print(SCM obj, SCM port)   /* Generic printing of string ports */
 {
   char buffer[MAX_PATH_LENGTH + 20];
 
-  sprintf(buffer, "#[%s-string-port %lx%s]", 
+  sprintf(buffer, "#[%s-string-port %lx%s]",
 	  ISPORTP(obj) ? "input" : "output",
 	  (unsigned long) obj,
 	  PORT_IS_CLOSEDP(obj) ? " (closed)" : "");
@@ -148,33 +196,44 @@ static void sport_release(SCM port)
 
 
 /*===========================================================================*\
- * 
+ *
  * 			Input ports
- * 
+ *
 \*===========================================================================*/
+enum kind_port {SREAD_C, SREAD, SWRITE};
 
 static struct port_obj *
-make_sport(SCM str, int init_len, int flags)
+make_sport(enum kind_port kind,  SCM str, int init_len, int flags)
 {
   struct sstream  *ss = STk_must_malloc(sizeof(struct sstream));
   SCM res;
 
   /* Initialize the stream part */
-  if (str) {
-    /* this is a input string */
-    PORT_BASE(ss) = STRING_CHARS(str);
-    PORT_CNT(ss)  = init_len;
-    PORT_STR(ss)  = str;
-  } else {
-    /* This is an output port */
-    PORT_BASE(ss) = STk_must_malloc(init_len);
-    PORT_CNT(ss)  = 0;
-    PORT_STR(ss)  = STk_false;
+  switch (kind) {
+    case SREAD:   /* this is a input string */
+		  {
+		    char *s = STRING_CHARS(str);
+
+		    PORT_BASE(ss) = s;
+		    PORT_END(ss)  = s + init_len;
+		    PORT_STR(ss)  = str;
+		    break;
+		  }
+    case SREAD_C: /* this is a input string (from a C string) */
+      		  PORT_BASE(ss) = (char *) str;
+		  PORT_END(ss)  = (char *) str + init_len;
+		  PORT_STR(ss)  = str;
+		  break;
+    case SWRITE:  /* This is an output string */
+      		  PORT_BASE(ss) = PORT_END(ss) = STk_must_malloc_atomic(init_len);
+		  PORT_STR(ss)  = STk_false;
+		  break;
   }
-  
+
   PORT_PTR(ss)     = PORT_BASE(ss);
   PORT_BUFSIZE(ss) = init_len;
 
+
   /* Initialize now the port itsef */
   NEWCELL(res, port);
   PORT_STREAM(res)	= ss;
@@ -182,7 +241,9 @@ make_sport(SCM str, int init_len, int flags)
   PORT_UNGETC(res) 	= EOF;
   PORT_LINE(res)	= 1;
   PORT_POS(res)		= 0;
-  
+  PORT_FNAME(res)	= "string port";
+  PORT_CLOSEHOOK(res)	= STk_false;
+
   PORT_PRINT(res)	= sport_print;
   PORT_RELEASE(res)	= sport_release;
   PORT_GETC(res)	= Sgetc;
@@ -191,16 +252,27 @@ make_sport(SCM str, int init_len, int flags)
   PORT_CLOSE(res)	= Sclose;
   PORT_PUTC(res)	= Sputc;
   PORT_PUTS(res)	= Sputs;
+  PORT_PUTSTRING(res)	= Sputstring;
+  PORT_NPUTS(res)	= Snputs;
   PORT_FLUSH(res)	= Sflush;
   PORT_BREAD(res)	= Sread;
   PORT_BWRITE(res)	= Swrite;
-  
+  PORT_SEEK(res)	= Sseek;
+
   return (struct port_obj *) res;
-} 
+}
+
+/*
+ * open-input-string with a C string ...
+ */
+SCM STk_open_C_string(char *str)
+{
+  return (SCM) make_sport(SREAD_C, (SCM) str, strlen(str), PORT_READ);
+}
 
 
 /*
-<doc ext open-input-string
+<doc EXT open-input-string
  * (open-input-string str)
  *
  * Returns an input string port capable of delivering characters from
@@ -209,14 +281,13 @@ doc>
  */
 DEFINE_PRIMITIVE("open-input-string", open_input_string, subr1, (SCM s))
 {
-  ENTER_PRIMITIVE(open_input_string);
   if (!STRINGP(s)) error_bad_string(s);
-  return (SCM) make_sport(s, STRING_SIZE(s), PORT_READ);
+  return (SCM) make_sport(SREAD, s, STRING_SIZE(s), PORT_READ);
 }
 
 
 /*
-<doc ext open-output-string
+<doc EXT open-output-string
  * (open-output-string)
  *
  * Returns an output string port capable of receiving and collecting characters.
@@ -224,13 +295,12 @@ doc>
  */
 DEFINE_PRIMITIVE("open-output-string", open_output_string, subr0,(void))
 {
-  return (SCM) make_sport((SCM) NULL, START_ALLOC_SIZE, PORT_WRITE);
+  return (SCM) make_sport(SWRITE, (SCM) NULL, START_ALLOC_SIZE, PORT_WRITE);
 }
 
 
-
 /*
-<doc ext get-output-string
+<doc EXT get-output-string
  * (get-output-string port)
  *
  * Returns a string containing all the text that has been written on the
@@ -242,34 +312,25 @@ DEFINE_PRIMITIVE("open-output-string", open_output_string, subr0,(void))
  * @end lisp
 doc>
  */
-SCM STk_get_output_string(SCM port)
+DEFINE_PRIMITIVE("get-output-string", get_output_string, subr1, (SCM port))
 {
   struct port_obj* p;
+  char *base;
 
-  if (! OSPORTP(port)) STk_error("bad string port ~S", port);
-#ifdef FIXME
-  // Que dit SRFI-6?
-  //  if (PORT_FLAGS(port) & PORT_CLOSED) 
-  // Err("get-output-string: string port is closed", port);
-#endif
-  p = PORT_STREAM(port);
-  return STk_makestring(PORT_CNT(p), PORT_BASE(p));
-}
+  if (! OSPORTP(port)) STk_error_bad_port(port);
 
-
-DEFINE_PRIMITIVE("get-output-string", scheme_get_output_string, subr1, (SCM port))
-{
-  ENTER_PRIMITIVE(scheme_get_output_string);
-  return STk_get_output_string(port);
+  p    = PORT_STREAM(port);
+  base = PORT_BASE(p);
+  return STk_makestring(PORT_END(p) - base, base);
 }
 
 
 /*
-<doc ext input-string-port? output-string-port?
+<doc EXT input-string-port? output-string-port?
  * (input-string-port? obj)
  * (output-string-port? obj)
  *
- * Returns |#t| if |obj| is an input string port or output string port 
+ * Returns |#t| if |obj| is an input string port or output string port
  * respectively, otherwise returns #f.
 doc>
  */
@@ -288,7 +349,7 @@ int STk_init_sport(void)
 {
   ADD_PRIMITIVE(open_input_string);
   ADD_PRIMITIVE(open_output_string);
-  ADD_PRIMITIVE(scheme_get_output_string);
+  ADD_PRIMITIVE(get_output_string);
   ADD_PRIMITIVE(input_string_portp);
   ADD_PRIMITIVE(output_string_portp);
   return TRUE;
