@@ -2,8 +2,8 @@
  *
  * s y s t e m . c				-- System relative primitives
  *
- * Copyright  1994-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
+ * Copyright © 1994-2011 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
  *
  * Permission to use, copy, modify, distribute,and license this
  * software and its documentation for any purpose is hereby granted,
@@ -16,7 +16,7 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 29-Mar-1994 10:57
- * Last file update:  5-Jun-2003 16:28 (eg)
+ * Last file update: 27-May-2011 22:47 (eg)
  */
 
 #include <unistd.h>
@@ -25,13 +25,24 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <dirent.h>
-#include <time.h> 
+#include <time.h>
 #include "stklos.h"
+#include "struct.h"
 
 #ifndef MAXBUFF
 #  define MAXBUFF 1024
 #endif
 
+#ifdef WIN32
+#  define TIME_DIV_CONST 10000
+#else
+#  define TIME_DIV_CONST 1000
+#endif
+
+
+static SCM exit_procs = STk_nil; /* atexit functions */
+static SCM date_type, time_type;
+
 /******************************************************************************
  *
  * Utilities
@@ -52,24 +63,47 @@ static void error_bad_int_or_out_of_bounds(SCM val)
   STk_error("bad integer ~S (or out of range)", val);
 }
 
-static void error_cannot_copy(SCM f1, SCM f2)
+static void error_win32_primitive(void)
 {
-  STk_error("cannot copy file ~S to ~S", f1, f2);
+  STk_error("Win32 primitive not available on this system");
 }
 
+static void error_posix(SCM obj1, SCM obj2)
+{
+  if (obj1 && obj2)
+    STk_error("%s: ~A ~A", strerror(errno), obj1, obj2);
+  else
+    if (obj1)
+      STk_error("%s: ~A", strerror(errno), obj1);
+    else
+      STk_error("%s", strerror(errno));
+}
 
 static SCM my_access(SCM path, int mode)
 {
   if (!STRINGP(path)) error_bad_path(path);
-  return MAKE_BOOLEAN(access(STRING_CHARS(path), mode) == 0);
+  return MAKE_BOOLEAN(access(STk_expand_file_name(STRING_CHARS(path)), mode) == 0);
 }
 
+static int my_stat(SCM path, struct stat *s)
+{
+  if (!STRINGP(path)) error_bad_path(path);
+  return stat(STk_expand_file_name(STRING_CHARS(path)), s);
+}
+
+static int my_lstat(SCM path, struct stat *s)
+{
+  if (!STRINGP(path)) error_bad_path(path);
+  return lstat(STk_expand_file_name(STRING_CHARS(path)), s);
+}
+
+
 
 int STk_dirp(const char *path)
 {
   struct stat buf;
-  
-  if (stat(path, &buf) >= 0) 
+
+  if (stat(path, &buf) >= 0)
     return S_ISDIR(buf.st_mode);
   return FALSE;
 }
@@ -81,12 +115,12 @@ int STk_dirp(const char *path)
 //EG: {
 //EG:   char *p, *q, dir[MAX_PATH_LENGTH];
 //EG:   struct stat buf;
-//EG:  
+//EG:
 //EG:   if (ISABSOLUTE(exec)) {
 //EG:     strncpy(path, exec, MAX_PATH_LENGTH);
 //EG:     return;
 //EG:   }
-//EG:   
+//EG:
 //EG:   /* the executable path may be specified by relative path from the cwd. */
 //EG:   /* Patch suggested by Shiro Kawai <shiro@squareusa.com> */
 //EG:   if (strchr(exec, DIRSEP) != NULL) {
@@ -95,39 +129,39 @@ int STk_dirp(const char *path)
 //EG:     absolute(dir, path);
 //EG:     return;
 //EG:   }
-//EG: 
-//EG: #ifdef FREEBSD 
+//EG:
+//EG: #ifdef FREEBSD
 //EG:   /* I don't understand why this is needed */
 //EG:   if (access(path, X_OK) == 0) {
 //EG:     stat(path, &buf);
 //EG:     if (!S_ISDIR(buf.st_mode)) return;
-//EG:   }  
+//EG:   }
 //EG: #endif
-//EG: 
+//EG:
 //EG:   p = getenv("PATH");
 //EG:   if (p == NULL) {
 //EG:     p = "/bin:/usr/bin";
 //EG:   }
-//EG: 
+//EG:
 //EG:   while (*p) {
 //EG:     /* Copy the stuck of path in dir */
 //EG:     for (q = dir; *p && *p != PATHSEP; p++, q++) *q = *p;
 //EG:     *q = '\000';
-//EG: 
-//EG:     if (!*dir) { 
+//EG:
+//EG:     if (!*dir) {
 //EG:       /* patch suggested by Erik Ostrom <eostrom@vesuvius.ccs.neu.edu> */
 //EG:       getcwd(path, MAX_PATH_LENGTH);
 //EG:       sprintf(path + strlen(path), "%c%s", DIRSEP, exec);
 //EG:     }
 //EG:     else
 //EG:       sprintf(path, "%s%c%s", dir, DIRSEP, exec);
-//EG: 
+//EG:
 //EG:     sprintf(path, "%s%c%s", dir, DIRSEP, exec);
 //EG:     if (access(path, X_OK) == 0) {
 //EG:       stat(path, &buf);
 //EG:       if (!S_ISDIR(buf.st_mode)) return;
 //EG:     }
-//EG: 	 
+//EG:
 //EG:     /* Try next path */
 //EG:     if (*p) p++;
 //EG:   }
@@ -142,25 +176,87 @@ int STk_dirp(const char *path)
  *
  ******************************************************************************/
 
+/*
+<doc EXT winify-file-name
+ * (winify-file-name fn)
+ *
+ * On Win32 system, when compiled with the Cygwin environment,
+ * file names are internally represented in a POSIX-like internal form.
+ * |Winify-file-bame| permits to obtain back the Win32 name of an interned
+ * file name
+ * @lisp
+ * (winify-file-name "/tmp")
+ *     => "C:\\\\cygwin\\\\tmp"
+ * (list (getcwd) (winify-file-name (getcwd)))
+ *     => ("//saxo/homes/eg/Projects/STklos"
+ *         "\\\\\\\\saxo\\\\homes\\\\eg\\\\Projects\\\\STklos")
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("winify-file-name", winify_filename, subr1, (SCM f))
+{
+#ifdef WIN32
+  char expanded[2 * MAX_PATH_LENGTH];
+
+  if (!STRINGP(f)) error_bad_string(f);
+
+  cygwin_conv_to_win32_path(STRING_CHARS(f), expanded);
+  return STk_Cstring2string(expanded);
+#else
+  error_win32_primitive();
+  return STk_void;
+#endif
+}
+
+
+/*
+<doc EXT posixify-file-name
+ * (posixify-file-name fn)
+ *
+ * On Win32 system, when compiled with the Cygwin environment,
+ * file names are internally represented in a POSIX-like internal form.
+ * |posixify-file-bame| permits to obtain the interned file name from
+ * its external form.
+ * file name
+ * @lisp
+ * (posixify-file-name "C:\\\\cygwin\\\\tmp")
+ *        => "/tmp"
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("posixify-file-name", posixify_filename, subr1, (SCM f))
+{
+#ifdef WIN32
+  char expanded[2 * MAX_PATH_LENGTH];
+
+  if (!STRINGP(f)) error_bad_string(f);
+
+  cygwin_conv_to_posix_path(STRING_CHARS(f), expanded);
+  return STk_Cstring2string(expanded);
+#else
+  error_win32_primitive();
+  return STk_void;
+#endif
+}
+
+
 /*
 <doc EXT expand-file-name
  * (expand-file-name path)
- * 
- * @cindex tilde expansion
+ *
  * |Expand-file-name| expands the filename given in |path| to
- * an absolute path. 
+ * an absolute path.
  * @lisp
- *   ;; Current directory is ~eg/STklos (i.e. /users/eg/STklos)
+ *   ;; Current directory is ~eg/stklos (i.e. /users/eg/stklos)
  *   (expand-file-name "..")            => "/users/eg"
  *   (expand-file-name "~eg/../eg/bin") => "/users/eg/bin"
- *   (expand-file-name "~/STklos)"      => "/users/eg/STk"
- * @end lisp 
+ *   (expand-file-name "~/stklos)"      => "/users/eg/stk"
+ * @end lisp
 doc>
 */
 
 DEFINE_PRIMITIVE("expand-file-name", expand_fn, subr1, (SCM s))
 {
-  ENTER_PRIMITIVE(expand_fn);
   if (!STRINGP(s)) error_bad_string(s);
   return STk_Cstring2string(STk_expand_file_name(STRING_CHARS(s)));
 }
@@ -171,15 +267,13 @@ DEFINE_PRIMITIVE("expand-file-name", expand_fn, subr1, (SCM s))
  * (canonical-file-name path)
  *
  * Expands all symbolic links in |path| and returns its canonicalized
- * absolute path name. The resulting path does not have symbolic links. 
- * If |path| doesn't designate a valid path name, |canonical-file-name| 
- * returns |f|.
+ * absolute path name. The resulting path does not have symbolic links.
+ * If |path| doesn't designate a valid path name, |canonical-file-name|
+ * returns |#f|.
 doc>
 */
 DEFINE_PRIMITIVE("canonical-file-name", canonical_path, subr1, (SCM str))
 {
-  ENTER_PRIMITIVE(canonical_path);
-
   if (!STRINGP(str)) error_bad_string(str);
   return STk_resolve_link(STRING_CHARS(str), 0);
 }
@@ -196,10 +290,9 @@ DEFINE_PRIMITIVE("getcwd", getcwd, subr0, (void))
   char buf[MAX_PATH_LENGTH], *s;
   SCM z;
 
-  ENTER_PRIMITIVE(getcwd);
-
   s = getcwd(buf, MAX_PATH_LENGTH);
-  if (!s) STk_error("cannot determine current directory");
+  if (!s)
+    error_posix(NULL, NULL);
   z = STk_Cstring2string(buf);
 
   return z;
@@ -215,13 +308,44 @@ doc>
 */
 DEFINE_PRIMITIVE("chdir", chdir, subr1, (SCM s))
 {
-  ENTER_PRIMITIVE(chdir);
-
   if (!STRINGP(s)) error_bad_path(s);
-  
-  if (chdir(STk_expand_file_name(STRING_CHARS(s))))
-    STk_error("cannot change directory to ~S", s);
- 
+
+  if (chdir(STk_expand_file_name(STRING_CHARS(s))) != 0)
+    error_posix(s, NULL);
+
+  return STk_void;
+}
+
+/*
+<doc EXT make-directory
+ * (make-directory dir)
+ *
+ * Create a directory with name |dir|.
+doc>
+*/
+DEFINE_PRIMITIVE("make-directory", make_directory, subr1, (SCM path))
+{
+  mode_t mask;
+
+  if (!STRINGP(path)) error_bad_path(path);
+  umask(mask = umask(0));
+  if (mkdir(STRING_CHARS(path), 0777) != 0)
+    error_posix(path, NULL);
+  return STk_void;
+}
+
+/*
+<doc EXT remove-directory
+ * (remove-directory dir)
+ *
+ * Delete the directory with name |dir|.
+doc>
+*/
+DEFINE_PRIMITIVE("remove-directory", remove_directory, subr1, (SCM path))
+{
+  if (!STRINGP(path)) error_bad_path(path);
+  if (rmdir(STRING_CHARS(path)) != 0)
+    error_posix(path, NULL);
   return STk_void;
 }
 
@@ -231,7 +355,7 @@ DEFINE_PRIMITIVE("chdir", chdir, subr1, (SCM s))
  * (getpid)
  *
  * Returns the system process number of the current program (i.e. the
- * Unix @i{pid}) as an integer.
+ * Unix ,(emph "pid")) as an integer.
 doc>
 */
 DEFINE_PRIMITIVE("getpid", getpid, subr0, (void))
@@ -250,8 +374,6 @@ doc>
 */
 DEFINE_PRIMITIVE("system", system, subr1, (SCM com))
 {
-  ENTER_PRIMITIVE(system);
-  
   if (!STRINGP(com)) error_bad_string(com);
   return MAKE_INT(system(STRING_CHARS(com)));
 }
@@ -261,7 +383,7 @@ DEFINE_PRIMITIVE("system", system, subr1, (SCM com))
  * (file-is-directory?  string)
  * (file-is-regular?    string)
  * (file-is-readable?   string)
-x * (file-is-writable?   string)
+ * (file-is-writable?   string)
  * (file-is-executable? string)
  * (file-exists?        string)
  *
@@ -274,11 +396,7 @@ DEFINE_PRIMITIVE("file-is-directory?", file_is_directoryp, subr1, (SCM f))
 {
   struct stat info;
 
-  ENTER_PRIMITIVE(file_is_directoryp);
-
-  if (!STRINGP(f)) error_bad_path(f);
-  if (stat(STRING_CHARS(f), &info) != 0) return STk_false;
-
+  if (my_stat(f, &info) != 0) return STk_false;
   return MAKE_BOOLEAN((S_ISDIR(info.st_mode)));
 }
 
@@ -287,80 +405,91 @@ DEFINE_PRIMITIVE("file-is-regular?", file_is_regularp, subr1, (SCM f))
 {
   struct stat info;
 
-  ENTER_PRIMITIVE(file_is_regularp);
-
-  if (!STRINGP(f)) error_bad_path(f);
-  if (stat(STRING_CHARS(f), &info) != 0) return STk_false;
-
+  if (my_stat(f, &info) != 0) return STk_false;
   return MAKE_BOOLEAN((S_ISREG(info.st_mode)));
 }
 
 
 DEFINE_PRIMITIVE("file-is-readable?", file_is_readablep, subr1, (SCM f))
 {
-  ENTER_PRIMITIVE(file_is_readablep);
   return my_access(f, R_OK);
 }
 
 
 DEFINE_PRIMITIVE("file-is-writable?", file_is_writablep, subr1, (SCM f))
 {
-  ENTER_PRIMITIVE(file_is_writablep);
   return my_access(f, W_OK);
 }
 
 
 DEFINE_PRIMITIVE("file-is-executable?", file_is_executablep, subr1, (SCM f))
 {
-  ENTER_PRIMITIVE(file_is_executablep);
   return my_access(f, X_OK);
 }
 
 
 DEFINE_PRIMITIVE("file-exists?", file_existsp, subr1, (SCM f))
 {
-  ENTER_PRIMITIVE(file_existsp)
-  return my_access(f, F_OK);
+  struct stat info;
+
+  return MAKE_BOOLEAN(my_lstat(f, &info) == 0);
+}
+
+/*
+<doc EXT file-size
+ * (file-size string)
+ *
+ * Returns the size of the file whose path name is given in
+ * |string|.If |string| denotes a file which does not exist,
+ * |file-size| returns |#f|.
+doc>
+ */
+
+DEFINE_PRIMITIVE("file-size", file_size, subr1, (SCM f))
+{
+  struct stat info;
+
+  if (my_stat(f, &info) != 0) return STk_false;
+  return STk_long2integer((long) info.st_size);
 }
 
 /*
 <doc EXT glob
  * (glob pattern ...)
  *
- * |Glob| performs file name ``globbing'' in a fashion similar to the 
+ * |Glob| performs file name ``globbing'' in a fashion similar to the
  * csh shell. |Glob| returns a list of the filenames that match at least
  * one of |pattern| arguments.  The |pattern| arguments may contain
  * the following special characters:
- * @itemize @bullet
- * @item |?| Matches any single character.
- * @item |*| Matches any sequence of zero or more characters.
- * @item |[chars]| Matches any single character in |chars|. 
- * If chars contains a sequence of the form |a-b| then any character 
- * between |a| and |b| (inclusive) will match.
- * @item  |\\| Matches the character |x|.
- * @item |@{a,b,...@}| Matches any of the strings |a|, |b|, etc.
- * @end itemize
- * 
- * As with csh, a '.' at the beginning of a file's name or just after 
- * a '/ must be matched explicitly or with a |@{@}| construct.  
+ * ,(itemize
+ * (item [|?| Matches any single character.])
+ * (item [|*| Matches any sequence of zero or more characters.])
+ * (item [|\[chars\]| Matches any single character in |chars|.
+ * If chars contains a sequence of the form |a-b| then any character
+ * between |a| and |b| (inclusive) will match.])
+ * (item [|\\x| Matches the character |x|.])
+ * (item [|{a,b,...}| Matches any of the strings |a|, |b|, etc.])
+ * )
+ *
+ * As with csh, a '.' at the beginning of a file's name or just after
+ * a '/ must be matched explicitly or with a |@{@}| construct.
  * In addition, all '/' characters must be matched explicitly.
- * 
+ * @l
  * If the first character in a pattern is '~' then it refers to
  * the home directory of the user whose name follows the '~'.
  * If the '~' is followed immediately by `/' then the value of
  * the environment variable HOME is used.
- *
+ * @l
  * |Glob| differs from csh globbing in two ways.  First, it does not
  * sort its result list (use the |sort| procedure if you want the list
  * sorted).
- * Second, |glob| only returns the names of files that actually exist; 
- * in csh no check for existence is made unless a pattern contains a 
- * |?|, |*|, or |@{@}| construct.
+ * Second, |glob| only returns the names of files that actually exist;
+ * in csh no check for existence is made unless a pattern contains a
+ * |?|, |*|, or |\[\]| construct.
 doc>
 */
 DEFINE_PRIMITIVE("glob", glob, vsubr, (int argc, SCM *argv))
 {
-  ENTER_PRIMITIVE(glob);
   return STk_do_glob(argc, argv);
 }
 
@@ -370,47 +499,75 @@ DEFINE_PRIMITIVE("glob", glob, vsubr, (int argc, SCM *argv))
  * (remove-file string)
  *
  * Removes the file whose path name is given in |string|.
- * The result of |remove-file| is @emph{void}.
+ * The result of |remove-file| is ,(emph "void").
 doc>
 */
 DEFINE_PRIMITIVE("remove-file", remove_file, subr1, (SCM filename))
 {
-  ENTER_PRIMITIVE(remove_file);
-  
   if (!STRINGP(filename)) error_bad_string(filename);
   if (remove(STRING_CHARS(filename)) != 0)
-    STk_error("cannot remove ~S", filename);
+    error_posix(filename, NULL);
   return STk_void;
 }
- 
+
 
 /*
 <doc EXT rename-file
  * (rename-file string1 string2)
  *
  * Renames the file whose path-name is |string1| to a file whose path-name is
- * |string2|. The result of |rename-file| is @emph{void}.
+ * |string2|. The result of |rename-file| is ,(emph "void").
 doc>
 */
 DEFINE_PRIMITIVE("rename-file", rename_file, subr2, (SCM filename1, SCM filename2))
 {
-  ENTER_PRIMITIVE(rename_file);
-  
   if (!STRINGP(filename1)) error_bad_string(filename1);
   if (!STRINGP(filename2)) error_bad_string(filename2);
   if (rename(STRING_CHARS(filename1), STRING_CHARS(filename2)) != 0)
-    STk_error("cannot rename file ~S in ~S", filename1, filename2);
+    error_posix(filename1, filename2);
   return STk_void;
 }
- 
+
+
+/*
+<doc EXT directory-files
+ * (directory-files path)
+ *
+ * Returns the list of the files in the directory |path|. Directories ,(q ".")
+ * and ,(q "..") don't appear in the result.
+doc>
+*/
+DEFINE_PRIMITIVE("directory-files", directory_files, subr1, (SCM dirname))
+{
+  char *path;
+  DIR* dir;
+  SCM res = STk_nil;
+  struct dirent *d;
+
+  if (!STRINGP(dirname)) error_bad_string(dirname);
+  path = STk_expand_file_name(STRING_CHARS(dirname));
+  dir  = opendir(path);
+  if (!dir) error_posix(dirname, NULL);
+
+  for (d = readdir(dir); d ; d = readdir(dir)) {
+    if (d->d_name[0] == '.')
+      if ((d->d_name[1] == '\0') || (d->d_name[1] == '.' && d->d_name[2] == '\0'))
+	continue;
+    res = STk_cons(STk_Cstring2string(d->d_name), res);
+  }
+  closedir(dir);
+  return STk_dreverse(res);
+}
+
+
 
 /*
 <doc EXT copy-file
  * (copy-file string1 string2)
  *
  * Copies the file whose path-name is |string1| to a file whose path-name is
- * |string2|. If the file |string2| already exists, its content prior 
- * the call to |copy-file| is lost. The result of |copy-file| is @emph{void}.
+ * |string2|. If the file |string2| already exists, its content prior
+ * the call to |copy-file| is lost. The result of |copy-file| is ,(emph "void").
 doc>
 */
 DEFINE_PRIMITIVE("copy-file", copy_file, subr2, (SCM filename1, SCM filename2))
@@ -419,25 +576,21 @@ DEFINE_PRIMITIVE("copy-file", copy_file, subr2, (SCM filename1, SCM filename2))
   int f1, f2, n;
 
   /* Should I use sendfile on Linux here? */
-  ENTER_PRIMITIVE(copy_file);
-  
   if (!STRINGP(filename1)) error_bad_string(filename1);
   if (!STRINGP(filename2)) error_bad_string(filename2);
-  
+
   f1 = open(STRING_CHARS(filename1), O_RDONLY);
+  if (f1 == -1) error_posix(filename1, NULL);
   f2 = open(STRING_CHARS(filename2), O_WRONLY|O_CREAT|O_TRUNC, 0666);
-  
-  if ((f1==-1) || (f2==-1)) {
-    error_cannot_copy(filename1, filename2);
-  }
+  if (f2 == -1) error_posix(filename2, NULL);
 
   while ((n = read(f1, buff, MAXBUFF)) > 0) {
     if ((n < 0) || (write(f2, buff, n) < n)) {
-      close(f1); close(f2); 
-      error_cannot_copy(filename1, filename2);
+      close(f1); close(f2);
+      error_posix(filename1, filename2);
     }
   }
-  
+
   close(f1); close(f2);
   return STk_void;
 }
@@ -456,57 +609,99 @@ DEFINE_PRIMITIVE("temporary-file-name", tmp_file, subr0, (void))
 {
 #ifdef WIN32
   char buff[MAX_PATH_LENGTH], *s;
-  
+
   s = tmpnam(buff);
   return s ? STk_Cstring2string(s) : STk_false;
 #else
   static int cpt=0;
   char buff[MAX_PATH_LENGTH];
+  MUT_DECL(tmpnam_mutex);
 
+  MUT_LOCK(tmpnam_mutex);
   for ( ; ; ) {
     sprintf(buff, "/tmp/stklos%05x", cpt++);
     if (cpt > 100000)		/* arbitrary limit to avoid infinite search */
-      return STk_false; 
+      return STk_false;
     if (access(buff, F_OK) == -1) break;
   }
+  MUT_UNLOCK(tmpnam_mutex);
 
   return STk_Cstring2string(buff);
 #endif
 }
 
 
+/*
+<doc EXT register-exit-function!
+ * (register-exit-function! proc)
+ *
+ * This function registers |proc| as an exit function. This function will
+ * be called when the program exits. When called, |proc| will be passed one
+ * parmater which is the status given to the |exit| function (or 0 if the
+ * programe terminates normally). The result of  |register-exit-function!|
+ * is undefined.
+ * @lisp
+ * (let* ((tmp (temporary-file-name))
+ *        (out (open-output-file tmp)))
+ *   (register-exit-function! (lambda (n)
+ *                              (when (zero? n)
+ *                                (remove-file tmp))))
+ *   out)
+ * @end lisp
+doc>
+*/
+MUT_DECL(at_exit_mutex);	 /* The exit mutex */
+
+DEFINE_PRIMITIVE("register-exit-function!", at_exit, subr1, (SCM proc))
+{
+  if (STk_procedurep(proc) == STk_false) STk_error("bad procedure ~S", proc);
+
+  MUT_LOCK(at_exit_mutex);
+  exit_procs = STk_cons(proc, exit_procs);
+  MUT_UNLOCK(at_exit_mutex);
+  return STk_void;
+}
+
+
+DEFINE_PRIMITIVE("%pre-exit", pre_exit, subr1, (SCM retcode))
+{
+  /* Execute the at-exit handlers */
+  MUT_LOCK(at_exit_mutex);
+  for (  ; !NULLP(exit_procs); exit_procs = CDR(exit_procs))
+    STk_C_apply(CAR(exit_procs), 1, retcode);
+  MUT_UNLOCK(at_exit_mutex);
+
+  /* Flush all bufers */
+  STk_close_all_ports();
+
+  return STk_void;
+}
+
 /*
 <doc EXT exit
- * (exit) 
+ * (exit)
  * (exit ret-code)
  *
  * Exits the program with the specified integer return code. If |ret-code|
  * is omitted, the program terminates with a return code of 0.
+ * If  program has registered exit functions with |register-exit-function!|,
+ * they are called (in an order which is the reverse of their call order).
 doc>
 */
-DEFINE_PRIMITIVE("exit", quit, subr01, (SCM retcode))
+DEFINE_PRIMITIVE("exit", exit, subr01, (SCM retcode))
 {
   long ret = 0;
-  
-  ENTER_PRIMITIVE(quit);
 
   if (retcode) {
     ret = STk_integer_value(retcode);
     if (ret == LONG_MIN) STk_error("bad return code ~S", retcode);
+  } else {
+    retcode = MAKE_INT(0);
   }
-#ifdef FIXME
-//EG:  /* Execute all the terminal thunks of pending dynamic-wind */
-//EG:  STk_unwind_all();
-//EG:
-//EG:  /* call user finalization code */
-//EG:  STk_user_cleanup();
-//EG:
-//EG:#if defined(WIN32) && defined(USE_SOCKET)
-//EG:  /* Unregister the interpreter from Winsock */
-//EG:  WSACleanup();  
-//EG:#endif
-#endif
+
+  STk_pre_exit(retcode);
   exit(ret);
+
   return STk_void; /* never reached */
 }
 
@@ -516,7 +711,7 @@ DEFINE_PRIMITIVE("exit", quit, subr01, (SCM retcode))
  * (machine-type)
  *
  * Returns a string identifying the kind of machine which is running the
- * program. The result string is of the form 
+ * program. The result string is of the form
  * |[os-name]-[os-version]-[processor-type]|.
 doc>
 */
@@ -527,35 +722,21 @@ DEFINE_PRIMITIVE("machine-type", machine_type, subr0, (void))
 
 
 #ifdef FIXME
-//EG: PRIMITIVE STk_random(SCM n)
-//EG: {
-//EG:   if (NEXACTP(n) || STk_negativep(n) == STk_true || STk_zerop(n) == STk_true)
-//EG:     Err("random: bad number", n);
-//EG:   return STk_modulo(STk_makeinteger(rand()), n);
-//EG: }
-//EG: 
-//EG: PRIMITIVE STk_set_random_seed(SCM n)
-//EG: {
-//EG:   if (NEXACTP(n)) Err("set-random-seed!: bad number", n);
-//EG:   srand((unsigned int) STk_integer_value_no_overflow(n));
-//EG:   return STk_unsepecified;
-//EG: }
-//EG: 
 //EG: #ifndef HZ
 //EG: #define HZ 60.0
 //EG: #endif
-//EG: 
+//EG:
 //EG: #ifdef CLOCKS_PER_SEC
 //EG: #  define TIC CLOCKS_PER_SEC
-//EG: #else 
+//EG: #else
 //EG: #  define TIC HZ
 //EG: #endif
-//EG: 
+//EG:
 //EG: PRIMITIVE STk_get_internal_info(void)
 //EG: {
 //EG:   SCM z = STk_makevect(7, STk_nil);
 //EG:   long allocated, used, calls;
-//EG: 
+//EG:
 //EG:   /* The result is a vector which contains
 //EG:    *	0 The total cpu used in ms
 //EG:    *	1 The number of cells currently in use.
@@ -565,9 +746,9 @@ DEFINE_PRIMITIVE("machine-type", machine_type, subr0, (void))
 //EG:    *    5 Total time used in the gc
 //EG:    *	6 A boolean indicating if Tk is initialized
 //EG:    */
-//EG: 
+//EG:
 //EG:   STk_gc_count_cells(&allocated, &used, &calls);
-//EG: 
+//EG:
 //EG:   VECT(z)[0] = STk_makenumber(STk_my_time());
 //EG:   VECT(z)[1] = STk_makeinteger(used);
 //EG:   VECT(z)[2] = STk_makeinteger(allocated);
@@ -579,7 +760,7 @@ DEFINE_PRIMITIVE("machine-type", machine_type, subr0, (void))
 //EG: #else
 //EG:   VECT(z)[6] = STk_false;
 //EG: #endif
-//EG:   
+//EG:
 //EG:   STk_alloc_cells = 0;
 //EG:   return z;
 //EG: }
@@ -587,7 +768,25 @@ DEFINE_PRIMITIVE("machine-type", machine_type, subr0, (void))
 
 
 /*
-<doc EXT clock 
+<doc EXT date
+ * (date)
+ *
+ * Returns the current date in a string
+doc>
+*/
+DEFINE_PRIMITIVE("date", date, subr0, (void))
+{
+  time_t t   = time(NULL);
+  char * res;
+
+  res = ctime(&t);
+  res[strlen(res)-1] = '\0'; /* to "delete" the newline added by asctime */
+  return STk_Cstring2string(res);
+}
+
+
+/*
+<doc EXT clock
  * (clock)
  *
  * Returns an approximation of processor time, in milliseconds, used so far by the
@@ -596,163 +795,163 @@ doc>
  */
 DEFINE_PRIMITIVE("clock", clock, subr0, (void))
 {
-  return STk_double2real((double) clock() / CLOCKS_PER_SEC * 1000.0);
+  return STk_double2real((double) clock() /
+			 CLOCKS_PER_SEC * (double) TIME_DIV_CONST);
 }
 
 /*
-<doc EXT current-time 
- * (current-time)
+<doc EXT current-seconds
+ * (current-seconds)
  *
- * Returns the time since the Epoch (that is 00:00:00 UTC, January 1, 1970), 
+ * Returns the time since the Epoch (that is 00:00:00 UTC, January 1, 1970),
  * measured in seconds.
 doc>
  */
-DEFINE_PRIMITIVE("current-time", current_time, subr0, (void))
+DEFINE_PRIMITIVE("current-seconds", current_seconds, subr0, (void))
 {
   return STk_ulong2integer(time(NULL));
 }
 
 /*
-<doc EXT full-current-time
- * (full-current-time)
- * 
- * Returns the time of the day as a pair where 
- * @itemize -
- * @item the  first element is the time since the Epoch 
- * (that is 00:00:00 UTC, January 1, 1970),  measured in seconds. 
- * 
- * @item the second element is the number of microseconds in the given 
- * second.
- * @end itemize
+<doc current-time
+ * (current-time)
+ *
+ * Returns a time object corresponding to the current time.
 doc>
 */
-DEFINE_PRIMITIVE("full-current-time", full_current_time, subr0, (void))
+DEFINE_PRIMITIVE("current-time", current_time, subr0, (void))
 {
   struct timeval now;
-  gettimeofday(&now, NULL); 
+  SCM argv[3];
 
-  return STk_cons(STk_long2integer(now.tv_sec),
-		  STk_long2integer(now.tv_usec));
+  gettimeofday(&now, NULL);
+
+  argv[2] = time_type;
+  argv[1] =  STk_long2integer(now.tv_sec);
+  argv[0] =  STk_long2integer(now.tv_usec);
+  return STk_make_struct(3, &argv[2]);
 }
-		  
+
+
+
+/*
+<doc EXT sleep
+ * (sleep n)
+ *
+ * Suspend the execution of the program for at |ms| milliseconds. Note that due
+ * to system clock resolution, the pause may be a little bit longer. If a
+ * signal arrives during the pause, the execution may be resumed.
+ *
+doc>
+*/
+DEFINE_PRIMITIVE("sleep", sleep, subr1, (SCM ms))
+{
+  long n = STk_integer_value(ms);
+  struct timespec ts;
+
+  if (n == LONG_MIN)
+    error_bad_int_or_out_of_bounds(ms);
+
+  ts.tv_sec  = n / 1000;
+  ts.tv_nsec = (n % 1000) * 1000000;
+
+  nanosleep(&ts, NULL);
+  return STk_void;
+}
+
+
 /*
 <doc EXT seconds->date
- * (seconds->date sec)
- * 
- * Returns a keyword list for the date given by |sec| (a date based on the 
- * Epoch). The keyed values returned are
- * @itemize -
- * @item second : 0 to 59 (but can be up to 61 to allow for leap seconds)
- * @item minute : 0 to 59
- * @item hour : 0 to 23
- * @item day : 1 to 31
- * @item month : 1 to 12
- * @item year : e.g., 2002
- * @item week-day : 0 (Sunday) to 6 (Saturday)
- * @item year-day : 0 to 365 (365 in leap years)
- * @item dst? : #t (daylight savings time) or #f
- * @item time-zone-offset : the difference between Coordinated Universal Time
- * (UTC) and local standard time in seconds.
- * @end itemize
- * @noindent
- * Example:
- * @lisp
- * (seconds->date (current-time))   => (:second 49 :minute 32 :hour 23
- *                                      :day 2 :month 4 :year 2002 
- *                                      :week-day 2 :year-day 91 
- *                                      :dst #t :time-zone-offset -3600)
- * @end lisp
+ * (seconds->date n)
+ *
+ * Convert the date |n| expressed as a number of seconds since the Epoch
+ * to a date.
 doc>
 */
 DEFINE_PRIMITIVE("%seconds->date", seconds2date, subr1, (SCM seconds))
 {
-  int overflow; 
-  SCM v, *vect;
+  int overflow;
+  SCM argv[11];
   struct tm *t;
   time_t tt;
 
-  ENTER_PRIMITIVE(seconds2date);
   tt = (time_t) STk_integer2int32(seconds, &overflow);
-  
+
   if (overflow) error_bad_int_or_out_of_bounds(seconds);
-  
-  t        = localtime(&tt);
 
-  v 	   = STk_makevect(20, STk_false);
-  vect     = VECTOR_DATA(v);
-  
-  vect[0]  = STk_makekey("second");
-  vect[1]  = MAKE_INT(t->tm_sec);
-  vect[2]  = STk_makekey("minute");
-  vect[3]  = MAKE_INT(t->tm_min);
-  vect[4]  = STk_makekey("hour");
-  vect[5]  = MAKE_INT(t->tm_hour);
-  vect[6]  = STk_makekey("day");
-  vect[7]  = MAKE_INT(t->tm_mday);
-  vect[8]  = STk_makekey("month");
-  vect[9]  = MAKE_INT(t->tm_mon + 1);
-  vect[10] = STk_makekey("year");
-  vect[11] = MAKE_INT(1900 + t->tm_year);
-  vect[12] = STk_makekey("week-day");
-  vect[13] = MAKE_INT(t->tm_wday);
-  vect[14] = STk_makekey("year-day");
-  vect[15] = MAKE_INT(t->tm_yday);
-  vect[16] = STk_makekey("dst");
-  vect[17] = MAKE_BOOLEAN(t->tm_isdst > 0);
-  vect[18] = STk_makekey("time-zone-offset");
-  vect[19] = STk_long2integer(timezone);
-
-  return STk_vector2list(v);
+  t = localtime(&tt);
+  argv[10]  = date_type;
+  argv[9]  = MAKE_INT(t->tm_sec);
+  argv[8]  = MAKE_INT(t->tm_min);
+  argv[7]  = MAKE_INT(t->tm_hour);
+  argv[6]  = MAKE_INT(t->tm_mday);
+  argv[5]  = MAKE_INT(t->tm_mon + 1);
+  argv[4]  = MAKE_INT(t->tm_year + 1900);
+  argv[3]  = MAKE_INT(t->tm_wday);
+  argv[2]  = MAKE_INT(t->tm_yday + 1);
+  argv[1]  = MAKE_INT(t->tm_isdst);
+#ifdef DARWIN
+  argv[0]  = MAKE_INT(0);	/* Cannot figure how to find the timezone */
+#else
+  argv[0]  = STk_long2integer(timezone);
+#endif
+  return STk_make_struct(11, &argv[10]);
 }
 
-DEFINE_PRIMITIVE("%date->seconds", date2seconds, subr1, (SCM v))
+
+/*
+<doc EXT date->seconds
+ * (date->seconds d)
+ *
+ * Convert the date |d| to the number of seconds since the ,(emph "Epoch").
+doc>
+*/
+DEFINE_PRIMITIVE("date->seconds", date2seconds, subr1, (SCM date))
 {
   struct tm t;
-  SCM *vect;
   time_t n;
+  SCM *p;
 
-  ENTER_PRIMITIVE(date2seconds);
-  
-  if (!(VECTORP(v) && VECTOR_SIZE(v) == 7)) STk_error("bad date vector ~S", v);
-  vect = VECTOR_DATA(v);
+  if (!STRUCTP(date) || STRUCT_TYPE(date) != date_type)
+    STk_error("bad date ~S", date);
 
-  t.tm_sec   = STk_integer_value(vect[0]);
-  t.tm_min   = STk_integer_value(vect[1]);
-  t.tm_hour  = STk_integer_value(vect[2]);
-  t.tm_mday  = STk_integer_value(vect[3]);
-  t.tm_mon   = STk_integer_value(vect[4]) - 1;
-  t.tm_year  = STk_integer_value(vect[5]) - 1900;
-  t.tm_isdst = (vect[6] != STk_false);
+  p = (SCM *) &(STRUCT_SLOTS(date));
+  t.tm_sec   = STk_integer_value(*p++);
+  t.tm_min   = STk_integer_value(*p++);
+  t.tm_hour  = STk_integer_value(*p++);
+  t.tm_mday  = STk_integer_value(*p++);
+  t.tm_mon   = STk_integer_value(*p++) - 1;
+  t.tm_year  = STk_integer_value(*p++) - 1900;
+  t.tm_isdst = -1;			/* to ignore DST */
 
   n = mktime(&t);
-  if (n == (time_t)(-1)) STk_error("cannot convert date to seconds (~S)", v);
-  
-  return STk_ulong2integer((long) n);
+  if (n == (time_t)(-1)) STk_error("cannot convert date to seconds (~S)", date);
+
+  return STk_double2real((double) n);
 }
 
 
+
 DEFINE_PRIMITIVE("%seconds->string", date2string, subr2, (SCM fmt, SCM seconds))
 {
   char buffer[1024];
   struct tm *p;
-  time_t tt;  
+  time_t tt;
   int len, overflow;
 
-  ENTER_PRIMITIVE(date2string);
-  
   tt = (time_t) STk_integer2int32(seconds, &overflow);
-  
+
   if (!STRINGP(fmt)) error_bad_string(fmt);
   if (overflow)      error_bad_int_or_out_of_bounds(seconds);
-  
+
   p   = localtime(&tt);
   len = strftime(buffer, 1023, STRING_CHARS(fmt), p);
 
   if (len > 0)
     return STk_Cstring2string(buffer);
-  else 
+  else
     STk_error("buffer too short!");
-  
+
   return STk_void; /* never reached */
 }
 
@@ -760,17 +959,21 @@ DEFINE_PRIMITIVE("%seconds->string", date2string, subr2, (SCM fmt, SCM seconds))
 /*
 <doc EXT running-os
  * (running-os)
- * 
- * Returns the name of the underlying Operating System which is running 
- * the program. 
- * The value returned by |runnin-os| is a symbol. For now, this procedure 
- * returns either |unix| or |windows|.
+ *
+ * Returns the name of the underlying Operating System which is running
+ * the program.
+ * The value returned by |runnin-os| is a symbol. For now, this procedure
+ * returns either |unix|, |windows|, or |cygwin-windows|.
 doc>
 */
 DEFINE_PRIMITIVE("running-os", running_os, subr0, (void))
 {
 #ifdef WIN32
+# ifdef __CYGWIN32__
+  return STk_intern("cygwin-windows");
+# else
   return STk_intern("windows");
+# endif
 #else
   return STk_intern("unix");
 #endif
@@ -785,11 +988,11 @@ DEFINE_PRIMITIVE("running-os", running_os, subr0, (void))
  *
  * Looks for the environment variable named |str| and returns its
  * value as a string, if it exists. Otherwise, |getenv| returns |#f|.
- * If |getenv| is called without parameter, it returns the list of 
- * all the environment variables accessible from the program as an 
+ * If |getenv| is called without parameter, it returns the list of
+ * all the environment variables accessible from the program as an
  * A-list.
  * @lisp
- * (getenv "SHELL")   
+ * (getenv "SHELL")
  *      => "/bin/zsh"
  * (getenv)
  *      => (("TERM" . "xterm") ("PATH" . "/bin:/usr/bin") ...)
@@ -814,11 +1017,10 @@ static SCM build_posix_environment(char **env)
 DEFINE_PRIMITIVE("getenv", getenv, subr01, (SCM str))
 {
   char *tmp;
-  
-  ENTER_PRIMITIVE(getenv);
+
   if (str) {		/* One parameter: find the value of the given variable */
     if (!STRINGP(str)) error_bad_string(str);
-    
+
     tmp = getenv(STRING_CHARS(str));
     return tmp ? STk_Cstring2string(tmp) : STk_false;
   } else {		/* No parameter: give the complete environment */
@@ -832,23 +1034,73 @@ DEFINE_PRIMITIVE("getenv", getenv, subr01, (SCM str))
  * (setenv! var value)
  *
  * Sets the environment variable |var| to |value|. |Var| and
- * |value| must be strings. The result of |setenv!| is @emph{void}.
+ * |value| must be strings. The result of |setenv!| is ,(emph "void").
 doc>
  */
 DEFINE_PRIMITIVE("setenv!", setenv, subr2, (SCM var, SCM value))
 {
   char *s;
-  ENTER_PRIMITIVE(setenv);
-  if (!STRINGP(var)) 		      STk_error("variable ~S is not a string", var);
+
+  if (!STRINGP(var)) 		      error_bad_string(var);
   if (strchr(STRING_CHARS(var), '=')) STk_error("variable ~S contains a '='", var);
   if (!STRINGP(value)) 		      STk_error("value ~S is not a string", value);
 
-  s = STk_must_malloc(strlen(STRING_CHARS(var))   + 
+  s = STk_must_malloc(strlen(STRING_CHARS(var))   +
 		      strlen(STRING_CHARS(value)) + 2); /* 2 because of '=' & \0 */
   sprintf(s, "%s=%s", STRING_CHARS(var), STRING_CHARS(value));
   putenv(s);
   return STk_void;
 }
+/*
+<doc EXT unsetenv!
+ * (unsetenv! var)
+ *
+ * Unsets the environment variable |var|. |Var| must be a string.
+ * The result of |unsetenv!| is ,(emph "void").
+doc>
+ */
+DEFINE_PRIMITIVE("unsetenv!", unsetenv, subr1, (SCM var))
+{
+#ifndef SOLARIS
+  if (!STRINGP(var)) error_bad_string(var);
+  unsetenv(STRING_CHARS(var));
+  return STk_void;
+#else
+  /* not exactly the same since getenv will not return #f after that */
+  return STk_setenv(var, STk_Cstring2string(""));
+#endif
+}
+
+/*
+<doc EXT hostname
+ * (hostname)
+ *
+ * Return the  host name of the current processor as a string.
+doc>
+*/
+DEFINE_PRIMITIVE("hostname", hostname, subr0, (void))
+{
+  char buff[256];
+
+  if (gethostname(buff, 256) < 0)
+    buff[255] = '0';
+  return STk_Cstring2string(buff);
+}
+
+/*
+<doc EXT pause
+ * (pause)
+ *
+doc>
+*/
+DEFINE_PRIMITIVE("pause", pause, subr0, (void))
+{
+  pause();
+  return STk_void;
+}
+
+
+
 
 
 /*
@@ -861,32 +1113,96 @@ DEFINE_PRIMITIVE("%library-prefix", library_prefix, subr0, (void))
   return STk_Cstring2string(PREFIXDIR);
 }
 
+DEFINE_PRIMITIVE("%shared-suffix", shared_suffix, subr0, (void))
+{
+  return STk_Cstring2string(SHARED_SUFFIX);
+}
+
+DEFINE_PRIMITIVE("%shared-library-suffix", shared_library_suffix, subr0, (void))
+{
+  return STk_Cstring2string(SHARED_LIB_SUFFIX);
+}
+
+DEFINE_PRIMITIVE("%chmod", change_mode, subr2, (SCM file, SCM value))
+{
+  long mode = STk_integer_value(value);
+
+  if (!STRINGP(file))   error_bad_path(file);
+  if (mode < 0 || mode > 0777) error_bad_int_or_out_of_bounds(value);
+
+  return MAKE_BOOLEAN(chmod(STRING_CHARS(file), mode) == 0);
+}
+
+
+DEFINE_PRIMITIVE("%big-endian?", big_endianp, subr0, (void))
+{
+  int i = 1;
+  char *p = (char *)&i;
+
+  return MAKE_BOOLEAN(p[0] != 1);
+}
 
 
 int STk_init_system(void)
 {
+  SCM current_module = STk_STklos_module;
+
+  /* Create the system-date structure-type */
+  date_type =  STk_make_struct_type(STk_intern("%date"),
+				    STk_false,
+				    LIST10(STk_intern("second"),
+					   STk_intern("minute"),
+					   STk_intern("hour"),
+					   STk_intern("day"),
+					   STk_intern("month"),
+					   STk_intern("year"),
+					   STk_intern("week-day"),
+					   STk_intern("year-day"),
+					   STk_intern("dst"),
+					   STk_intern("tz")));
+  STk_define_variable(STk_intern("%date"), date_type, current_module);
+
+  /* Create the time structure-type */
+  time_type =  STk_make_struct_type(STk_intern("%time"),
+				    STk_false,
+				    LIST2(STk_intern("second"),
+					  STk_intern("microsecond")));
+  STk_define_variable(STk_intern("%time"), time_type, current_module);
+
+  /* Declare primitives */
   ADD_PRIMITIVE(clock);
+  ADD_PRIMITIVE(date);
+  ADD_PRIMITIVE(current_seconds);
   ADD_PRIMITIVE(current_time);
-  ADD_PRIMITIVE(full_current_time);
+  ADD_PRIMITIVE(sleep);
   ADD_PRIMITIVE(seconds2date);
   ADD_PRIMITIVE(date2seconds);
   ADD_PRIMITIVE(date2string);
   ADD_PRIMITIVE(running_os);
   ADD_PRIMITIVE(getenv);
   ADD_PRIMITIVE(setenv);
+  ADD_PRIMITIVE(unsetenv);
+  ADD_PRIMITIVE(hostname);
   ADD_PRIMITIVE(library_prefix);
+  ADD_PRIMITIVE(shared_suffix);
+  ADD_PRIMITIVE(shared_library_suffix);
+  ADD_PRIMITIVE(change_mode);
 
   ADD_PRIMITIVE(getcwd);
   ADD_PRIMITIVE(chdir);
+  ADD_PRIMITIVE(make_directory);
+  ADD_PRIMITIVE(remove_directory);
+  ADD_PRIMITIVE(directory_files);
   ADD_PRIMITIVE(getpid);
   ADD_PRIMITIVE(system);
-    
+
   ADD_PRIMITIVE(file_is_directoryp);
   ADD_PRIMITIVE(file_is_regularp);
   ADD_PRIMITIVE(file_is_readablep);
   ADD_PRIMITIVE(file_is_writablep);
   ADD_PRIMITIVE(file_is_executablep);
   ADD_PRIMITIVE(file_existsp);
+  ADD_PRIMITIVE(file_size);
   ADD_PRIMITIVE(glob);
   ADD_PRIMITIVE(expand_fn);
   ADD_PRIMITIVE(canonical_path);
@@ -895,8 +1211,15 @@ int STk_init_system(void)
   ADD_PRIMITIVE(rename_file);
   ADD_PRIMITIVE(copy_file);
   ADD_PRIMITIVE(tmp_file);
-  ADD_PRIMITIVE(quit);
+  ADD_PRIMITIVE(pre_exit);
+  ADD_PRIMITIVE(exit);
+  ADD_PRIMITIVE(at_exit);
   ADD_PRIMITIVE(machine_type);
 
+  ADD_PRIMITIVE(winify_filename);
+  ADD_PRIMITIVE(posixify_filename);
+
+  ADD_PRIMITIVE(pause);
+  ADD_PRIMITIVE(big_endianp);
   return TRUE;
 }
